/*
**  ADDEL
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
**  {mis,Mddel}=addel(X,vsz,prior,quadr,h)  To compute deletion
**                                          diagnostics for linear
**                                          and/or quadratic
**                                          discriminant analysis
**
*/


/*
**
** ADDEL
**
** PURPOSE:  to compute deletion diagnostics for
**           LDA or QDA
**
**
**
** FORMAT:   {mis,MDdel}=addel(X,vsz,prior,quadr,h)
**
** INPUT:     X     = (nn1+nn2+...+nng) x p matrix containing
**                    the units belonging to the different universes
**            vsz   = g x 1 (where g is the number of groups) vector
**                   which contains the sizes of the
**                   groups
**                   e.g.: if vsz =50|40|30
**            prior = g x 1 vector containing prior probabilities
**            quadr = scalar dummy variable
**                   if discr <> 1 QDA is produced
**                   else  LDA is used
**            h    = scalar dummy variable
**                   if d <> 1 no plot is produced
**                   if d=1 a plot of each column of matrix
**                   scodel  calculated
**                   (potential outlier is a unit which
**                    shows a sharp drop in the plot)
** OUTPUT:    mis   = N x 3 matrix
**                   1st col. contains the indexes of the
**                   units,
**                   2nd col. difference in overall miscl. prob.
**                   3rd col. difference in individual miscl. prob.
**           MDdel =
**                   1st col. contains the indexes of the
**                   units,
**                   2nd col. deletion Mahalanobis distances
**                   MD for unit i when unit i is not in the sample
**                   3rd col. Generalization of Wilks statistics
**                   to the case of multiple samples
**                   (Within group deviance matrix is used)
**                   det(W_(j))/det(W)  j=1, ..., n
**
** REMARK:   This method is very good when only one outlier is present
**           in the data, but it can suffer masking and swamping
**           problems when there are several outliers.
**           For robust methods see procedures FWDLDA.g and FWDQDA.g
**
**     
**
*/


proc (2)=ADdel(X,vsz,prior,quadr,h);
      local seq,i,id,ident,n,bsb,Xd,j,mis,g,vsz1,postT,
            sum,sz,post1,lnd,M,MD,eqva,modsz,MDdel,WT,BT,W,B;

  n=rows(X); g=rows(vsz);
  seq=seqa(1,1,n);

           id=0;
           i=1; do until i==g+1;
           id=id|reshape(i,vsz[i],1);
             i=i+1;
           endo;
          @ Ident Identifier vector  @
          Ident=id[2:n+1];
          vsz1=0|vsz|10^40;
   mis=seq~miss(zeros(n,2),0);
   Mddel=mis;

  {WT,BT}=vardec(X,vsz);

    if quadr==1;
      {postT}=qdasimpl(X,vsz,prior);
    else;
      {postT}=ldasimpl(X,vsz,prior);
    endif;
   sum=sumc(1-postT);


        vsz1=0|vsz;
 i=1;    do until i==n+1;
    bsb=setdif(seq,i,1);
    Xd=X[bsb,.];  sz=vsz;
     j=1; do until j==g+1;
       if i>sumc(vsz1[1:j]) and i<=sumc(vsz1[1:j+1]);
         sz[j]=vsz[j]-1;
         goto modsz;
       else;
       endif;
        j=j+1;
    endo;
       modsz:
      @   "III";;   i;
                   sz; wait;    @
     if quadr==1;
        {post1,Lnd,M,MD,eqva}=qda(Xd,sz,X,vsz,prior);
     else;
        {post1,Lnd,M,MD,eqva}=lda(Xd,sz,X,vsz,prior);
     endif;

      {W,B}=vardec(Xd,sz);

      mis[i,2:3]=(sum-sumc(1-post1))~(postT[i]-post1[i]);
      MDdel[i,2:3]=sqrt(MD[i])~det(W);

    i=i+1;
  endo;

      @ In third col. we have Wilks test @
       Mddel[.,3]=MDdel[.,3]/det(Wt);

    if h==1;
       _pdate="";
          title("Deletion values for overall miscl. prob.");
      j=2; do until j==4;
           iplot(mis[.,j],0);
          title("Deletion values for individual miscl. prob.");
      j=j+1;
     endo;
          title("Deletion Mahalanobis distances");
           iplot(MDdel[.,2],0);

          title("Wilks test");
           iplot(MDdel[.,3],0);
    else;
    endif;

      retp(mis,MDdel);
    endp;

