/*
**   ANOVA.G
**   written by MARCO RIANI
**
**  FORMAT                      PURPOSE                     Line
** =======================================================================
** {tst}=ANOVA(X,sz,d)         Tests for                         17
**                           equality of multivariate
**                           means (given that var cov
**                           of the different groups are
**                           equal (Mardia et al. 1979:
**                           pag. 138))
*/

/*
**
** ANOVA
**
** PURPOSE:  To test the hypothesis of equal mean among groups
**
** FORMAT:   {tst}=ANOVA(X,sz,d)
**
** INPUT:     X = n x p matrix of data
**           sz = g x 1 vector which contains the sizes of the groups
**            d = scalar.
**                if d=1  program displays on the screen the value of
**                the tests
**                else nothing is printed on the screen
**
** OUTPUT:    tst = 3 x 1 vector
**                 tst[1]= Wilks Lambda  det(W)/(det(W+B))
**                       W=Within groups deviance matrix
**                       B=Between groups deviance matrix
**                         See Rencher (1995) p. 190.
**                 tst[2] = Likelihood ratio test
**                          -((n-g)-(p-g+2)/2)*ln( tst[1] );  @ L =
**                          See (Mardia et al. 1979: pag. 138))
**                 tst[3] = genralized Hotelling multiplied
**                          by (n-g)/(g-1);
**                          If g==2 tst[3] is the usual Hotelling
**                          statistic
**                          See (Rencher p. 188)
**
** REMARK: the output on the screen is written using Rencher notation
**         In other terms
**         v_E=n-g       v_H=g-1
**
** See also: VARDEC
**
*/


proc (1)=anova(X,sz,d);
 local W,B,la,g,T,n,p,va,v_H,v_E,GH,tst,l;
    n=rows(X);
    p=cols(X);
    g=rows(sz);
    v_H=g-1;
    v_E=sumc(sz)-g;

 {W,B}=vardec(X,sz);
 T=W+B;
   {va}=eig(inv(W)*B);
    va=rev(sortc(va,1));  @
    va=va[1:minc({p,g-1})]; @
    La=prodc(1/(1+va));           @  La= Wilks Lambda; @
    L=-((n-g)-(p-g+2)/2)*ln(La);  @ L = Likelihood ratio test @
    GH=sumc(va)*v_E/v_H;          @ Generalized Hotelling *v_E/v_H @
    tst=La|L|GH;
  if d==1;

      "---------------------------------------------------------------";
      "    TEST OF EQUALITY AMONG MEANS " ;
      "---------------------------------------------------------------";
      "Wilks Lambda is:";; La;
       "This value must be compared with a Lambda r.v. with";
      "p=";; p;;",";;"v_H=";; v_H;;",";;"v_E=";; v_E;;
       "See Table A.9 of Rencher p. 501-516";
      "---------------------------------------------------------------";
       "Likelihood ratio test is:  ";; L;
       "This value must be compared with a \Chi^2 r.v. with";
        p*(g-1);;  "degrees of freedom";
      "---------------------------------------------------------------";
       "Generalized Hotelling (Lawley Hotelling) *(v_E/v_H) is:";; GH;
       "This value must be compared with a Hotelling T^2 r.v. with";
      "p=";; p;;",";;"v_H=";; v_H;;",";;"v_E=";; v_E;;
       "See Table A.9 of Rencher p. 497-516";
      "---------------------------------------------------------------";

  else;
  endif;
     retp(tst);
endp;

