/*
**   BOXPLOTB.G
**   written by MARCO RIANI & ALDO CORBELLINI
**
**  FORMAT                         PURPOSE                     Line
** =======================================================================
** {out,cent,Hul,Spl}=boxplotb(X,m,d)      To draw a bivariate  66
**                                         boxplot
**
*/

/*
**
** PURPOSE: To superimpose a bivariate boxplot on
**          a scatter diagram
**
** FORMAT:   {out,cent,Hul,Spl}=boxplotb(X,m.d)
**
** INPUT:     X = n x 2 data matrix
**                1st col = x coordinates
**                2nd col = y coordiantes
**            m = scalar which defines the theoretical
**                confidence level of the outer contour
**                For example 1.68 corresponds to a 99%
**                confidence contour if the data are normally
**                distributed
**                REMARK: if m=0 the outer contour coincides with
**                the inner contour
**            d = scalar
**                if d==1 a bivariate boxplot is produced on the screen
**                else no plot is produced
**
** OUTPUT:  out = l x 3 matrix which contains the coordinates
**                of the units which lie outside the outer spline curve
**                1st col = x coordinates
**                2nd col = y coordiantes
**                3rd col = labels of the units
**         cent = 2 x 1 vector containing the coordinates
**                of the robust centroid
**                cent[1] = x coordinate
**                cent[2] = y coordinate
**          Hul = r x k matrix containing the coordinates
**                of the hulls in each step
**                The first two columns contain the coordinates
**                of the outermost hull
**                The last two columns contain the coordinates
**                of the 50% hull
**                REMARK1: The number of hulls is k/2
**                REMARK2: The coordinates of the points of the last row
**                         concide with the coordinates of the point
**                         belonging in the first row
**                REMARK3: (r-1) is the number points forming
**                         the hull with the maximum number of points
**           Spl= 205 x 4 matrix containing the coordinates
**                of the inner and outer spline
**                The first two columns refer to the inner spline
**                The last  two columns refer to the outer spline
**
** REMARK: if no unit lies outside the outer spline out is a
**          scalar equal to 0
**
** See also: HULL,SPLINEM,OUTC
**
*/

proc (4)=boxplotb(X,m,d);
local seq,H50,zleft,Xfin,zleftf,czz,zz,out,i,
      tmp,Hul,Ispl,Ospl,Spl,cent,sel;

     if m < 0;
      errorlog "Coefficient m must not be smaller than 0";
       end;
     endif;

     seq=seqa(1,1,rows(X));
      X=X~seq;
     {Hul,zleft}=hull(X,0.5);

       @  zleft= k x 3 matrix containing
                1st and 2nd cols = x and y coordinates
                     of the points which lie inside the 50% hull
                3rd col = row numbers of these points
         The first two colums refer to the outermost hull
         the last two columns to the 50% hull @
        tmp=packr(Hul[.,cols(Hul)-1:cols(Hul)]);
        tmp=tmp[1:rows(tmp)-1,.];

       {Ispl}=splinem(tmp);
   @ Find robust centroid using the arithmetic means of
     the observations which lie inside the 50 % spline @
     cent=meanc(Ispl);
    {Xfin}=outc(zleft,cent,Ispl);
   @ "Xfin";   Xfin; wait;   @


    @ Xfin contains the units which are inside the 50% hull
      but outside the 50% spline @
if ismiss(zleft)==0;
    if Xfin==0 and cols(Xfin)==1;
       zleftf=zleft;                      @ units inside the 50% spline @
       Czz=X[setdif(seq,zleft[.,3],1),.]; @ units outside the 50% spline @
    else;
       zz=setdif(zleft[.,3],Xfin[.,3],1);
       zleftf=X[zz,.];            @ units inside the 50% spline @
       Czz=X[setdif(seq,zz,1),.]; @ units outside the 50% spline @
    endif;
     @ centroid computed using the arithmetic means of
       the units which lie inside the 50% spline @
     cent=meanc(zleftf[.,1:2]);
else;
    czz=miss(0,0);
    @ "Warning: no unit inside 50% spline"; wait; @
endif;
    
           @   ZLEFTF; WAIT;   @

     @ Ospl = l x 2 matrix containing the coordinates
       of the outer spline @
     Ospl=(Ispl-cent')*(m+1)+cent';

     @ out contains the coordinates of the units which
       lie outside the outer spline @
     {out}=outc(Czz,cent,Ospl);

      spl=Ispl~Ospl;
      if d==1;
                 graphset;
                  begwind;
                  window(1,1,1);
                  makewind(80,6.855,0,0,1);
                  scale(X[.,1]|spl[.,3],X[.,2]|spl[.,4]);
                         _psym=cent[1]~cent[2]~
                          4~7~1.5~1;
                          _plctrl=0;
                         @ Plot 50% spline @
                        xy(spl[.,1],spl[.,2]);
                         @ Plot outer spline @
                       xy(Spl[.,3],Spl[.,4]);
                        _plctrl=-1;
                         @ plot all points @
                          xy(X[.,1],X[.,2]);
                  endwind;
        else;
        endif;
 retp(out,cent,hul,spl);
endp;

