/*
**  CONFLR (Confirmatory analysis using likelihood ratio)
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {signLr,Un}=conflr(X,lar,sel,la,th,d);
**
**                                To produce sign sqrt of lik ratio
**                                in each step of the fwd search
**                                for each element of vector la
**                                and each column of matrix X
**                                specified in vector sel
**
*/

/*
** CONFFLR
**
** PURPOSE:  To produce sign sqrt of lik ratio
**           in each step of the fwd search
**           for each element of vector la
**           and each column of matrix X
**           specified in vector sel
**
**
**
** FORMAT:   {signLr,Un}=ConfLr(X,lar,sel,la,th,d);
**
** INPUT:     X = n x p data matrix; n observations
**                and p variables
**          lar = p x 1 vector which specifies the reasonable
**                value of the transformation parameter
**                for each variable
**          sel = c x 1 (2) vector or scalar 0
**                if sel=vector indexes of the variables
**                             (VAR) for which
**                             you want to have the sign sqrt lik ratio
**                             test
**                if sel=0 (scalar zero) all variables
**                         are considered
**                If sel has two columns, the second column contains
**                the true numbers of the columns whose value
**                             of lambda has to be tested
**          la  = l x 1 vector or scalar missing value
**                if la=vector specifies for which values
**                      of the transformation parameters
**                      you want to monitor the sign sqrt lik ratio test
**                if la=scalar missing value then
**                        la=(-1,-0.5,0,0.5,1)'
**          th  = constant (>0) which specifies the confidence
**                level of the bivariate ellipses
**                (it must be chosen according to the F
**                distribution with 2 and N-2 d.f.)
**          d   = scalar
**                if d=1 a plot of lik ratio is created
**                elseif d==2 a plot of sign sqrt lik. ratio is created
**                else no plot is created on the screen
**                Every panel of the plot refers to a VAR
**                (element of vector sel)
**                Each curve in each panel refers to an
**                element of vector la
**
** OUTPUT: sqrtLr = (n-m+1) x (l*c+1) matrix.
**                  1st col  =  fwd search index
**                  2nd col  =  Sign sqrt Lik. ratio
**                              referred to sel[1] and la[1]
**                  3rd col  =  Sign sqrt Lik. ratio
**                              referred to sel[1] and la[2]
**                    ....
**                  l*c+1 col = Sign sqrt Lik. ratio
**                              referred to sel[c] and la[l]
**             Un = (n-m) x (l*c+1)
**                  1st col  =  fwd search index
**                  2nd col  =  Units included in bsb for the fwd search
**                              referred to sel[1] and la[1]
**                  3rd col  =  Units included in bsb for the fwd search
**                              referred to sel[1] and la[2]
**                    ....
**                  l*c+1 col = Units included in bsb for the fwd search
**                              referred to sel[c] and la[l]
**                  REMARK:
**                  If more than one unit joins the subset than
**                  than the number of these units is collapsed into
**                  one big number
**                  For example if n<100 and Un[34,2]=7623 means that
**                  when m (subset size) = Un[34,1] units 76 and 23
**                  have joined the subset
**
**
*/


 proc (2)=conflr(X,lar,sel,la,th,d);

   local seq,n,p,bsb,Xt,nla,nsel,sqrtLr,U,Un,jj,ii,G,Mp,out,f,Xr,X1,
         Lmax,nlr,a,b,lr,sign,i,Lx,In,ast,l,on,j,nr,truelab,pmslr,lari;

 n=rows(X);
 p=cols(X);
 seq=SEQA(1,1,n);


 if la==miss(0,0);
   la=-1|-0.5|0|0.5|1;  @ vector la specifies for which values of lambda
                          you want to compute the lik. ratio @
 else;
 endif;

 if sel==0 and rows(sel)==1;
  sel=seqa(1,1,p);
 else;
  if cols(sel)==2;
     truelab=sel[.,2];
     sel=sel[.,1];
   else;
     truelab=sel[.,1];
   endif;     
 endif;

  nla=rows(la);
  nsel=rows(sel);


 @ Initialize scalar nr which might bu used for matrix Un @
    nr=100;
  if n<100;
  elseif nr<1000;
    nr=1000;
  else;
    nr=10000;
  endif;

@ Matrix msign will contain the fwd searches for sign sqrt of lik. ratio. @
 SqrtLr=seq~miss(zeros(n,nla*nsel),0);
@ Un= n x (nla*nsel+1) matrix
  units enterd the subset for each value of vector la and each
  value of vector sel @
  Un=SqrtLr;


 @ Step 1: transform the variables with the reasonable values you have found
           and, if necessary, rescale them @
  Xr=X; @ Initialize Xr @
    j=1; do until j==p+1;
                G=exp(meanc(ln(X[.,j])));
                @ G=meanc(X[.,j]); @
         if lar[j]==0;
                Xr[.,j]=G*ln(X[.,j]);
        else;
                Xr[.,j]=X[.,j]^(lar[j])/(G^(lar[j]-1));
        endif;
     /**              rescale:
             if maxc(Xr[.,j])< 5;
         maxc(Xr[.,j]);
  "jkljkljl"; wait;
                Xr[.,j]=Xr[.,j]*10;
                     goto rescale;
                 else;
             endif;    ***/
        j=j+1;
    endo;


jj=1; do until jj==nsel+1;
  Xt=Xr;  @ transform for procedures medb and unibiv @
  X1=Xr;
  ii=1; do until ii==nla+1;
        @ Step 1: Transform the appropriate column of matrix X
            for procedure medb and unibiv @
           @ Xt = Matrix which contains all variables transformed
             as specified in vector lar and variable
             sel[jj] trasformed as specified by la[ii] @

             G=exp(meanc(ln(X[.,sel[jj]])));
             G=meanc(X[.,sel[jj]]);
           if la[ii] NE 0;
               Xt[.,sel[jj]]= X[.,sel[jj]]^la[ii]/(G^(la[ii]-1));
              else;
               Xt[.,sel[jj]]= G*ln(X[.,sel[jj]]);
             endif;

              @ X1= Matrix which contains all variables transformed
                as specified in vector lar and variable
                sel[jj] untrasformed @
               X1[.,sel[jj]]= X[.,sel[jj]];
    @  maxc(x1)~minc(x1); wait;   @
      @  X1=X1*100;   @

         @ Step 2: Calculate matrix of medians @
              {mP}=medb(Xt,0);

         @ Step 3: Calculate initial subset @
            {out}=unibiv(Xt,th,mP,0);
            f=out[.,1];
            bsb=setdif(seq,f,1);


         @ Step 4: Compute MLE of transformation parameters for
                   the variables specified in vector sel[jj] @
            {lmax}=fwdmle(X1,bsb,sel[jj]~la[ii]);
                lmax; wait;  
         @  lmax=real(lmax);  @

         @ Step 5: Compute Likelihood ratio test statistics @
         {a,b,lr,Lx,In,U}=flr(X1,bsb,sel[jj]~la[ii],lmax[.,2:p+1],n+1);
          nlr=rows(lr);
          sign=zeros(nlr,1);

               i=1; do until i==nlr+1;
                       if  lmax[i,sel[jj]+1]-la[ii] >0;
                          sign[i]=sqrt(lr[i,2]);
                       else;
                          sign[i]=-sqrt(lr[i,2]);
                       endif;
                  i=i+1;
               endo;

               SqrtLr[lr[.,1],(ii+1)+(jj-1)*nla]=real(sign);

                 if (U[.,3]$==miss(0,0))==1;
                 else;
                    U=missrv(U,zeros(rows(U),cols(U)));
                    U=(delif(U',sumc(U).==0)');
                       j=3; do until j==cols(U)+1;
                         U[.,2]=U[.,2]+nr^(j-2)*U[.,j];
                     j=j+1;
                    endo;
                  endif;
               Un[U[.,1],(ii+1)+(jj-1)*nla]=U[.,2];
      ii=ii+1;
    endo;
  jj=jj+1;
 endo;
 
@ sqrtlr; wait; @
 SqrtLr=packr(SqrtLr);
 Un=packr(Un);

if d==1 or d==2;
              if d==1;
                  ast=6.63|0;
              else;
                 ast=2.58|0|-2.58;
                 ytics(-4,4,1,1);
                 ytics(-6,6,2,1);
              endif;
                  l=rows(ast);
                  on=ones(l,1);
           _pline=on~6*on~zeros(l,1)~ast~(n+1)*on~ast~2*on~ 2*on~0.4*on;
           _pnumht=0.3;
           _plegctl =3|4|250|1800;
           
             pmslr="";
            i=1; do until i==rows(la)+1;
               pmslr= pmslr $+ "la" $+ ftocv(la[i],3,2) $+ "\000";
               i=i+1;
            endo;
         
          _plegstr=pmslr;
          _plwidth=8;
          _pltype=6|3|2|5|1|4;
      
            begwind;
            if nsel<=3;
              window(nsel,1,0);
            elseif nsel<=6;
              window(ceil(nsel/2),2,0);
            elseif nsel<=12;
              window(ceil(nsel/3),3,0);
            else;
              window(6,4,0);
            endif;
            makewind(80,6.8555,0,0,1);
               i=1; do until i==nsel+1;
                  lari=lar[sel[i]];
                  @ lari;  wait; @
                   _PMSGSTR ="y" $+ FTOCV(truelab[i],2,0) $+ "H0:la=" $+ ftocv(lari,3,1);
                       _PMSGCTL=0.7~6~0.3~0~2~2;
                 if d==1;
                       xy(sqrtLr[.,1],sqrtLr[.,nla*(i-1)+2:nla*i+1]^2);
                 else;
                  xy(sqrtLr[.,1],sqrtLr[.,nla*(i-1)+2:nla*i+1]);
                 endif;
            nextwind;
               i=i+1;
        endo;
        endwind;
  else;
  endif;
/*****
_plegctl =3|4|250|1800;
_plegstr = " la=-1 \0"\
           " la=-0.5 \0"\
           " la=0 \0"\
           " la=0.25 \0"\
           " la=0.5 \0"\
           " la=1 \0"\
           " la6 \0"\
           " la7 \0"\
           " la8 \0";
                  ast=6.63;
                  l=rows(ast);
                  on=ones(l,1);
           _pline=on~6*on~zeros(l,1)~ast~(n+1)*on~ast~2*on~2*on~0.4*on;
           _pnumht=0.3;
       ytics(0,10,2,1);

      xy(mlr[.,1],mlr[.,2:nla+1]);

_plegctl =3|2|300|2500;
_plegstr = " la=-1 \0"\
           " la=-0.5 \0"\
           " la=0 \0"\
           " la= 0.5 \0"\
           " la=1 \0";

 ytics(-7,7,1,1);
 ast=2.58;
 ast=ast|-ast;
 l=rows(ast);
 on=ones(2,1);
 _pline=on~6*on~zeros(2,1)~ast~(n+1)*on~ast~2*on~2*on~0.4*on;

    xy(msign[.,1],real(msign[.,2:nla+1]));
 ******/
   retp(sqrtLr,Un);
endp;

