/*
**  CONFLRAD (Confirmatory analysis using likelihood ratio for
**  Discriminant Analysis)
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {signLr,Un}=conflrAD(X,lar,sel,la,th,vsz,dd,quadr,d)
**
**                                To produce sign sqrt of lik ratio
**                                in each step of the fwd search
**                                for each element of vector la
**                                and each column of matrix X
**                                specified in vector sel
**
*/

/*
** CONFLR
**
** PURPOSE:  To produce sign sqrt of lik ratio
**           in each step of the fwd search
**           for each element of vector la
**           and each column of matrix X
**           specified in vector sel
**
**
**
** FORMAT:   {signLr,Un}=conflrAD(X,lar,sel,la,th,vsz,dd,quadr,d)
**
** INPUT:     X = n x p data matrix; n observations
**                and p variables
**          lar = p x 1 vector which specifies the reasonable
**                value of the transformation parameter
**                for each variable
**          sel = c x 1 vector or scalar 0
**                if sel=vector indexes of the variables
**                             (VAR) for which
**                             you want to have the sign sqrt lik ratio
**                             test
**                if sel=0 (scalar zero) all variables
**                         are considered
**          la  = l x 1 vector or scalar missing value
**                if la=vector specifies for which values
**                      of the transformation parameters
**                      you want to monitor the sign sqrt lik ratio test
**                if la=scalar missing value then
**                        la=(-1,-0.5,0,0.5,1)'
**          th  = constant (>0) which specifies the confidence
**                level of the bivariate ellipses
**                (it must be chosen according to the F
**                distribution with 2 and N-2 d.f.)
**          vsz = g x 1 vector which contains sizes of the groups
**          dd  = 2 x 1 vector specifying which kind of forward search
**                you want
**                if dd[1]=0 inclusion is based on MD without constraints
**                if dd[1]=1 inclusion is based on the difference
**                           among the sizes of the groups
**                  else     inclusion is based on the ratio
**                           among the sizes of the groups
**                element dd[2] is a suboption when dd[1] /= 0
**                if dd[2]=0  when you include a unit from
**                            universe j basic subset for
**                            universe k (k ne q) can change
**                 else       when you include a unit from
**                            universe j basic subset for
**                            universe k (k ne q) cannot change
**        quadr = scalar which specifies whether you want to use
**                linear or quadratic discriminant analysis
**                quadr=1 ==> LDA
**                else ===> QDA
**          d   = scalar
**                if d=1 a plot of lik ratio is created
**                elseif d==2 a plot of sign sqrt lik. ratio is created
**                else no plot is created on the screen
**                Every panel of the plot refers to a VAR
**                (element of vector sel)
**                Each curve in each panel refers to an
**                element of vector la
**
** OUTPUT: sqrtLr = (n-m+1) x (l*c+1) matrix.
**                  1st col  =  fwd search index
**                  2nd col  =  Sign sqrt Lik. ratio
**                              referred to sel[1] and la[1]
**                  3rd col  =  Sign sqrt Lik. ratio
**                              referred to sel[1] and la[2]
**                    ....
**                  l*c+1 col = Sign sqrt Lik. ratio
**                              referred to sel[c] and la[l]
**             Un = (n-m) x (l*c+1)
**                  1st col  =  fwd search index
**                  2nd col  =  Units included in bsb for the fwd search
**                              referred to sel[1] and la[1]
**                  3rd col  =  Units included in bsb for the fwd search
**                              referred to sel[1] and la[2]
**                    ....
**                  l*c+1 col = Units included in bsb for the fwd search
**                              referred to sel[c] and la[l]
**                  REMARK:
**                  If more than one unit joins the subset than
**                  than the number of these units is collapsed into
**                  one big number
**                  For example if n<100 and Un[34,2]=7623 means that
**                  when m (subset size) = Un[34,1] units 76 and 23
**                  have joined the subset
**
**
*/


 proc (2)=conflrAD(X,lar,sel,la,th,vsz,dd,quadr,d);

   local seq,n,p,bsb,Xt,nla,nsel,sqrtLr,U,Un,jj,ii,G,Mp,out,f,Xr,X1,
         Lmax,nlr,a,b,lr,sign,i,ast,l,on,j,nr,like,ch,mis,
         vv,X2,sz,sz1,Ibsb;

 n=rows(X);
 p=cols(X);
 g=rows(vsz);
 seq=seqa(1,1,n);

 if la==miss(0,0);
   la=-1|-0.5|0|0.5|1;  @ vector la specifies for which values of lambda
                          you want to compute the lik. ratio @
 else;
 endif;

 if sel==0 and rows(sel)==1;
  sel=seqa(1,1,p);
 else;
 endif;

  nla=rows(la);
  nsel=rows(sel);

 @ Initialize scalar nr which might bu used for matrix Un @
  if n<100;
    nr=100;
  elseif n<1000;
    nr=1000;
  else;
    nr=10000;
  endif;

@ Matrix msign will contain the fwd searches for sign sqrt of lik. ratio. @
 SqrtLr=seq~miss(zeros(n,nla*nsel),0);
@ Un= n x (nla*nsel+1) matrix
  units enterd the subset for each value of vector la and each
  value of vector sel @
  Un=SqrtLr;

 @ Step 1: transform the variables with the reasonable values you have found
           and, if necessary, rescale them @
  Xr=X; @ Initialize Xr @
    j=1; do until j==p+1;
         if lar[j]==0;
                Xr[.,j]=ln(X[.,j]);
        else;
                Xr[.,j]=X[.,j]^(lar[j]);
        endif;
                   rescale:
             if maxc(Xr[.,j])< 5;
                Xr[.,j]=Xr[.,j]*10;
                     goto rescale;
                 else;
             endif;
        j=j+1;
    endo;

jj=1; do until jj==nsel+1;
  Xt=Xr;  @ transform for procedures medb and unibiv @
  X1=Xr;
        @ Step 1: Transform the appropriate column of matrix X
            for procedure medb and unibiv @
         @ Step 2: Calculate matrix of medians @
         @ Step 3: Calculate initial subset @
         @ Step 4: Compute MLE of transformation parameters for
                   the variables specified in vector sel[jj] @
         @ Step 5: Compute Likelihood ratio test statistics @
  ii=1; do until ii==nla+1;
           @ Xt = Matrix which contains all variables transformed
             as specified in vector lar and variable
             sel[jj] trasformed as specified by la[ii] @
           if la[ii] NE 0;
               Xt[.,sel[jj]]= X[.,sel[jj]]^la[ii];
              else;
               Xt[.,sel[jj]]= ln(X[.,sel[jj]]);
             endif;

              @ X1= Matrix which contains all variables transformed
                as specified in vector lar and variable
                sel[jj] untrasformed @
               X1[.,sel[jj]]= X[.,sel[jj]];

  sz=0; @ this vector will contain the sizes of the basic subsets @
  vv=0|vsz;
  f=0; @ This vector will contain potential outliers @
  j=1; do until j==g+1;
     X2=Xt[sumc(vv[1:j])+1:sumc(vv[1:j+1]),.];
      {mP}=medb(X2,0);
      {out}=unibiv(X2,th,mP,0);
      if ismiss(out)==1;
        sz=sz|0;
      else;
        f=f|(out[.,1]+sumc(vv[1:j]));
        sz=sz|rows(out[.,1]);
      endif;
    j=j+1;
  endo;
f=f[2:rows(f)];
sz=sz[2:rows(sz)];
sz1=vsz-sz;    @ sizes of the groups forming the basic subset @
Ibsb=setdif(seq,f,1);  @ units forming initial basic subset @
           ch=sel[jj]~la[ii];
           mis=0;
 @  VCX(iBSB); WAIT;  @
 if quadr==1;
    {lmax}=fwdmleQD(Ibsb,sz1,X1,vsz,dd,ch,mis);
     {lr,like,U}=flrQD(Ibsb,sz1,X1,vsz,dd,ch,lmax[.,2:p+1],mis,n+1);
  else;
    {lmax}=fwdmleLD(Ibsb,sz1,X1,vsz,dd,ch,mis);
     @ lmax[.,4]'; wait; @
    {lr,like,U}=flrLD(Ibsb,sz1,X1,vsz,dd,ch,lmax[.,2:p+1],mis,n+1);
@   LMAX~lR; wait; @
  endif;

          nlr=rows(lr);
          sign=zeros(nlr,1);
               i=1; do until i==nlr+1;
                       if  lmax[i,sel[jj]+1]-la[ii] >0;
                          sign[i]=sqrt(lr[i,2]);
                       else;
                          sign[i]=-sqrt(lr[i,2]);
                       endif;
                  i=i+1;
               endo;

               SqrtLr[lr[.,1],(ii+1)+(jj-1)*nla]=sign;

                 if (U[.,3]$==miss(0,0))==1;
                 else;
                    U=missrv(U,zeros(rows(U),cols(U)));
                    U=(delif(U',sumc(U).==0)');
                       j=3; do until j==cols(U)+1;
                         U[.,2]=U[.,2]+nr^(j-2)*U[.,j];
                     j=j+1;
                    endo;
                  endif;
               Un[U[.,1],(ii+1)+(jj-1)*nla]=U[.,2];
      ii=ii+1;
    endo;
  jj=jj+1;
 endo;
 SqrtLr=packr(SqrtLr);
 Un=packr(Un);

if d==1 or d==2;
              if d==1;
                  ast=6.63|0;
              else;
                 ast=2.58|0|-2.58;
                 ast=1.96|0|-1.96;
              endif;
                  l=rows(ast);
                  on=ones(l,1);
           _pline=on~6*on~zeros(l,1)~ast~(n+1)*on~ast~2*on~ 2*on~0.4*on;
           _pnumht=0.3;
           _plegctl =3|4|250|1800;
    if rows(la)==5;
           _plegstr = " la=-1 \0"\
                      " la=-0.5 \0"\
                      " la=0 \0"\
                      " la=0.5 \0"\
                      " la=1 \0";
    else;
           _plegstr = " la1 \0"\
                      " la2 \0"\
                      " la3 \0"\
                      " la4 \0"\
                      " la5 \0"\
                      " la6 \0"\
                      " la7 \0"\
                      " la8 \0";
    endif;

            begwind;
            if nsel<=3;
              window(nsel,1,0);
            elseif nsel<=6;
              window(ceil(nsel/2),2,0);
            elseif nsel<=12;
              window(ceil(nsel/3),3,0);
            else;
              window(6,4,0);
            endif;
            makewind(80,6.8555,0,0,1);
               i=1; do until i==nsel+1;
                            _PMSGSTR =0 $+ "X" $+ FTOCV(sel[i],1,0);
                       _PMSGCTL=0.7~6~0.3~0~2~2;
                 if d==1;
                       xy(sqrtLr[.,1],sqrtLr[.,nla*(i-1)+2:nla*i+1]^2);
                 else;
                  xy(sqrtLr[.,1],sqrtLr[.,nla*(i-1)+2:nla*i+1]);
                 endif;
            nextwind;
               i=i+1;
        endo;
        endwind;
  else;
  endif;
   retp(sqrtLr,Un);
endp;

