/*
**   EQVAR.G
**   written by MARCO RIANI
**
**  FORMAT                      PURPOSE                     Line
** =======================================================================
** {LRc}=eqvar(X,sz,d)     Test of homogeneity of
**                         covariances
**                         See Mardia et al. 1979: p. 140
**                         or Rencher (1995) p.280-284
*/

/*
**
** EQVAR
**
** PURPOSE:  to test whether var cov matrices of different groups
**           can be considered equal
**
** FORMAT:   {LRc}=eqvar(X,sz,d)
**
** INPUT:     X = n x p matrix of data
**           sz = g x 1 vector which spcifies the sizes of the groups
**            d = scalar.
**                if d=1  program displays on the screen the value of
**                the test using and not using Box correction and
**                the degres of freedom
**                else program computes test using Box correction
**                and nothing is printed on the screen
**
** OUTPUT:  LRc = value of likelihood ratio test statistic
**            Remark: Box correction is used
**
**
**  WARNING: if the absolute value of the determinant of the
**           var cov matrix of any group is less than 1e-20
**           program reports a missing value for LR test
**
*/


proc (1)=eqvar(X,sz,d);
 local n,p,g,i,VZ,VZi,dVZi,sVZi,LR,LRc,gam;
    n=rows(X);         @ n=rows of data set @
      if sumc(sz) ne n;
       errorlog "The sum of the elements of vector sz is not equal to
                the number of rows of matrix X.";
        end;
      endif;
    p=cols(X);
    g=rows(sz);
    sz=0|sz;
    @ Initialize VZ (this matrix will contain the
      pooled var cov matrix @
    VZ=0;
    @ Initialize sVZi (this matrix which will contain the
      quantity \sum_{i=1}^g (n_i-1) \ln | \S_i| @
    sVZi=0;

  i=1; do until i==g+1;
    @ VZi unbiased variance covariance matrix for group i @
    VZi=vcx(X[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]);
    VZ=VZ+VZi*(sz[i+1]-1);
    dVZi= abs(det(VZi));
     if dVZi < 1e-40;
       Lrc=miss(0,0);
       goto end;
     else;
       sVZi= sVZi+(sz[i+1]-1)*ln(dVZi);
     endif;
        i=i+1;
   endo;
      VZ=VZ/(n-g);
  @ LR= Likelihood ratio test statistic @
  LR= (n-g)*ln(det(VZ))-sVZi;
  @ gam = correction factor due to Box (1949)
   see Mardia et al. p. 140
   or Rencher (1995) p. 282 eq. 7.20
  Remark:  -2 ln M for Rencher corresponds to eq. 5.3.27 for Mardia @
  gam=1-(2*p^2+3*p-1)/(6*(p+1)*(g-1))*( sumc(1/(sz[2:g+1]-1)) -1/(n-g));
  LRc=LR*gam;
  if d==1;
  "/*****************************************************************/";
  "LR test of homogeneity of covariances (without Box corr.)";;  LR;
  "LR test of homogeneity of covariances using Box corr.";; LRc;
  " This value must be compared with a X^2 r.v. with";
  /*  p(p+1)(g-1)/2   degrees of freedom  */
  0.5*p*(p+1)*(g-1);; "degrees of freedom";
  "/*****************************************************************/";
  else;
  endif;
  end:
 retp(LRc);
endp;



