/*
**  FLR (Forward likelihood ratio)
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {pos,pot,lr,likmax,In,Un}=flr(X,bsb,ch,lamax,d);
**
**                                To monitor the likelihood ratio
**                                in every step of the forward procedure
**                                The purpose is to test the null
**                                hypothesis that the transformation
**                                parameters are as those specified in
**                                vector la.
**
**
*/

/*
** FLR
**
** PURPOSE:  The principal purpose is to monitor the likelihood ratio in
**           every step of the forward procedure. The purpose is to test
**           the null
**           hypothesis that the transformation parameters are as those
**           specified in vector la.
**           It also produces average, maximum of Mahalanobis distances
**           of clean (and min of noclean) in  each step of
**           the forward procedure
**
** FORMAT:   {pos,pot,lr,likmax,In,Un}=flr(X,bsb,ch,lamax,d);
**
** INPUT:     X = n x p data matrix; n observations
**                and p variables
**          bsb = m x 1 vector. It contains
**                the indexes of the m units which
**                form the initial subset
**                Initial centroid is determined using only
**                these units.  (1 \leq m \leq n)
**       ch     = c x 2 (3) matrix or scalar 0
**                1st column = indexes of the variables (VAR)
**                             to be tested for transformation
**                2nd column = values of transformation
**                             parameters
**                3rd column = true numbers of the columns whose value
**                             of lambda has to be tested
**                REMARK: if ch=0 (scalar zero) you can test
**                         the null hypothesis that no variable
**                         must be transformed
**                EXAMPLE: if ch=(1~0)|
**                               (4~0.5)
**                         LR concerns VAR 1 and 4
**                         The null hypothesis is:
**                         log transformation for VAR 1
**                         and sqrt transformation for VAR 4
**                EXAMPLE: if ch=(1~0~3)|
**                               (4~0.5~7)
**                         LR concerns VAR 1 and 4
**                         The null hypothesis is:
**                         log transformation for VAR 1
**                         and sqrt transformation for VAR 4
**                         The true labels for the variables are 3 and 7
**        lamax = (n-m+1) x p matrix which contains (in the rows) the
**                values of MLE transformation parameters in the last
**                (n-m+1) steps of the forward search
**                REMARK1: the last row of matrix lamax refers to
**                MLE when m=n (last step of the fwd search)
**           d  = l x 1 vector which enables the user to monitor the
**                likelihood grids of transformation parameters
**                for the variables specified in the first column
**                of matrix ch
**                For examples if d={24,33} you can monitor the
**                likelihood grid when there are respectively 24 and 33
**                observations in bsb
**
** OUTPUT:  pos =  (n-m+1) x 3 matrix.  The first column of pos
**                 contains the
**                 dimension  of the sample used to compute Mahalanobis
**                 distances Columns 2,3 respectively contain
**                 Max and Average of Mahalanobis distances of the units
**                 forming the basic subset
**          pot =  (n-m) x 2 matrix. The first column contains the
**                 dimension of the sample used in order to compute
**                 Mahalanobis distances, the second
**                 reports the minimum of Mahalanobis distances among
**                 the units which form the group of potential
**                 outliers
**                 REMARK: if m=n pot is a missing value
**            lr = (n-m+1 x 2) matrix. The first column contains the
**                 dimension of
**                 the basic subset, the second the values of the
**                 likelihood ratio in each
**                 step of the forward search
**                 The reference distribution is a \Chi^2 r.v.
**                 with c degress of freedom (c=rows(ch))
**        likmax = (n-m+1 x 2) matrix. the first column contains the
**                 size of the basic subset, the second the value of the
**                 maximazied likelihood
**           In  = (n-m+1) x 11 matrix
**                 1st col = fwd search index
**                 2nd col = unit with m-th MD (max MD in bsb)
**                 3rd col = unit with (m+1)-th MD (min MD among
**                           the units not in bsb)
**                  ......
**                 11th col = unit with (mm+10)th MD
**           Un = (n-m) x 11 Matrix which contains the unit(s)
**                included in bsb at each step of the fwd search
**                REMARK: in every step the new subset is compared
**                with the old subset. Un contains the unit(s)
**                which are in the new subset but not in the old one
**
**
**        WARNING: the number of windows for the likelihood grids
**                 cannot exceed 24
**
**    See Also: FWDMLE, RFLR, FLRLD, FLRQD
**
**
**
*/


 proc (6)=flr(X,bsb,ch,lamax,d);

   local seq,ym,vc,nbsb,la,Xt,sel,
       nn,p,mm,i,gg,pos,pot,zs,min,mala,ii,lr,mess,In,Un,Unit,
       j,G,z,Xc,dZ,dZmax,Zmax,la1,gri,lat,k,zt,Likmax,
       lim,confint,on,nla1,ol,trlab;

nn=rows(X);
p=cols(X);
seq=SEQA(1,1,nn);

mm= rows(bsb);
min=mm;
@   The first column of pos contains the dimension
    of the sample used to compute Mahalanobis distances
    Columns 2,3 respectively contain
    Max and Average of Mahalanobis distances of the units
    forming the basic subset
    Column 4 contains Max of Mahalanobis distances of all
    the units  @
  pos=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,2);

  @  pot has two columns the first contains the dimension of the
     sample used in order to compute Mahalanobis distances, the second
     reports the minimum of Mahalanobis distances among the units
     which form the group of potential outliers @
   if mm<nn;
    pot=seqa(mm,1,nn-mm)~zeros(nn-mm,1);
   else;
    pot=miss(0,0);
   endif;
  @ Matrix lr which will contain the LR test in each step of the
    forward search @
  lr=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,1);

  @ matrix In will contain in the 2nd col. the unit with greatest MD
    in basic subset, in third col the one with the smallest MD among
    Noclean, in 4th col the unit with second smallest MD among nocl. @
   In=seqa(mm,1,nn-mm+1)~miss(zeros(nn-mm+1,10),0);

   if nn>mm;
      @  matrix Un will contain from 2nd col. to 10th the unit(s) which
      entered the basic subset @
      Un=seqa(mm+1,1,nn-mm)~miss(zeros(nn-mm,10),0);
   else;
      Un=miss(0,0);
  endif;




  @ Matrix likmax which will contain the max of the profile
  likelihood in each step of the
  forward search @
       Likmax=lr;


/*** mala in the column will contain MD  ***/
mala=seq~zeros(nn,1);


   @ ch[.,1] specifies the variables to analyze in the LR test
     ch[.,2] controls how the variable nust be transformed in the
     fwd search @
       if rows(ch)==1 and ch==0;
           sel=seqa(1,1,p);
           la=ones(p,1);
       else;
           sel=ch[.,1];
           la=ch[.,2];
           if cols(ch)==3;
              trlab=ch[.,3];
           else;
              trlab=ch[.,1];
           endif;
       endif;


          @ _Create Xt=matrix of transformed values to be used
             in the forward search (_for ordering MD) @
          Xt=X;


        j=1; do until j==rows(sel)+1;
                          @ la; wait;
                              j;
                              sel[j];  wait;
                              la[sel[j]]; wait;   @
                G=exp(meanc(ln(X[.,j])));
                              @  "GG";   G; wait; @

           if la[j] NE 0;
                  Xt[.,sel[j]]= X[.,sel[j]]^la[j] @  /
                          G^(la[j]-1) @ ;
              else;
                  Xt[.,sel[j]]= @ G* @ ln(X[.,sel[j]]);
             endif;

             @ if maxc(Xt[.,sel[j]])-minc(Xt[.,sel[j]])<0.1;
                  Xt[.,sel[j]]= 10*Xt[.,sel[j]]/maxc(Xt[.,sel[j]]);
              else;
             endif;  @

           j=j+1;
        endo;
       @    "mina nd max";
      minc(Xt);
      maxc(Xt);  wait;   @


mm=mm; do until mm==nn+1;

  Z=X[bsb,.];
  Zmax=Z;
 j=1;  do until j==rows(la)+1;
    G=exp(meanc(ln(Z[.,sel[j]])));
     if la[j] NE 0;
             Z[.,sel[j]]=(Z[.,sel[j]]^la[j]-1)/(la[j]*G^(la[j]-1));
       else;
             Z[.,sel[j]]= G*ln(Z[.,sel[j]]);
      endif;

    if lamax[mm-min+1,sel[j]] NE 0;
             Zmax[.,sel[j]]=(Zmax[.,sel[j]]^lamax[mm-min+1,sel[j]]-1)/(
                    lamax[mm-min+1,sel[j]]*G^(lamax[mm-min+1,sel[j]]-1));
       else;
             Zmax[.,sel[j]]=G*ln(Zmax[.,sel[j]]);
      endif;

    j=j+1;
  endo;



     if maxc(lamax[rows(lamax),.]')-minc(lamax[rows(lamax),.]')<1E-12;
       ol=1;
     else;
       ol=0;
     endif;



       dZ=det(vcx(Z));
       dZmax=det(vcx(Zmax));

if mm==57;
  "Z and Zmax";
   dZ;
   dZmax;
wait;
else;
endif;

@ "mmmm zmax ";dzmax; wait;   @
            
          if abs(dZmax)<1e-12;
                 lr[mm-min+1,2]=miss(0,0);
                 Likmax[mm-min+1,2]=miss(0,0);
          else;
                lr[mm-min+1,2]=mm*ln(dZ/dZmax);
                Likmax[mm-min+1,2]=-mm*ln(det(vcx(Zmax)));

          endif;

       @ Likelihood grids for variables speeified in the
         first column of matrix ch @
       if maxc(mm.==d)==1;
       
         @ vector on will be used in matrix _pline to draw
          99% confidence bands of transformation parameter @
         on=ones(3,1);
         
         if ol==0;
            begwind;
            if rows(la)<=3;
              window(1,rows(la),0);
            elseif rows(la)<=6;
              window(2,3,0);
            elseif rows(la)<=9;
              window(3,3,0);
            elseif rows(la)<=12;
              window(3,4,0);
            elseif rows(la)<=15;
              window(3,5,0);
            elseif rows(la)<=24;
              window(4,6,0);
            else;
              window(5,6,0);
            endif;
            makewind(80,6.8555,0,0,1);

             la1=seqa(-2,0.01,401);
              la1=seqa(-1,0.01,260); 
             nla1=rows(la1);
             xtics(la1[1],la1[nla1],0.1,1);
             Gri=zeros(rows(la1),1);

               i=1; do until i==rows(la)+1;
               
                   @ vector lim will contain the x coordinates of the
                     confidence interval for lambda
                     lim[1,1] = lower limit
                     lim[2,1] = lower limit
                     lim[3,1] = mle   @
                   lim=miss(zeros(3,1),0);
               
                     G=exp(meanc(ln(X[bsb,sel[i]])));
                     Zt=Zmax;   k=1;   confint=0;
                        j=1; do until j==nla1+1;
                         if abs(la1[j]) >1e-5;
                           Zt[.,sel[i]]=(X[bsb,sel[i]]^la1[j]-1
                                  )/(la1[j]*G^(la1[j]-1));
                         else;
                           Zt[.,sel[i]]=G *ln(X[bsb,sel[i]]);
                         endif;

                             Gri[k]=-mm*ln(det(vcx(Zt)));
                         @ Likmax[mm-min+1,.]; wait; @  
                               @ The following _if finds the conf. interval
                              _for each transformation parameter @
                            if Likmax[mm-min+1,2]-Gri[k]<=0.5*3.84
                                   and confint==0;
                                    lim[1]=la1[j];
                                    confint=1;
                            elseif Likmax[mm-min+1,2]-Gri[k]>=0.5*3.84
                                   and confint==1;
                                    lim[2]=la1[j];
                                  confint=2;
                                         @  "DIFFF";
                                     lim[1]-lamax[mm-min+1,i];
                                     lamax[mm-min+1,i]-lim[2]; wait; @
                            else;
                            endif;
                             
                           j=j+1;   k=k+1;
                        endo;
                      _pnumht=0.3;
                      if i==1;
                      @ minc(gri); maxc(gri); "jkl"; wait;  @ 
                       ytics(round(minc(Gri))-30,round(maxc(gri))+5,50,1);
                       _plwidth=10;
                      else;
                       endif;
                      
                          mess= "m=" $+ ftocv(mm,3,0) $+
                          "y" $+ FTOCV(trlab[i],2,0);
                            _PMSGSTR=mess;
                       _PMSGCTL=0.7~1~0.5~0~2~2;
                       
                        lim[3]=lamax[mm-min+1,i];
                _pline=on~6*on~lim~-1e+10*on~lim~1e+10*on~2*on~
                        2*on~0.4*on;
                  @ " variable";; i; "Gri"; la1~Gri; wait;    @
                  xy(la1,Gri);
                  
                  "lim for variable";; i;
                  lim;
            nextwind;
               i=i+1;
               endo;
            endwind;
            graphset;
          else; @ else given that lambda is the same for all the
                  variables just produce one plot @

            begwind;
              window(1,1,0);
            makewind(80,6.8555,0,0,1);

             la1=seqa(-5,0.1,122);
             nla1=rows(la1);
             xtics(la1[1],la1[nla1],0.2,1);
             Gri=zeros(rows(la1),1);

                     G=((exp(meanc(ln(X[bsb,.]))))');
                     Zt=Zmax;   k=1;
                        j=1; do until j==nla1+1;
                         if abs(la1[j]) >1e-5;
                           Zt=(X[bsb,.]^la1[j]-1
                                  )./(la1[j].*G^(la1[j]-1));
                         else;
                           Zt=G .*ln(X[bsb,.]);
                         endif;

                             Gri[k]=-mm*ln(det(vcx(Zt)));
                           j=j+1;   k=k+1;
                        endo;
                  @    _pnumht=0.3; @
                      if j==1;
                       ytics(round(minc(Gri))-30,round(maxc(gri))+5,2,1);
                      else;
                       endif;
                   mess= 0 $+ "m=" $+ ftocv(mm,2,0);
                            _PMSGSTR =mess;
                       _PMSGCTL=0.7~1~0.3~0~2~2;
                  @ " variable";; i; "Gri"; la1~Gri; wait;    @
                 /*******
                   output file=C:\gauss\sec\sbn44.prn reset;
                       la1~Gri;
                   output file=C:\gauss\sec\sbn44.prn off;
                        **************/
                  xy(la1,Gri);
            endwind;
            graphset;
          endif;
         else;
       endif;
       


      @ Order all MD and select the smallest m+1 @
      Xc=Xt[bsb,.];
      ym=meanc(Xc);

      vc=invpd(vcx(Xc));

      i=1;  do until i==nn+1;
          mala[i,2]=sqrt((Xt[i,.]-ym')*vc*(Xt[i,.]'-ym));
            i=i+1;
      endo;

   zs=sortc(mala,2);
   zs=zs|miss(zeros(10,2),0);
   @ Store in matrix In unit ranked mm-th (mm+9)-th @
   IN[mm-min+1,2:11]=zs[mm:mm+9,1]';

   @ pos has 3 columns: the first contains the dimension of clean data
      set,
    Columns 2, and 3 respectively contain
    Max, Average, forming the basic subset @

                @   maximum       average  @
   pos[mm-min+1,2:3]=zs[mm,2]~meanc(zs[1:mm,2]);


     if mm<nn;
     @ pot contains minimum of Mahalanobis distances among the units
       which form the group of potential outliers @
       pot[mm-min+1,2]=zs[mm+1,2];

     @ the dimension of the clean data set increases of one unit.
       vector bsb contains the indexes of the units
       forming the new basic subset @
     nbsb=zs[1:mm+1,1];

             @ Store in matrix Un the unit(s) included in this step @
             unit=setdif(nbsb,bsb,1);
             if rows(unit)<=10;
              UN[mm-min+1,2:rows(unit)+1]=(unit');
             else;
              "Warning: interchange greater than 10";
              UN[mm-min+1,2:11]=unit[1:10]';
             endif;
      bsb=nbsb;

                      else;
     endif;

                    mm=mm+1;
                  endo;
  @   graphset;  @
   retp(pos,pot,lr,likmax,In,Un);
endp;

