/*
**  FLRLD (Forward likelihood ratio for linear discriminant analysis)
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {lr,likmax,Un}=flrLD(bsb,sz,X,vsz,dd,ch,lamax,mis,d)
**
**                                To monitor the likelihodd ratio
**                                in every step of the forward procedure
**                                The purpose is to test the null
**                                hypothesis that the transformation
**                                parameters are as those specified in
**                                vector la. This program takes into
**                                account the presence of different
**                                universes
**
*/

/*
** FLRLD
**
** PURPOSE:  The principal purpose is to monitor the likelihood ratio in
**           every step of the forward procedure. The purpose is to test
**           the null
**           hypothesis that the transformation parameters are as those
**           specified in vector la.
**           It also produces average, maximum of Mahalanobis distances
**           of clean (and min of noclean) in  each step of
**           the forward procedure
**
** FORMAT:   {lr,likmax,Un} = flrLD(bsb,sz,X,vsz,dd,ch,lamax,mis,d)
**
** INPUT:    bsb = mm x 1 vector containing the units
**                belonging to the initial subset
**            sz = g x 1 vector which contains
**                the dimension of the g subgroups forming
**                the initial basic subset
**                   e.g.: if sz =50|40|30
**                   then size of 1st gr. is 50
**                   then size of 2nd gr. is 40
**                   then size of 3rd gr. is 30
**            X = nn x p. Full matrix
**          vsz = g x 1 vector which contains sizes of the groups
**          dd  = 2 x 1 vector specifying which kind of forward search
**                you want
**                if dd[1]=0 inclusion is based on MD without constraints
**                if dd[1]=1 inclusion is based on the difference
**                           among the sizes of the groups
**                  else     inclusion is based on the ratio
**                           among the sizes of the groups
**                element dd[2] is a suboption when dd[1] /= 0
**                if dd[2]=0  when you include a unit from
**                            universe j basic subset for
**                            universe k (k ne q) can change
**                 else       when you include a unit from
**                            universe j basic subset for
**                            universe k (k ne q) cannot change
**          ch  = c x 2 matrix or scalar 0
**                1st column = indexes of the variables (VAR)
**                             to be tested for transformation
**                2nd column = values of transformation
**                             parameters
**                REMARK: if ch=0 (scalar zero) you can test
**                         the null hypothesis that no variable
**                         must be transformed
**                EXAMPLE: if ch=(1~0)|
**                               (4~0.5)
**                         LR concerns VAR 1 and 4
**                         The null hypothesis is:
**                         log transformation for VAR 1
**                         and sqrt transformation for VAR 4
**        lamax = (k+1) x p matrix which contains (in the rows) the
**                values of MLE transformation parameters in the last
**                k+1 steps of the forward search
**                REMARK: the last row of matrix lamax refers to
**                MLE when m=n (last step of the fwd search)
**          mis = scalar.
**                If mis=1 we search the transformation parameters
**                which minimize the overall probability of
**                misallocation
**                else  we search the transformation parameters
**                which maximize the likelihood.
**           d  = l x 1 vector which enables the user to monitor the
**                likelihood grid of the parameters.
**                For examples if l={24,33} you can monitor the
**                likelihood grid when there are respectively 24 and 33
**                observations in the basic subset
**
** OUTPUT:    lr = (n-m+1 x 2) matrix. The first column contains the
**                 dimension of
**                 the basic subset, the second the values of the
**                 likelihood ratio in each
**                 step of the forward search
**        likmax = (n-m+1 x 2) matrix. the first column contains the
**                 size of the basic subset, the second the value of the
**                 maximazied likelihood
**                 Likmax= -0.5*( mm*p*ln(2*pi)
**                         +mm*ln(|W/n|) +mm*p );
**                 W= within groups sum of squares and product matrix
**                     W= \sum_i^g \hat{\Sigma_i}*n_i
**           Un  = n-m x 10 matrix
**                 1st col = fwd search index
**                 2nd col:10th col = unit(s) which entered the basic
**                 subset
**
*/


 proc (3)=flrLD(bsb,sz,X,vsz,dd,ch,lamax,mis,d);

   local seq,ym,vc,
       nn,n,p,mm,i,zs,clean,min,mala,y,ii,lr,mess,
       j,G,z,Xc,dZ,dZmax,Zmax,la1,gri,lat,k,zt,Likmax,nla1,Un,unit,
       eqsz,m,vZ,iVZ,id,Ident,la0,Le1,xr,de,a,Te,ord,ab1,nbsb,GG,
       Teall,sel,Grit,pt,ptmax,spt,sptmax,Xt,la,Xcc,lim,confint,on;

      nn=rows(X);
      p=cols(X);
      seq=SEQA(1,1,nn);
      mm= rows(bsb);
      min=mm;
      g=rows(sz);    @ g= number of groups @

  @ Matrix lr which will contain the LR test in each step of the
    forward search @
  lr=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,1);

   if nn>mm;
      @  matrix Un will contain from 2nd col. to 10th the unit(s) which
      entered the basic subset @
      Un=seqa(mm+1,1,nn-mm)~miss(zeros(nn-mm,9),0);
   else;
      Un=miss(0,0);
  endif;

  @ Matrix likmax which will contain the max of the profile
  likelihood in each step of the
  forward search @
       Likmax=lr;


  eqsz=seqa(1,1,g)~sz;    @ For the iris data for example
                                 eqsz= 1  37
                                       2  39
                                       3  43
                        Later matrix eqsz will bew changed @

    @  Matrix M contains (in the rows) the vector of the means for
       the groups
       first row refers to first group
       ...
       last row refers to the last group  @
    M=zeros(g,p);

    @ Ident = n x 1 vector which contains the labels of the units
      referred to the different groups @
           id=0;
           i=1; do until i==g+1;
           id=id|reshape(i,vsz[i],1);
             i=i+1;
           endo;
          Ident=id[2:nn+1];

    @ Mala= nn x 2 matrix whose 2nd col contains MD of
      unit from its own centroid @
     Mala=seq~zeros(nn,1);

    Le1=eqsz[.,1]~zeros(g,1);
     @ e.g. for the iris
           1   0
    Le1=   2   0
           3   0   @

          vsz=0|vsz;
          sz=0|sz;
      @        0                  0
         vsz = 50           sz=   36
               50                 37
               50                 36   @

   @ ch[.,1] specifies the variables to analyze in the LR test
     ch[.,2] controls how the variable nust be transformed in the
     fwd search @
       if rows(ch)==1 and ch==0;
           sel=seqa(1,1,p);
           la=ones(p,1);
       else;
           sel=ch[.,1];
           la=ch[.,2];
       endif;
       @ Divide each each column by its maximum @
       @  X=X./(maxc(X)'); @

          @ Create Xt=matrix of transformed values to be used
             in the forward search (for ordering MD) @
          Xt=X;

        j=1; do until j==rows(sel)+1;
           if la[j] NE 0;
                  Xt[.,sel[j]]= X[.,sel[j]]^la[j];
              else;
                  Xt[.,sel[j]]=ln(X[.,sel[j]]);
             endif;
           j=j+1;
        endo;

  Xcc=X[bsb,.];  @ Xcc = Subset of original data @
  Xc=Xt[bsb,.];  @ Xc = subset of transformed data @

 mm=mm; do until mm==nn+1;

  @   vcx(Xc); wait;     @
  Z=Xcc;
  Zmax=Z;
  j=1;  do until j==rows(la)+1;
    GG=exp(meanc(ln(Xcc[.,sel[j]])));
     if la[j] NE 0;
             Z[.,sel[j]]=(Xcc[.,sel[j]]^la[j]-1)/(la[j]*GG^(la[j]-1));
       else;
             Z[.,sel[j]]=GG *ln(Xcc[.,sel[j]]);
      endif;

    if lamax[mm-min+1,j] NE 0;
             Zmax[.,sel[j]]=(Xcc[.,sel[j]]^lamax[mm-min+1,sel[j]]-1)/(
              lamax[mm-min+1,sel[j]]*GG^(lamax[mm-min+1,sel[j]]-1));
       else;
             Zmax[.,sel[j]]=GG*ln(Xcc[.,sel[j]]);
      endif;

    j=j+1;
  endo;






      @ dZ= Value of the function under the null hypothesis @
      @ dZmax= Value of the function under maximum likelihood @
       dZ=0;  dZmax=0;
  if mis==1;

  {pt}=ldasimpl(Z,sz[2:g+1],ones(g,1)/g);
  {ptmax}=ldasimpl(Zmax,sz[2:g+1],ones(g,1)/g);
    i=1; do until i==g+1;
        @ mis matrix which contains the overall prob. of misclass.
          for the various groups @
         spt=pt[sumc(sz[1:i])+1:sumc(sz[1:i+1]),1];
         sptmax=ptmax[sumc(sz[1:i])+1:sumc(sz[1:i+1]),1];
         dZ= dZ+ sumc(1-spt);
         dZmax= dZmax+ sumc(1-sptmax);
          i=i+1;
     endo;

      @ LR=m*ln(|W(l0)|/|W(l)|)  @
       lr[mm-min+1,2]= dZmax-dZ;
      Likmax[mm-min+1,2]= dZmax;
  else;
    i=1; do until i==g+1;
      dZ= dZ+
      sz[i+1]*(vcxm(Z[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]));

      dZmax= dZmax+
      sz[i+1]*(vcxm(Zmax[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]));

          i=i+1;
     endo;
           @ Likmax[mm-min+1,2]= -0.5*( mm*p*ln(2*pi)
                         +mm*ln((1/mm)^p * dZmax) +mm*p ); @
       Likmax[mm-min+1,2]=-mm*ln(det(dZmax/sumc(sz)));
 
      DZmax=det(dZmax);
      dZ=det(dZ);

      @ LR=m*ln(|W(l0)|/|W(l)|)  @
      lr[mm-min+1,2]= mm*ln(dZ/dZmax);

  endif;




   @    "vcx(zmax)";  det(vcx(Zmax));  wait;    @
       @ Plot LIkelihood grids if requested @
        if maxc(mm.==d)==1;
         if mis==1;
            begwind;
            window(2,3,0);
            makewind(80,6.8555,0,0,1);

        

             la1=seqa(-2.5,0.1,50);
             nla1=rows(la1);
             xtics(la1[1],la1[nla1],0.1,1);
             Gri=zeros(nla1,1);
               i=1; do until i==rows(sel)+1;
                     GG=exp(meanc(ln(Xcc[.,sel[i]])));
                     Zt=Zmax;   k=1; confint=0;
                        j=1; do until j ==nla1+1;
                     Grit=0;
                          if abs(la1[j]) >1e-7;
                    Zt[.,sel[i]]=(Xcc[.,sel[i]]^la1[j]-1)/(la1[j]*GG^(la1[j]-1));
                          else;
                     Zt[.,sel[i]]=GG *ln(Xcc[.,sel[i]]);
                           endif;

                       {pt}=ldasimpl(Zt,sz[2:g+1],ones(g,1)/g);
                         ii=1; do until ii==g+1;
                              spt=pt[sumc(sz[1:ii])+1:sumc(sz[1:ii+1]),1];
                              Grit= grit+ sumc(1-spt);
                               ii=ii+1;
                          endo;
                             Gri[k]= Grit;
                               j=j+1;   k=k+1;
                           endo;

                       _pnumht=0.3;
                    if i==1;
                     @  scale(miss(0,0),Gri);
                       xtics(la1[1],la1[nla1],0.1,1);
                      @
                       ytics(round(minc(Gri))-4,round(maxc(gri))+3,
                    (ceil(maxc(Gri))-trunc(minc(Gri)))/10,1);
                    else;
                    endif;
                         mess= 0 $+ "m=" $+ ftocv(mm,3,0) $+ " "$+
                               "X"$+FTOCV(sel[i],1,0);
                            _PMSGSTR=mess;
                       _PMSGCTL=0.7~1~0.3~0~2~2;
                   /* " variable";; i; "Gri"; la1~Gri; wait; */
                  xy(la1,Gri);
            nextwind;
               i=i+1;
               endo;
            endwind;
            graphset;
          else;  @ if mis not equal 1 @
             @ vector on will be used in matrix _pline to draw
               99% confidence bands of transformation parameter @
              on=ones(3,1);
            graphset;
            begwind;
            if rows(la)<=3;
              window(2,rows(la),0);
            elseif rows(la)<=4;
              window(2,2,0);
            elseif rows(la)<=6;
              window(2,3,0);
            elseif rows(la)<=12;
              window(3,4,0);
            elseif rows(la)<=15;
              window(3,5,0);
            else;
              window(4,6,0);
            endif;
            makewind(80,6.8555,0,0,1);

                @ vector lim will contain the x coordinates of the
                     confidence interval for lambda
                     lim[1,1] = lower limit
                     lim[2,1] = lower limit
                     lim[3,1] = mle   @
                   lim=miss(zeros(3,1),0);

             la1=seqa(-1,0.01,251);
             nla1=rows(la1);
             xtics(round(la1[1]),round(la1[nla1]),0.1,0);
             xtics(-1,1.4,0.1,0);

             Gri=zeros(nla1,1);
               i=1; do until i==rows(sel)+1;
                     GG=exp(meanc(ln(Xcc[.,sel[i]])));
                     Zt=Zmax;   k=1;   confint=0;

                        j=1; do until j ==nla1+1;
                     Grit=0;
                          if abs(la1[j]) >1e-7;
                             Zt[.,sel[i]]=(Xcc[.,sel[i]]^la1[j]-1)
                              /(la1[j]*GG^(la1[j]-1));
                          else;
                             Zt[.,sel[i]]=GG *ln(Xcc[.,sel[i]]);
                           endif;
                          ii=1; do until ii==g+1;
                             Grit= Grit +
                              sz[ii+1]*(vcxm(Zt[(
                               sumc(sz[1:ii])+1):sumc(sz[1:ii+1]),.]));
                                ii=ii+1;
                           endo;
                             Gri[k]= -mm*ln(det(Grit/sumc(sz)));


@ "mm";; mm;
i; 
likmax[mm-min+1,2];
gri[k]; wait;
la1[j]; 
Likmax[mm-min+1,2];
Gri[k]; 
 wait; @

                             @ The following if finds the conf. interval
                              for each transformation parameter @
                            if Likmax[mm-min+1,2]-Gri[k]<=0.5*3.84
                                   and confint==0;
                                    lim[1]=la1[j];
                                    confint=1;
                           @ Likmax[mm-min+1,2]; Gri[k]; la1[j]; wait; @
                            elseif Likmax[mm-min+1,2]-Gri[k]>=0.5*3.84
                                   and confint==1;
                                    lim[2]=la1[j];
                                  confint=2;
                           @ Likmax[mm-min+1,2]; Gri[k]; la1[j]; wait; @
                           lim[3]=lamax[mm-min+1,i];
                               @   "mm-min+1";; mm-min+1;  wait;
                                  lamax[mm-min+1,.]; wait;
                                           "DIFFF";
                                     lim[1]-lamax[mm-min+1,i];
                                     lamax[mm-min+1,i]-lim[2]; wait; @
                            else;
                            endif;


                               j=j+1;   k=k+1;
                           endo;
                       _pnumht=0.3;
                    if i==1;
                       scale(miss(0,0),gri);
                       xtics(-1,1.4,0.1,0);
                    else;
                    endif;
mess= 0 $+ "m=" $+ ftocv(mm,3,0) $+ " " $+ "y" $+ ftocv(sel[i],1,0);
                            _PMSGSTR=mess;
                       _PMSGCTL=0.7~1~0.3~0~2~2;
                   /* " variable";; i; "Gri"; la1~Gri; wait; */
   _pline=on~6*on~lim~-1e+10*on~lim~1e+10*on~2*on~
                        2*on~0.4*on;
                  xy(la1,Gri);
            nextwind;
               i=i+1;
               endo;
            endwind;
            graphset;
         endif;
        else;
       endif;



    @ Compute mean, pooled varcov, and
       (inv(varcov)) for each population  @
     VZ=0;
    i=1; do until i==g+1;
      M[i,.]=(meanc(Xc[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]))';
      VZ= Vz + vcxm(Xc[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]) * sz[i+1];
       @  VZ; wait;    @
          i=i+1;
     endo;
      VZ=VZ/(mm-g);  @ Pooled var cov matrix @
      iVZ=invpd(VZ);

                  @ Compute Vector of MD from each appropriate
                    population @

      i=1;  do until i==nn+1;
                      @ MD of each unit from every population @
         Xr=(Xt[i,.]-M[ident[i],.]);
         Mala[i,2]=sqrt(Xr*iVz*(Xr'));
            i=i+1;
      endo;

     @   "vcxmala";   vcx(Mala); wait; @

        zs=sortc(Mala,2);

  if mm<nn;
    @ if dd[1] /= 0 Include alternatively a unit from each group
          according to difference (if dd[1]=1)
          or according to ratio (if dd[1] ne 1) @
    if dd[1] /= 0;
       @ Matrix Teall contains sorted MD for each of the g populations @
       j=1; do until j==g+1;
            @ Te is a matrix whose 2nd col contains MD for
              universe j ordered @
         if j==1;
           Teall=sortc(mala[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),.],2);
         else;
           Teall=Teall|sortc(mala[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),.],2);
         endif;
         j=j+1;
       endo;
       @ Second col. of eqsz contains dimension of basic subsets @
       @ Vector de contains the index of the universes
         whose size must be increased @
      if dd[1]==1;
       a=vsz[2:g+1]-eqsz[.,2];
       de=delif(eqsz[.,1],a.<maxc(a));
      else;
       a=eqsz[.,2]./vsz[2:g+1];
       de=delif(eqsz[.,1],a.>minc(a));
      endif;
                               @ vsz[2:g+1]~eqsz[.,2];
                                de;  wait;      @


        Le1=de~zeros(rows(de),1); @ 1 \leq rows(Le1) \leq g
                                 where g is the number of the groups @
       @ Among the subsets which have comparable size insert the unit
        with the smallest Mahalanobis distance @
       i=1; do until i==rows(de)+1;
           j=de[i];
            @ Te is a matrix whose 2nd col contains MD for
              universe j ordered @
      @ Te=sortc(mala[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),.],2); @
            @ second col. of matrix Le1 will contain MD which occupy
              position mm+1 in the groups of equal sizes.
              sz is a vector which contains the sizes of the BSBs
              REMARK: sz has a 0 in position [1] this is the reason
              why in the next line we write sz[j+1] @
           Le1[i,2]=Teall[sumc(vsz[1:j])+sz[j+1]+1,2];
           i=i+1;
       endo;
      ord=sortc(Le1,2);
      ab1=ord[1,1];  @ ab1 is the group with the smallest MD
                     among the groups of comparable size @


       eqsz[ab1,2]=eqsz[ab1,2]+1; @ dim. of basic subset for univ.
                                       ab1 is increased by one unit @

             @ Store in matrix Un the unit(s) included in this step @
     Te=Teall[sumc(vsz[1:ab1])+1:sumc(vsz[1:ab1])+sz[ab1+1]+1,1];
         if dd[2]==0;
           @ select all groups apart the ab1-th @
           sel=setdif(eqsz[.,1],ab1,1);
           nbsb=Te;
            i=1; do until i==g;
              j=sel[i];
              nbsb=nbsb|Teall[sumc(vsz[1:j])+1:sumc(vsz[1:j])+sz[j+1],1];
             i=i+1;
            endo;
         else;
             nbsb=delif(bsb,bsb.>= sumc(vsz[1:ab1])+1
                       .and bsb .<=sumc(vsz[1:ab1+1]));
             @ nbsb=nbsb|Te[1:sz[ab1+1]+1,1];  @
              nbsb=nbsb|Te;
          endif;
           @  Te=Te[.,1]|miss(zeros(10,1),0);    @
            sz[2:g+1]=eqsz[.,2];   @ new sizes for new basic subset @

    else;
             zs=zs[.,1]|miss(zeros(10,1),0);
             nbsb=zs[1:mm+1,1];
              @ Create  sz[2:g+1]= vector which contains
              the new sizes of the groups  @
              j=1; do until j==g+1;
                sz[j+1]=rows(selif(nbsb,
                nbsb.>= sumc(vsz[1:j])+1 .and nbsb .<= sumc(vsz[1:j+1])));
               j=j+1;
              endo;
    endif;
            @  vcx(bsb); vcx(nbsb); wait; @
             @ Store in matrix Un the unit(s) included in this step @
             unit=setdif(nbsb,bsb,1);
             if rows(unit)<=9;
              UN[mm-min+1,2:rows(unit)+1]=(unit');
             else;
              "Warning: interchange greater than 10";
              UN[mm-min+1,2:10]=unit[1:9]';
             endif;
    bsb=sortc(nbsb,1);
    Xc=Xt[bsb,.];
    Xcc=X[bsb,.];
   else;
  endif;
 mm=mm+1;
endo;

   retp(lr,likmax,Un);
endp;

