/*
**  FLRQD (Forward likelihood ratio for quadr. discriminant analysis)
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {lr,likmax,Un}=flrQD(bsb,sz,X,vsz,dd,ch,lamax,d)
**
**                                To monitor the likelihodd ratio
**                                in every step of the forward procedure
**                                The purpose is to test the null
**                                hypothesis that the transformation
**                                parameters are as those specified in
**                                vector la. This program takes into
**                                account the presence of different
**                                universes
**
*/

/*
** FLRQD
**
** PURPOSE:  The principal purpose is to monitor the likelihood ratio in
**           every step of the forward procedure. The purpose is to test
**           the null
**           hypothesis that the transformation parameters are as those
**           specified in vector la.
**           It also produces average, maximum of Mahalanobis distances
**           of clean (and min of noclean) in  each step of
**           the forward procedure
**
** FORMAT:   {lr,likmax,Un} = flrQD(bsb,sz,X,vsz,dd,ch,lamax,d)
**
** INPUT:    bsb = mm x 1 vector containing the units
**                belonging to the initial subset
**            sz = g x 1 vector which contains
**                the dimension of the g subgroups forming
**                the initial basic subset
**                   e.g.: if sz =50|40|30
**                   then size of 1st gr. is 50
**                   then size of 2nd gr. is 40
**                   then size of 3rd gr. is 30
**            X = nn x p. Full matrix
**          vsz = g x 1 vector which contains sizes of the groups
**          dd  = scalar specifying which kind of forward search
**                you want
**                if dd=0 inclusion is based on MD without constraints
**                if dd=1 inclusion is based on the difference
**                           among the sizes of the groups
**                  else     inclusion is based on the ratio
**                           among the sizes of the groups
**          ch  = c x 2 matrix or scalar 0
**                1st column = indexes of the variables (VAR)
**                             to be tested for transformation
**                2nd column = values of transformation
**                             parameters
**                REMARK: if ch=0 (scalar zero) you can test
**                         the null hypothesis that no variable
**                         must be transformed
**                EXAMPLE: if ch=(1~0)|
**                               (4~0.5)
**                         LR concerns VAR 1 and 4
**                         The null hypothesis is:
**                         log transformation for VAR 1
**                         and sqrt transformation for VAR 4
**        lamax = (k+1) x p matrix which contains (in the rows) the
**                values of MLE transformation parameters in the last
**                k+1 steps of the forward search
**                REMARK: the last row of matrix lamax refers to
**                MLE when m=n (last step of the fwd search)
**          mis = scalar.
**                If mis=1 we search the transformation parameters
**                which minimize the overall probability of
**                misallocation
**                else  we search the transformation parameters
**                which maximize the likelihood.
**           d  = l x 1 vector which enables the user to monitor the
**                likelihood grid of the parameters.
**                For examples if l={24,33} you can monitor the
**                likelihood grid when there are respectively 24 and 33
**                observations in the basic subset
**
** OUTPUT:    lr = (n-m+1 x 2) matrix. The first column contains the
**                 dimension of
**                 the basic subset, the second the values of the
**                 likelihood ratio in each
**                 step of the forward search
**        likmax = (n-m+1 x 2) matrix. the first column contains the
**                 size of the basic subset, the second the value of the
**                 maximazied likelihood
**                 likmax= -0.5*( m*p* \ln(2*\pi)+
**                         \sum_i^g n_i \ln |\hat{\Sigma_i}| +mm*p )
**           Un  = n-m x 10 matrix
**                 1st col = fwd search index
**                 2nd col:10th col = unit(s) which entered the basic
**                 subset
**
*/


 proc (3)=flrQD(bsb,sz,X,vsz,dd,ch,lamax,mis,d);

   local seq,ym,vc,
       nn,n,p,mm,i,zs,clean,min,mala,y,ii,lr,mess,
       j,G,z,Xc,dZ,dZmax,Zmax,la1,gri,lat,k,zt,Likmax,nla1,Un,unit,
       eqsz,m,vZ,iVZ,id,Ident,la0,Le1,xr,de,a,Te,ord,ab1,nbsb,GG,
       grit,pt,spt,Xt,la,sel,Xcc;
      nn=rows(X);
      p=cols(X);
      seq=SEQA(1,1,nn);
      mm= rows(bsb);
      min=mm;
      g=rows(sz);    @ g= number of groups @

  @ Matrix lr which will contain the LR test in each step of the
    forward search @
  lr=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,1);

   if nn>mm;
      @  matrix Un will contain from 2nd col. to 10th the unit(s) which
      entered the basic subset @
      Un=seqa(mm+1,1,nn-mm)~miss(zeros(nn-mm,9),0);
   else;
      Un=miss(0,0);
  endif;
  @ Matrix likmax which will contain the max of the profile
  likelihood in each step of the
  forward search @
       Likmax=lr;


  eqsz=seqa(1,1,g)~sz;    @ For the iris data for example
                                 eqsz= 1  37
                                       2  39
                                       3  43
                        Later matrix eqsz will bew changed @

    @  Matrix M contains (in the rows) the vector of the means for
       the groups
       first row refers to first group
       ...
       last row refers to the last group  @
    M=zeros(g,p);

    @  Matrix VZ contains the varcov matrices of every
       group
       first p rows refer to first group
       ...
       last p rows refer to last group  @
    VZ=zeros(p*g,p);

    @  Matrix iVZ contains inv(varcov) of every
       group
       first p rows refer to first group
       ...
       last p rows refer to last group  @
    iVZ=zeros(p*g,p);
    @ Ident = n x 1 vector which contains the labels of the units
      referred to the different groups @
           id=0;
           i=1; do until i==g+1;
           id=id|reshape(i,vsz[i],1);
             i=i+1;
           endo;
          Ident=id[2:nn+1];

    @ Mala= nn x 2 matrix whose 2nd col contains MD of
      unit from its own centroid @
     Mala=seq~zeros(nn,1);

    Le1=eqsz[.,1]~zeros(g,1);
     @ e.g. for the iris
           1   0
    Le1=   2   0
           3   0   @

          vsz=0|vsz;
          sz=0|sz;
      @        0                  0
         vsz = 50           sz=   36
               50                 37
               50                 36   @

   @ ch[.,1] specifies the variables to analyze in the LR test
     ch[.,2] controls how the variable nust be transformed in the
     fwd search @
       if rows(ch)==1 and ch==0;
           sel=seqa(1,1,p);
           la=ones(p,1);
       else;
           sel=ch[.,1];
           la=ch[.,2];
       endif;

          @ Create Xt=matrix of transformed values to be used
             in the forward search (for ordering MD) @
          Xt=X;

        j=1; do until j==rows(sel)+1;
           if la[j] NE 0;
                  Xt[.,sel[j]]= X[.,sel[j]]^la[j];
              else;
                  Xt[.,sel[j]]=ln(X[.,sel[j]]);
             endif;
           j=j+1;
        endo;

  Xcc=X[bsb,.];  @ Xcc = Subset of original data @
  Xc=Xt[bsb,.];  @ Xc = subset of transformed data @
 mm=mm; do until mm==nn+1;

  Z=Xcc;
  Zmax=Z;
  j=1;  do until j==rows(la)+1;
    GG=exp(meanc(ln(Xcc[.,sel[j]])));
     if la[j] NE 0;
             Z[.,sel[j]]=(Xcc[.,sel[j]]^la[j]-1)/(la[j]*GG^(la[j]-1));
       else;
             Z[.,sel[j]]=GG *ln(Xcc[.,sel[j]]);
      endif;

    if lamax[mm-min+1,j] NE 0;
             Zmax[.,sel[j]]=(Xcc[.,sel[j]]^lamax[mm-min+1,sel[j]]-1)/(
       lamax[mm-min+1,sel[j]]*GG^(lamax[mm-min+1,sel[j]]-1));
       else;
             Zmax[.,sel[j]]=GG*ln(Xcc[.,sel[j]]);
      endif;

    j=j+1;
  endo;
   @    "vcx(zmax)";  det(vcx(Zmax));  wait;    @


       @ Plot LIkelihood grids if requested @
        if maxc(mm.==d)==1;
         if mis==1;
            begwind;
            window(2,3,0);
            makewind(80,6.8555,0,0,1);

             la1=seqa(-3,0.1,81);
             nla1=rows(la1);
             xtics(la1[1],la1[nla1],0.1,1);
             Gri=zeros(nla1,1);
               i=1; do until i==rows(sel)+1;
                     GG=exp(meanc(ln(Xcc[.,sel[i]])));
                     Zt=Zmax;   k=1;
                        j=1; do until j ==nla1+1;
                     Grit=0;
                          if abs(la1[j]) >1e-7;
                            Zt[.,sel[i]]=(Xcc[.,sel[i]]^la1[j]-1)
                              /(la1[j]*GG^(la1[j]-1));
                          else;
                             Zt[.,sel[i]]=GG *ln(Xcc[.,sel[i]]);
                          endif;

                       {pt}=qdasimpl(Zt,sz[2:g+1],ones(g,1)/g);
                         ii=1; do until ii==g+1;
                              spt=pt[sumc(sz[1:ii])+1:sumc(sz[1:ii+1]),1];
                              Grit= grit+ meanc(1-spt);
                               ii=ii+1;
                          endo;
                             Gri[k]= Grit;
                               j=j+1;   k=k+1;
                           endo;

                       _pnumht=0.3;
                    if i==1;
                      scale(miss(0,0),Gri); @
                    ytics(round(minc(Gri))-2,round(maxc(gri))+2,
                    (ceil(maxc(Gri))-trunc(minc(Gri)))/10,1); @
                    else;
                    endif;
                         mess= 0 $+ "m=" $+ ftocv(mm,3,0) $+ " "$+
                               "V"$+FTOCV(sel[i],1,0);
                            _PMSGSTR=mess;
                       _PMSGCTL=0.7~1~0.3~0~2~2;
                   /* " variable";; i; "Gri"; la1~Gri; wait; */
                  xy(la1,Gri);
            nextwind;
               i=i+1;
               endo;
            endwind;
            graphset;
         else;   @ Normal likelihood grids of transf. par. @
            begwind;
            window(2,3,0);
            makewind(80,6.8555,0,0,1);

             la1=seqa(-2,0.1,41);
             nla1=rows(la1);
             xtics(la1[1],la1[nla1],0.1,1);
             Gri=zeros(nla1,1);
               i=1; do until i==rows(sel)+1;
                     GG=exp(meanc(ln(Xcc[.,sel[i]])));
                     Zt=Zmax;   k=1;
                        j=1; do until j ==nla1+1;
                     Grit=0;
                          if abs(la1[j]) >1e-7;
                             Zt[.,sel[i]]=(Xcc[.,sel[i]]^la1[j]-1)/
                               (la1[j]*GG^(la1[j]-1));
                          else;
                        Zt[.,sel[i]]=GG *ln(Xcc[.,sel[i]]);
                           endif;
                          ii=1; do until ii==g+1;
                            Grit= Grit+
                            sz[ii+1]*ln(det(vcxm(Zt[(
                            sumc(sz[1:ii])+1):sumc(sz[1:ii+1]),.])));
                                ii=ii+1;
                           endo;
                             Gri[k]= -Grit;
                               j=j+1;   k=k+1;
                           endo;
                       _pnumht=0.3;
                    if i==1;
                       scale(miss(0,0),gri);
                    else;
                    endif;
                         mess= 0 $+ "m=" $+ ftocv(mm,3,0) $+ " "$+
                               "V"$+FTOCV(sel[i],1,0);
                            _PMSGSTR=mess;
                       _PMSGCTL=0.7~1~0.3~0~2~2;
                   /* " variable";; i; "Gri"; la1~Gri; wait; */
                  xy(la1,Gri);
            nextwind;
               i=i+1;
               endo;
            endwind;
            graphset;
         endif;
        else;
       endif;
      @ Z; wait; rows(Z); Z~X[bsb,.]; wait;   @
      @ dZ= Value of the function under the null hypothesis @
      @ dZmax= Value of the function under maximum likelihood @
       dZ=0;  dZmax=0;
    i=1; do until i==g+1;
     dZ= dZ+
        sz[i+1]*ln(det(vcxm(Z[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.])));

     dZmax= dZmax+
       sz[i+1]*ln(det(vcxm(Zmax[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.])));

          i=i+1;
     endo;
      @ " dZmax"; dZmax;  wait;   @
      lr[mm-min+1,2]= dZ-dZmax;     @ LR= 2(l_1 - l_0) @

      Likmax[mm-min+1,2]= -0.5*( mm*p*ln(2*pi)+dZmax+mm*p );

    @ Compute mean, varcov, and (inv(varcov)) for each population @
    i=1; do until i==g+1;
      M[i,.]=(meanc(Xc[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]))';
      VZ[(1+p*(i-1)):p*i,.]=vcx(Xc[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]);
   @   VZ[(1+p*(i-1)):p*i,.]; wait; @
      iVZ[(1+p*(i-1)):p*i,.]=invpd(VZ[(1+p*(i-1)):p*i,.]);
          i=i+1;
     endo;

                  @ Compute Vector of MD from each appropriate
                    population @

      i=1;  do until i==nn+1;
                      @ MD of each unit from every population @
         Xr=(Xt[i,.]-M[ident[i],.]);
         Mala[i,2]=Xr*iVz[((ident[i]-1)*p+1):(ident[i]*p),.]*(Xr');
            i=i+1;
      endo;

        zs=sortc(Mala,2);

  if mm<nn;

    @ if dd /= 0 Include alternatively a unit from each group
          according to difference (if dd=1)
          or according to ratio (if dd ne 1) @
    if dd /= 0;
       @ Second col. of eqsz contains dimension of basic subsets @
       @ Vector de contains the index of the universes
         whose size must be increased @
      if dd==1;
       a=vsz[2:g+1]-eqsz[.,2];
       de=delif(eqsz[.,1],a.<maxc(a));
      else;
       a=eqsz[.,2]./vsz[2:g+1];
       de=delif(eqsz[.,1],a.>minc(a));
      endif;
                            @    vsz[2:g+1]~eqsz[.,2];
                                de;  wait;   @

       Le1=de~zeros(rows(de),1); @ 1 \leq rows(Le1) \leq g
                                 where g is the number of the groups @
       @ Among the subsets which have comparable size insert the
         unit with the smallest Mahalanobis distance @
      i=1; do until i==rows(de)+1;
         j=de[i];
          @ Te is a matrix whose 2nd col contains MD ordered @
          Te=sortc(mala[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),.],2);
          @ second row of matrix Le1 will contain MD which occupies
            position mm+1 in the groups of equal sizes.
            sz is a vector which contains the sizes of the BSBs
            Remark sz has a 0 in position 1 this is the reason
            why in the next line we write sz[j+1] @
        Le1[i,2]=Te[sz[j+1]+1,2];
        i=i+1;
      endo;

      ord=sortc(Le1,2);
      ab1=ord[1,1];  @ ab1 is the group with the smallest MD
                     among the groups of comparable size @


       eqsz[ab1,2]=eqsz[ab1,2]+1; @ dim. of basic subset for univ.
                                       ab1 is increased by one unit @

      Te=sortc(mala[sumc(vsz[1:ab1])+1:sumc(vsz[1:ab1+1]),.],2);

             @ Store in matrix Un the unit(s) included in this step @
             nbsb=delif(bsb,bsb.>= sumc(vsz[1:ab1])+1
                       .and bsb .<=sumc(vsz[1:ab1+1]));
             nbsb=nbsb|Te[1:sz[ab1+1]+1,1];

            sz[2:g+1]=eqsz[.,2];   @ new sizes for new basic subset @

    else;

             zs=zs[.,1]|miss(zeros(10,1),0);
             nbsb=zs[1:mm+1,1];
              @ Create  sz[2:g+1]= vector which contains
              the new sizes of the groups  @
              j=1; do until j==g+1;
                sz[j+1]=rows(selif(nbsb,
                nbsb.>= sumc(vsz[1:j])+1 .and nbsb .<= sumc(vsz[1:j+1])));
               j=j+1;
              endo;

     endif;
     unit=setdif(nbsb,bsb,1);
     if rows(unit)<=9;
       Un[mm-min+1,2:rows(unit)+1]=(unit');
     else;
       "Warning interchange greater than 9 when mm=";;  mm;
       Un[mm-min+1,2:10]=(unit[1:9]');
     endif;

     bsb=sortc(nbsb,1);
     Xc=Xt[bsb,.];
     Xcc=X[bsb,.];
   else;
  endif;

 mm=mm+1;
endo;

   retp(lr,likmax,Un);
endp;

