/*
**  FWDGLM
**  written by MARCO RIANI
**
**  FORMAT                                    PURPOSE               Line
** =======================================================================
**
**  {BB,TT,Rglm,LEV,WEI,LIK,Rst,Coo,Sco,Un,diff,mIn}=
**                     fwdglm(X,y,q,freq,mint,tol,bsb,sh,stab,binary)
**                                            To perform a fwd search
**                                            for GLM models
**
**
*/


/*
**
** FWDGLM
**
** PURPOSE:  to perform a fwd search for GLM
**           In every step of the search a lot of
**           quantities are monitored
**
**
**
** FORMAT:
**
** {BB,TT,Rglm,LEV,WEI,LIK,Rst,Coo,Sco,Un,diff,mIn}
**                      = fwdglm(X,y,q,freq,mint,tol,bsb,sh,stab,binary)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             q  = scalar which specifies the kind of GLM
**                  q=1 ==> Poisson link
**                  q=2 ==> Logit link
**                  q=3 ==> Complementary Log Log link
**                  q=4 ==> Gamma link
**                  q=5 ==> Inverse Gaussian link
**                  q=6 ==> Probit link
**                  q=7 ==> Arc sin link
**                  q=8 ==> Modified Logit Link
**           freq = n x 1 vector of frequencies
**                  if fr = scalar 0 a n x 1 vector of 1s created
**           mint = scalar. If mint is equal to 1
**                  the null model is also considered
**            tol = tolerance level for parameter estimation
**             sh = scalar specifying the kind of GLM residuals
**                  to be used in the fwd search
**                  sh=1 ==> fwd search on simple residuals
**                  sh=2 ==> fwd search on Pearson residuals
**                  sh=3 ==> fwd search on deviance residuals
**                  sh=4 ==> fwd search on Studentized Pearson residuals
**                           McN p. 397 eq. (12.4)
**                  sh=5 ==> fwd search on Studentized Deviance residuals
**                           McN p. 397  eq. (12.5)
**           stab = scalar 
**                  stab=0 implies that every unit can leave the subset
**                  stab=1 implies that no unit can leave the subset
**                  stab=2 implies that every unit can leave the subset
**                         except those forming the initial subset
**                  stab=3 every unit can leave the subset but
**                         units of the initial subset cannot leave before m<= n*(3/4)
**
**          binary = scalar (it applies only when the response is 0/1    
**                   binary = 0 implies inclusion without constraints 
**                   binary = 1 units are included according to
**                         the difference between 0s and 1s
**                   binary =2 units are included according to
**                         the ratio between 0s and 1s 
**           
**
**
**
** OUTPUT:  BB = n-k x p matrix
**               beta coefficients in each step of the fwd search
**          tt = n-k x p matrix
**               t test in each step of the fwd seach
**        rGLM = n x n-k+1 matrix containins GLM residuals for all the
**               units in each step
**               the kind of glm residuls is specified in input sh
**        LEV  = n x n-k+1 matrix containing
**               diagonal elements of generalized Hat matrix
**               McN p. 397 eq. (12.3) in each step
**        WEI  = n x n-k+1 matrix containing
**               the weights associated to the n units in each step
**        LIK  = n-k  x 4 matrix which contains:
**               1st  col = forward search index
**               2nd  col = deviance
**               3rd  col = pseudo R^2
**               4th  col = dispersion parameter
**        RST  = matrix which will contain max glm residual inside
**               bsb in each step
**        COO  = matrix which will contain Cook distances
**               in each step
**        SCO  = matrix containing the value of the score test
**               in each step
**        UN   = matrix which will contain the units entered in
**               each step
**       diff  = matrix which contains the number of units entering the subset
**               in each step of the fwd search  
**        mIn  = units forning the complementary of basic subset in each step of the fwd search
**
**
**  REMARK1    this program does not add a column of 1 in matrix X to
**             create the intercept.
**
**  REMARK2    The iterative least squares method is used for
**             parameters estimation
**
**  REMARK3    gauss function trap is set equal to 1
**
*/


proc (12)=fwdglm(X,y,q,freq,mint,tol,bsb,sh,stab,binary);
   local    n,p,i,j,R,temp,mIn,BB,TT,PVT,LIK,RST,Coo,
            LEV, WEI,jk,unit,UN,mm,seq,eq_0,eq_1,k,
            B,Rs,W,Rsq,cnt,mX,vY,vfreq,Bib,Tl,Hi,Mo,gap,
            diff,Sco,Rglm,ncl,e,sor,Xb,oldbsb,ind,ord_0,
            ord_1,intr_0,intr_1,R_0,R_1,Rdif,ORD,ORDNCL;

  n=rows(X);
  p=cols(X);
  k=rows(bsb);

   temp=0;     @ specify the point you want to start the forward search @
   mo=rows(x)-K; @ mo= scalar. You can see the units belonging to nobsb
                 in the last mo steps of the forward search @

if freq==0;
   freq=ones(n,1);
 else;
endif;

seq=seqa(1,1,n);
   eq_0=selif(seq,y.==0);
   eq_1=selif(seq,y.==1);
   if binary==1;
      gap=rows(eq_1)-rows(eq_0);
   else;
      gap=rows(eq_1)/rows(eq_0);
   endif;

  R=seq~zeros(n,1); @ The second column of matrix R will contain the
                      GLM residuals at each step of the forward search @

  @ Diff = (n-p-temp) x nla matrix. It says how many new units join the basic
    subset in every step of the forward search. This index is called
    interchange index @
  Diff=zeros(n-p-temp,1);

  mIn=miss(zeros(mo,mo),0); @ matrix mIn contains the the units forming
                         the group of potential outliers in the last
                         mo steps of the forward procedure @


  @ Matrix BB will contain the beta coeff. in each step of the fwd search @
  BB=seqa(p,1,n-p+1)~miss(zeros(n-p+1,p),0);

  @ TT= (n-p) x (p+1) Matrix which will contain t tests for
    beta coeff. in each step of the fwd search @
  TT=miss(seqa(p+1,1,n-p)~zeros(n-p,p),0);

  @ Matrix PVT will contain pvalues for ttest of beta coeff. @
  PVT=TT;

  @ Lik= (n-p+1) x 4 Matrix
  1st col = fwd search index
  2nd col = deviance
  3rd col = pseudo R^2
  4th col = dispersion parameter  @
  Lik=seqa(p,1,n-p+1)~miss(zeros(n-p+1,3),0);


  @ Matrix RST will contain max deviance residual among BSB
    in each step of the fwd search @
  RST=seqa(p+1,1,n-p)~miss(zeros(n-p,1),0);

  @ Coo= (n-p) x 2 Matrix which contains Cook distance from old and new basic
    subset in each step of the fwd search  eq. 3.2.2 p. 24
    \beta_{(i)}= old basic subset
    \beta= new basic subset  @
  COO=seqa(p+1,1,n-p)~miss(zeros(n-p,1),0);
  @ Un = (n-p) x 10 matrix. In the second col. it contains the unit
    which entered the subset @
  UN=seqa(p+1,1,n-p)~miss(zeros(n-p,9),0);

  @ Sco= (n-p-1) x 2 Matrix which contains score test
    of goodness of link
     in each step of the fwd search  @
  Sco=seqa(p+2,1,n-p-1)~miss(zeros(n-p-1,1),0);

  @ Rglm =n x (n-k+1) matrix every row refers to a unit. Every column
    to a step in the fwd search. This matrix will contain residuals for
    every unit.
    REMARK: The kind of residuals used is specified in scalar search @
  Rglm=miss(zeros(n,n-k+1),0);

  @ LEV =n x (n-k+1) matrix every row refers to a unit. Every column
    to a step in the fwd search. This matrix will contain leverage values
    for the units belonging to BSB. @
  LEV=miss(zeros(n,n-k+1),0);

  @ WEI =n x (n-p+1) matrix every row refers to a unit. Every column
    to a step in the fwd search. This matrix will contain the weights for
    every unit used in IWLS estimation procedure  @
  WEI=miss(zeros(n,n-k+1),0);

   jk=0; @ this is a dummy index linked to the columns of matrix mIn @
   unit=0;


                               start:
                 mm=k+temp;
                  @ bsb contains the indexes of the units forming BSB @
                  mX=X[bsb,.]; @ initial basic subset for X @
                  vfreq=freq[bsb,.]; @ initial subset of frequency
                                        vector @

                   if rank(mX)<p;
                     "rank less than p";; wait;
                      temp=temp+1;
                      goto start;
                   else;
                   endif;

                  ncl=setdif(seq,bsb,1);
                  vY=y[bsb];   @ inital basic subset for y @

          mm=k+temp; do until mm==n+1;
          "mm";; mm;

              @ glm on basic subset @

                  {B,Rs,Hi,W,Rsq,cnt}=glm(mX,vY,q,vfreq,mint,tol);
                      if cnt==1;
                         "Attention convergence";
                         "NOT obtained when mm=";; mm;   wait;
                         goto star;
                      else;
                      endif;
                      @ Store beta coeff. @
                         BB[mm-p+1,2:p+1]=(B[.,1]');
                      @ Store dispersion parameter  @
                         Lik[mm-p+1,4]=Rsq[2,1];
                      @ Store R^2 @
                         Lik[mm-p+1,3]=Rsq[1,1];
                      @ Store deviance @
                         Lik[mm-p+1,2]=Rsq[1,2];

               @   COMPUTE RESIDUALS FOR THE OVERALL SAMPLE @
               @ res. for the overall sample
                 and beta based on units forming BSB @
             {RS}=predglm(X,y,q,freq,B[.,1]);
              e=RS[.,sh];
                r[.,2]=e.^2;    @ r_i =e_i^2 @

                  if mm>p+temp;

                      @ Store max glm residual among BSB @
                         sor=sortc(abs(e),1);
                         RST[mm-p,2]=sor[mm];

                      TT[mm-p,2:p+1]=(B[.,3]'); @ store t test  @
                      PVT[mm-p,2:p+1]=(B[.,4]'); @ store p values of t test  @
                      @ Compute Cook distance @
                      bib=BB[mm-p+1,2:p+1]-BB[mm-p,2:p+1];
                      Xb=mX./W;
                           Coo[mm-p,2]=bib*(Xb'Xb)*
                             (bib')/(p*Lik[mm-p+1,4]);
                  else;
                  endif;
                     if mm>p+temp+1;
                        {Tl}=scglm(mX,vY,q,vfreq,B[.,1],RSQ[2,1]);
                         @ STORE SCORE TEST @
                         Sco[mm-p-1,2]=Tl[1,3];
                       else;
                       endif;

         @ STORE LEVERAGE @
           LEV[bsb,mm-k+1]=Hi;

         @ Rglm= n x (nn-m+1) Matrix. STORE RESIDUALS @
           Rglm[.,mm-k+1]=   @ e @ r[.,2] ;

         @ WEI= n x (nn-m+1) Matrix. STORE WEIGHTS @
           WEI[bsb,mm-k+1]=w;


                    star:
         if mm<n;

              oldbsb=bsb;
             if (q==2 or q==3 or q==6) and (binary /=0 );
                      if stab==1;
                         ind=setdif(R[.,1],bsb,1);
                         R[bsb,2]=10^7*ones(rows(bsb),1);
                      else;
                      endif;
                   @ sort separately residuals whose response is = 0 , 1 @
                   @ ord_0=matrix l x 2
                           1st col =indexes of the units of univ. y=0
                           2nd col =residual of the units of univ. y=0 @
                      ord_0=sortc(R[eq_0,.],2);
                      ord_1=sortc(R[eq_1,.],2);
                   @  # of units of bsb whose response is = 0 , 1 @
                       intr_0=intrsect(eq_0,bsb,1);
                       intr_1=intrsect(eq_1,bsb,1);
                       r_0=rows(intr_0);
                       r_1=rows(intr_1);
                    @  "r0 and r1";;
                       r_0;; r_1;;   @
                     if binary==1;
                        rdif=r_1-r_0;
                     else;
                        rdif=r_1/r_0;
                     endif;
                      @  "rdif";   rdif~gap; wait;     @
                     if rdif >  gap;
                       @ insert inside the subset a unit from gr. y=0 @
                         if stab==1;
                            bsb=intr_0|ord_0[1,1]|intr_1;
                         else;
                            bsb=ord_0[1:r_0+1,1]|intr_1;
                         endif;
                     elseif rdif < gap;
                       @ insert inside the subset a unit from gr. y=1 @
                         if stab==1;
                            bsb=intr_0|intr_1|ord_1[1,1];
                         else;
                            bsb=intr_0|ord_1[1:r_1+1,1];
                         endif;
                     else;
                       @ insert the unit with the smallest residual @
                            if stab==1;
                              if ord_0[1,2]< ord_1[1,2];
                                  bsb=intr_0|ord_0[1,1]|intr_1;
                              else;
                                  bsb=intr_0|intr_1|ord_1[1,1];
                              endif;
                            else;
                              if ord_0[r_0+1,2]< ord_1[r_1+1,2];
                                  bsb=ord_0[1:r_0+1,1]|intr_1;
                              else;
                                  bsb=intr_0|ord_1[1:r_1+1,1];
                              endif;
                            endif;
                     endif;
             else;
                     if stab==1;
                         R[bsb,2]=10^15*ones(rows(bsb),1);
                         ord=sortc(R,2);
                         bsb=bsb|ord[1,1];
                      else;
                         ord=sortc(R,2);
                         bsb=ord[1:(mm+1),1];
                      endif;
             endif;
                     ncl=setdif(seq,bsb,1);
                     if mm>= n-mo and mm < n-1;
                       @ Include ncl in matrix mIn @
                        ordncl=sortc(R[ncl,.],2);
                        jk=jk+1;
                        mIn[1:rows(ordncl),jk]=ordncl[.,1];
                        else;
                     endif;

                   mX=X[bsb[.,1],.];  @ subset of X @
                   vY=Y[bsb[.,1]];    @ subset of untransformed Y @
                   vfreq=freq[bsb[.,1]]; @ subset of frequency vector @

                   Unit=setdif(bsb,oldbsb,1);
                   UN[mm-p+1,2:rows(unit)+1]=(unit');
                   diff[mm-p-temp+1,1]=rows(Unit);

          else;
         endif;

                mm=mm+1;
          endo;

   retp(BB,TT,Rglm,LEV,WEI,LIK,Rst,Coo,Sco,Un,diff,mIn);
endp;

