/*
**  FWDLDA
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {pos,pot,postF,In,Un,mea,LLN,ADpos,Adpot,eqcov,EIGE,VECT,mis}
**                            =fwdlda(bsb,sz,X,vsz,prior,dd);
**
**                                To identify influential
**                                observations in linear discriminant
**                                analysis.
**                                In every step of the fwd search
**                                a lot of quantities are monitored
**
*/

/*
** FWDLDA
**
** PURPOSE:  to monitor posterior probabilities in each step
**           of the forward search
**
** FORMAT: {pos,pot,postF,In,Un,mea,LLN,ADpos,Adpot,eqcov,EIGE,VECT,mis}
**                  =fwdlda(bsb,sz,X,vsz,prior,dd);
**
** INPUT:    bsb = mm x 1 vector containing the units
**                belonging to the initial subset
**            sz = g x 1 vector which contains
**                the dimension of the g subgroups forming
**                the initial basic subset
**                   e.g.: if sz =50|40|30
**                   then size of 1st gr. is 50
**                   then size of 2nd gr. is 40
**                   then size of 3rd gr. is 30
**            X = nn x p. Full matrix
**          vsz = g x 1 vector which contains sizes of the groups
**        prior = g x 1 vector containing prior probabilities
**          dd  = 2 x 1 vector specifying which kind of forward search
**                you want
**                if dd[1]=0 inclusion is based on MD without constraints
**                if dd[1]=1 inclusion is based on the difference
**                           among the sizes of the groups
**                  else     inclusion is based on the ratio
**                           among the sizes of the groups
**                element dd[2] is a suboption when dd[1] /= 0
**                if dd[2]=0  when you include a unit from
**                            universe j basic subset for
**                            universe k (k ne q) can change
**                 else       when you include a unit from
**                            universe j basic subset for
**                            universe k (k ne q) cannot change
** OUTPUT:  pos =  (nn-mm+1) x 3 matrix.  The first column of pos
**                 contains  the
**                dimension  of the sample used to compute Mahalanobis
**                distances Columns 2,3 respectively contain
**                Max and Median of Mahalanobis distances of the units
**                forming the basic subset
**          pot =  (nn-mm) x 2.The first column contains the
**                 dimension of the sample used in order to compute
**                 Mahalanobis distances, the second
**                 reports the minimum of Mahalanobis distances among
**                 the units which form the group of potential
**                 outliers
**        postF =  nn x (nn -mm+1) matrix of posterior probabilities in
**                 each step
**                 of the fwd search
**                 The first row is a forward search for the first unit
**           IN = (nn-mm) x 11 matrix.
**                1st col = forward search index
**                2nd col = unit with mm-th MD (max MD in bsb)
**                3rd col = unit with (mm+1)-th MD (min MD among NCL)
**                   ...
**                11th col = unit with (mm+9)-th MD
**                REMARK= the mm-th MD would be the mm_j-th
**                where mm_j is the part of basic subset referred to
**                universe j (j=1, ...,g) and mm= \sumc_{j=1}^g mm_j
**           Un = (nn-mm) x 11 Matrix which contains the unit(s)
**                included in bsb at each step of the fwd search
**                REMARK: in every step the new subset is compared
**                with the old subset. Un contains the unit present
**                in the new subset but not in the old one
**        ADpos = (nn-mm+1) x (3*g+1) matrix
**                It is equal to pos but is referred to each group
**                2nd 3rd and 4th cols contain Max bsb Median bsb and
**                     max MD for Univ1
**                5th 6th and 7th cols contain Max bsb Median bsb and
**                     max MD for Univ2
**                ...
**        ADpot = equal to pot but referred to each group
**                2nd col contain Min of MD for bsb1
**                3rd col contain Min of MD for bsb2
**                ...
**        eqcov = (nn-mm+1) x 2 matrix
**                test of equality among var cov matrices
**                in each step of the forward search
**        EIGE =  (nn-mm+1) x 2 matrix
**                2nd col contain fwd search index
**                3rd col contains the proportion explained by the
**                the first eigenvalue of the matrix inv(W)*B
**        VECT =  p x (nn-mm+1) matrix
**                the cols of this matrix contain the first eigenvector
**                in each step of the fwd search
**         mis =  (nn-mm+1) x (2*g+1) matrix
**                1st  col = forward search index
**                2nd  col = average prob of miscl. for group 1
**                3nd  col = average prob of miscl. for group 2
**                ...
**                (g+1)th  col= average prob. of miscl. for group g
**                (g+2)th  col = average prob of miscl. for cobsb_1
**                (g+3)th  col = average prob of miscl. for cobsb_2
**                ...
**                (2g+1)th col = average prob of miscl. for cobsb_g
**                where cobsb_j= complement of subset from univ. j
**                 (j=1,  ...g)
**
**       REMARK the coeff. of the eigenvectors are first
**              standardized as v'*W*w=1 and then modifed
**              v[i]=v[i]*sqrt(w_{i,i})
**
*/

 proc (13)=fwdlda(bsb,sz,X,vsz,prior,dd);

   local seq,seqp,va,ve,EIGE,VECT,W,B,ind,va1,neig,mis,spost,
       nn,n,mm,i,g,gg,pos,pot,zs,clean,min,mala,y,ii,j,zs1,
       post,seqT,postF,In,p,mea,Le1,LND,LLN,eqsz,me,md,te,
       a,ab,Z1,Z3,ab1,rTe,de,ord,Adpot,Adpos,Mon,eqcov,eqva,
       oldncl,ncl,unit,Un,nbsb,Z,Teall,sel,Tin,bsb_j;
g=rows(sz);    @ g= number of groups @
p=cols(X);
nn=rows(X);
seqT=seqa(1,1,nn);
seqp=seqa(1,1,p);
Z=X[bsb,.];
mm= rows(Z);  @ Dimension of initial basic subset @
min=mm;
oldncl=0; @ Initialize with a scalar the group of pot. outliers @

    @    The first column of pos contains the dimension
         of the sample used to compute Mahalanobis distances
         Columns 2,3 respectively contain
         Max and Average of Mahalanobis distances of the units
         forming the basic subset
         Column 4 contains Max of Mahalanobis distances of all
         the units  @
   pos=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,2);

   @     ADpos=(nn-mm+1) x (3*g+1) matrix
         1st col=forward search index
         2nd 3rd and 4th col = max median and average MD
                                for group 1 ... @
   ADpos=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,3*g);

   @    Mea = (nn-mm+1) x g*p matrix.
       The first column of pos contains the dimension
        of the sample used
        The first g columns contain the means of the p variables
        belonging to the first universe  ... @
   Mea=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,g*p);

   @  pot has two columns the first contains the dimension of the
       sample used in order to compute Mahalanobis distances, the second
       reports the minimum of Mahalanobis distances among the units
       which form the group of potential outliers @
   pot=seqa(mm,1,nn-mm)~zeros(nn-mm,1);

   @ ADpot = nn-mm x g matrix. first col=dimension index
       2nd col=min of MD of group 1 among potential outliers
       It is inizialized with missing values because same values in
       the last rows will be zero @
   ADpot=seqa(mm,1,nn-mm)~reshape(miss(0,0),nn-mm,g);


   @   eqcov = Matrix whose 2nd column contains the test of equality
       of the variance cov matrix
       at each step of the fwd search @
   eqcov=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,1);


   @   LLN= (nn-mm+1) x (g+1) matrix.
       1st col = index related to the steps of the forward search
       2nd col = ln(det(varcov)) for pooled var cov matrix @
   LLN=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,1);

   @   EIGE=(nn-mm+1) x 2 matrix
       1st col = index related to the steps of the forward search
       2nd col = proportion of var of inv(W)*B explained by the
                 the first eigenvalue @
   EIGE=LLN;


   @    In= Matrix whose 2nd column contains the unit in mm-th position
        at each step of the fwd search. The other cols refer to the
        units ranked (mm+1)-th (mm+10)-th in non dexreasing order @
   In=seqa(mm,1,nn-mm)~miss(zeros(nn-mm,10),0);

   @    Un= Matrix which contains the unit(s) included in
            bsb at each step of the fwd search @
   UN=seqa(mm+1,1,nn-mm)~miss(zeros(nn-mm,10),0);

   @ postf= nn x (nn-mm+1) matrix of posterior probabilities of
     belonging to its appropriate group in each step of the forward
     search @
   postf=zeros(nn,nn-mm+1);

   @ VECT=  p X (nn-mm+1) matrix it contains the values of the first
         standardized eigenvector in each step of the forward search @
   VECT=zeros(p,nn-mm+1);

   @ mis =  (nn-mm+1) x (g+1) matrix
            1st  col = forward search index
            2nd  col = average prob of miscl. for group 1
            3nd  col = average prob of miscl. for group 2
            ...
            (g+1)th  col= average prob. of miscl. for group g
            (g+2)th  col = average prob of miscl. for bsb_1
            (g+3)th  col = average prob of miscl. for bsb_2
            ...
            (2g+1)th col = average prob of micl. for bsb_g
            where bsb_j= subset from univ. j  (j=1, ...g) @
   mis=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,2*g);


  @ mala= nn x 2 matrix whose 2nd col contains MD of
       unit from its own centroid @
  mala=seqT~zeros(nn,1);

    eqsz=seqa(1,1,g)~sz;    @ For the iris data for example
                                 eqsz= 1  36
                                       2  37
                                       3  36             @
 /*****   Le1=eqsz[.,1]~zeros(g,1);
                                e.g. for the iris
                                       1   0
                                Le1=   2   0
                                       3   0
  *****/

          vsz=0|vsz;
          sz=0|sz;
 @         0           0
    vsz = 50    sz =  36
          50          37
          50          36       @


  do until mm==nn+1;  @ nn= numero tot. di unita' stat. @
   mm;
 @ Input of lda: Z = overall subset
                sz[2:g+1] = sizes of groups inside  subset Z
                 X = overall sample containing all units
               vsz[2:g+1] = sizes of the groups inside sample X
             prior = prior probabilities @
  {post,lnd,Me,MD,eqva}=lda(Z,sz[2:g+1],X,vsz[2:g+1],prior);

  {W,B}=vardec(Z,sz[2:g+1]);
  {va,ve}=eigv(invpd(W)*B);


  @ order eigenvalues and eigenvectors @
   ind=rev(sortc(seqp~va,2));

   va1=va[ind[.,1]]; @ vector va contains ordered eigenvalues @

   ve=ve[.,ind[1,1]]; @ vector ve contains first eigenvector @
   neig=minc(p|g-1);
   VECT[.,mm-min+1]=real(ve/sqrt(ve'W*ve/(rows(Z)-g)) .*
                    sqrt(diag(W)/(rows(Z)-g))   );

   EIGE[mm-min+1,2]=va1[1]/sumc(va1[1:neig]);

   @ Store in matrix eqcov the test of equality among var cov matrices @
   eqcov[mm-min+1,2]=eqva;


  @ MD= n x 1 vector of Mahalanobis distances of every unit from its
        appropriate population @
  @ mala = n x 2 matrix (first col = seq(1,1,n))  @
      mala[.,2]=MD;

    postf[.,mm-min+1]=post[.,1];  @ posterior probabilities @

    LLN[mm-min+1,2]=lnd;   @ log of det of var cov @

      j=1; do until j==g+1;
        @ mis= Matrix which contains misclassification prob.
               for each universe and each subset @
        spost=post[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),1];
        mis[mm-min+1,j+1]= @ prior[j]* @ meanc(1-spost);
        @ bsb_j = subset from Universe j @
        bsb_j=selif(bsb,bsb.> sumc(vsz[1:j])
                    .and bsb.<=sumc(vsz[1:j+1]));
   @ seqa(sumc(vsz[1:j])+1,1,vsz[j+1])';
    bsb_j';
    wait;  @
        bsb_j=setdif(seqa(sumc(vsz[1:j])+1,1,vsz[j+1]),bsb_j,1);
        @ bsb_j'; wait; @
         if ismiss(bsb_j)==1;
           mis[mm-min+1,g+j+1]= miss(0,0);
         else;
           mis[mm-min+1,g+j+1]= meanc(1-post[bsb_j,1]);
         endif;
        @ Mea=Matrix Mea will contain the means of the variables @
        Mea[mm-min+1,(j-1)*p+2:j*p+1]=Me[j,.];

        @ Mon has 2 cols.
           2nd  col contains sorted MD for universe j
           1st col contains the indexes of these sorted units @
        Mon=sortc(mala[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),.],2);

       @ cols of matrix ADpos contain max bsb median and max MD for
        every
        universe. cols 2 3 and 4 refer to 1st universe
                  cols 5 6 snd 7 refer to 2nd universe ...
                                  max bsb  median   global max @
       ADpos[mm-min+1,3*j-1:3*j+1]=Mon[sz[j+1],2]~Mon[trunc(sz[j+1]/2),2]
                                 ~maxc(Mon[.,2]);

                  if sz[j+1]<vsz[j+1];
                          @ ADpot contains minimum of Mahalanobis
                             distances among the units
                             which form the group of potential
                             outliers for each group
                             Minimun noclean  @
                         ADpot[mm-min+1,j+1]=Mon[sz[j+1]+1,2]; @
       ADpot[mm-min+1,j+1]=meanc(Mon[sz[j+1]+1:rows(Mon),2]);  @
                  else;
                  endif;
          j=j+1;
          endo;


      zs=sortc(mala,2);

    @ Matrix pos refers to all the units @
    @ pos has 3 columns: the first contains the dimension of clean data
      set, Columns 2, and 3 respectively contain Max, Average,
      forming the basic subset @
                    @   maximum       average  @
     pos[mm-min+1,2:3]=zs[mm,2]~meanc(zs[1:mm,2]);


 if mm<nn;
        @ pot contains minimum of Mahalanobis distances among the units
          which form the group of potential outliers @
        pot[mm-min+1,2]=zs[mm+1,2];

    @ if dd[1] /= 0 Include alternatively a unit from each group
          according to difference (if dd[1]=1)
          or according to ratio (if dd[1] ne 1) @
    if dd[1] /= 0;
       @ Matrix Teall contains sorted MD for each of the g populations @
       j=1; do until j==g+1;
         if j==1;
           Teall=sortc(mala[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),.],2);
         else;
           Teall=Teall|sortc(mala[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),.],2);
         endif;
         j=j+1;
       endo;
       @ Second col. of eqsz contains dimension of basic subsets @
       @ Vector de contains the index of the universes
         whose size must be increased @
      if dd[1]==1;
       a=vsz[2:g+1]-eqsz[.,2];
       de=delif(eqsz[.,1],a.<maxc(a));
      else;
       a=eqsz[.,2]./vsz[2:g+1];
       de=delif(eqsz[.,1],a.>minc(a));
      endif;
                               @ vsz[2:g+1]~eqsz[.,2];
                                de;  wait;      @


        Le1=de~zeros(rows(de),1); @ 1 \leq rows(Le1) \leq g
                                 where g is the number of the groups @
       @ Among the subsets which have comparable size insert the unit
        with the smallest Mahalanobis distance @
       i=1; do until i==rows(de)+1;
           j=de[i];
            @ second col. of matrix Le1 will contain MD which occupy
              position mm+1 in the groups of equal sizes.
              sz is a vector which contains the sizes of the BSBs
              REMARK: sz has a 0 in position [1] this is the reason
              why in the next line we write sz[j+1] @
           Le1[i,2]=Teall[sumc(vsz[1:j])+sz[j+1]+1,2];
           i=i+1;
       endo;
      ord=sortc(Le1,2);
      ab1=ord[1,1];  @ ab1 is the group with the smallest MD
                     among the groups of comparable size @


       eqsz[ab1,2]=eqsz[ab1,2]+1; @ dim. of basic subset for univ.
                                       ab1 is increased by one unit @

            @ Te is a matrix whose 2nd col contains MD for
              universe j ordered @
            Te=Teall[sumc(vsz[1:ab1])+1:sumc(vsz[1:ab1])+sz[ab1+1]+1,1];
            @ Tin=vector to be included in matrix In @
            Tin=Teall[sumc(vsz[1:ab1])+1:sumc(vsz[1:ab1+1]),1];
         if dd[2]==0;
           @ select all groups apart the ab1-th @
           sel=setdif(eqsz[.,1],ab1,1);
           nbsb=Te;
            i=1; do until i==g;
              j=sel[i];
              nbsb=nbsb|Teall[sumc(vsz[1:j])+1:sumc(vsz[1:j])+sz[j+1],1];
             i=i+1;
            endo;
         else;
             nbsb=delif(bsb,bsb.>= sumc(vsz[1:ab1])+1
                       .and bsb .<=sumc(vsz[1:ab1+1]));
             @ nbsb=nbsb|Te[1:sz[ab1+1]+1,1];  @
              nbsb=nbsb|Te;
          endif;
             unit=setdif(nbsb,bsb,1);
             if rows(unit)<=10;
              UN[mm-min+1,2:rows(unit)+1]=(unit');
             else;
              "Warning: interchange greater than 10";
              UN[mm-min+1,2:11]=unit[1:10]';
             endif;
             @ Store in matrix In unit ranked mm_j-th (mm_j+9)-th
             in terms of MD @
            @ Te=Te[.,1]|miss(zeros(10,1),0); @
            Tin=Tin|miss(zeros(10,1),0);
          @  In[mm-min+1,2:11]=Te[sz[ab1+1]:(sz[ab1+1]+9),1]';  @
            In[mm-min+1,2:11]=Tin[sz[ab1+1]:(sz[ab1+1]+9),1]';
            sz[2:g+1]=eqsz[.,2];   @ new sizes for new basic subset @

     else;
             zs=zs[.,1]|miss(zeros(10,1),0);
             nbsb=zs[1:mm+1,1];
              @ Create  sz[2:g+1]= vector which contains
              the new sizes of the groups  @
              j=1; do until j==g+1;
                sz[j+1]=rows(selif(nbsb,
                nbsb.>= sumc(vsz[1:j])+1 .and nbsb .<= sumc(vsz[1:j+1])));
               j=j+1;
              endo;

             @ Store in matrix Un the unit(s) included in this step @
             unit=setdif(nbsb,bsb,1);
             if rows(unit)<=10;
              UN[mm-min+1,2:rows(unit)+1]=(unit');
             else;
              "Warning: interchange greater than 10";
              UN[mm-min+1,2:11]=unit[1:10]';
             endif;
             @ Store in matrix In unit ranked mm-th (mm+9)-th @
             IN[mm-min+1,2:11]=zs[mm:mm+9]';
    endif;
    Z=X[sortc(nbsb,1),.];
    bsb=nbsb;
 else;
 endif;

   mm=mm+1;
 endo;
   retp(pos,pot,postF,In,UN,mea,LLN,ADpos,Adpot,eqcov,EIGE,VECT,mis);
endp;

