/*
**  FWDMLE
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {mL}=fwdmle(X,bsb,ch);              To create a matrix which
**                                     contains MLE of transformation
**                                     parameters in each step
**                                     of the forward search
**
*/

/*
**
** FWDMLE
**
** PURPOSE:  to create a matrix which contains Max. Lik. Est.
**           of transformation parameters at each step of
**           the forward search
**
**
** FORMAT:    {mL}=fwdmle(X,bsb,ch);
**
** INPUT:     X     = n x p matrix of the data
**           bsb    = m x 1 vector which contains the indexes of the
**                    units forming the initial subset
**           ch     = c x 2 vector
**                    1st column = indexes of the variables (VAR)
**                                 whose MLE of transformation
**                                 parameter is required
**                    2nd column = values of the transformation
**                                 parameters to be used for ordering MD
**                    REMARK: if ch=0 (scalar zero) maximization
**                             concerns all the variables and the fwd
**                             search is done using untranformed values
**                    EXAMPLE: if ch=(1~0)|
**                                   (4~0.5)
**                             Maximation concerns VAR 1 and 4
**                             Ordering of MD in each step of the fwd
**                             search uses log transformation for VAR 1
**                             and sqrt transformation for VAR 4
**
** OUTPUT:    mL    = (n-m+1)  x (p+1) matrix. In the first column
**                    it contains the dimension of the basic subset used
**                    in the forward search. From column 2 to (p+1) we
**                    can find max. lik. estimates of the transformation
**                    parameters in each step of the forward search
**
** REMARK:  the input vector f can be conveniently created by
**          program UNIBIV.G
**          The transformtion used in the maximization is 2.5*sin(\lambda)
**          The initial \lambda values are =1 (no transformation)
**
**
** See also: FLR,FWDMLER,FWDMLEQD,FWDMLELD,MULTOUT
**
** WARNING: two global variables named Xglobal and chglobal are created
**
**
*/

 clearg Xglobal,chglobal;


proc (1)=ass(la0);    /* procedure which transforms the parameters */
     local j,nla,la;
         nla=rows(la0);
         la=zeros(nla,1);
             @ j=1; do until j==nla+1;  @
                                  
                        la= @ 2.3*sin  ( @ la0 @ )  @ ;
                                       
                           @ la=2.5*la0./sqrt(1+la0^2); @
                        @ la[j]=  la0[j];   @
             @ j=j+1;
              endo; @
              @   la=la0;  @
    retp(la);
endp;


proc (1)=lik(la);   /* procedure which computes max. lik. */
     local j,nla,G,z,nXc,dZ,p,i;
             p=cols(Xglobal);
             nXc=rows(Xglobal);
             nla=rows(chglobal);
             Z=Xglobal;
             
                     
             {la}=ass(la);

                  G=exp(meanc(ln(Xglobal[.,chglobal])));
                   G=G'; @ Vector of geometric means @
                   @   just Z[.,chglobal] is tested @
             Z[.,chglobal]=(Xglobal[.,chglobal].^(la')-1
                              )./((la').*G.^((la')-1));
            @ Z[.,chglobal]=Z[.,chglobal]./(meanc(Z[.,chglobal])'); @                
       dZ=100*det(vcx(Z));
    @ dz; wait; @
   retp(dZ);
endp;

proc (1)=fwdmle(X,bsb,ch); /* procedure which computes fwd search */
       local nn,seq,  Plik, G, mm,la0,zs,mala,min,dZ,Xt,sel,j,la,
             psi,fvalue,grad,retcode,vc,i,ym,psin,p,prov,finale;
          nn=rows(X);
          p=cols(X);
          seq=SEQA(1,1,nn);
          mm= rows(bsb);
          min=mm;

   @ chglobal specifies the variables to analyze in the maximization @
       if ch==0 and rows(ch)==1;
           sel=seqa(1,1,p);
           la=ones(p,1);
       else;
           la=ch[.,2];
           sel=ch[.,1];
       endif;
           chglobal=sel;

          Plik=seqa(mm,1,nn-mm+1)~miss(zeros(nn-mm+1,p),0);

          /*  calculation of Mahal. dist. on initial clean data set */
           mala=seq~zeros(nn,1);
           la0= sin (ones(rows(chglobal),1));
           @ la0=la+1e-4; @
           
           @ la0= -5.1|-5.3|-0.7|0.01|-3.89|-3.38|-3.65; @
            @ 2*la+1e-2;  @

          @ create Xt=matrix of transformed values to be used
             in the forward search (for ordering MD) @
          Xt=X;

        j=1; do until j==rows(sel)+1;
                  G=exp(meanc(ln(X[.,sel[j]])));

           if la[j] ne 0;
                  Xt[.,sel[j]]= X[.,sel[j]]^la[j]/(G^(la[j]-1));
              else;
                  Xt[.,sel[j]]=ln(X[.,sel[j]]);
             endif;
           j=j+1;
        endo;

       mm=mm; do until mm==nn+1;  @ nn= numero tot. di unita' stat. @
       "mm=";; mm; 
               if mm>min;
               @ if abs(maxc(psi))>4; @
               
                  la0=psi;  
                  
               else;
                
               endif;

                  Xglobal=X[bsb,.];
                  @ Now rescale Xglobal for procedure OPTMUM @
                  Xglobal=Xglobal./(maxc(Xglobal)');
                  Xglobal=50*Xglobal;
                  
                  
                  {dZ}=lik(la0);

                  {psi,fvalue,grad,retcode}=OPTMUM(&lik,la0);
                  @   "DET RETTIFIED"; det(vcx(Xglobal));
                    "fvalue";;   fvalue; wait;  @
                    
                    
                   @ Retransform the MLE parameters @
                            {psin}=ass(psi);
                             
                 
                  if retcode ne 0 or maxc(abs(psin))>5;
                     "print retcode";; retcode; 
                     "DET RETTIFIED"; det(vcx(Xglobal));
                    "fvalue";;   fvalue; 
                    "mm";; mm;; @ wait;  @
                 goto finale;
                    "Warning convergence not obtained";
                    "Do not store anything";
                    "Vector of initial estimates in the next step
                    will be initialized with estimates before convergence";
                    psi=la0; 
                  else;
                     @ else store MLE of transformation parameters @                  
                  
                    Plik[mm-min+1, (chglobal)+1]=(psin');

                  
                  endif;
                  
                  @ output file =notconv  on;
                     "step";; mm; 
                  output file =notconv off;  @
                  
                  
                  
               ym=meanc(Xt[bsb,.]);
               vc=invpd(vcx(Xt[bsb,.]));

                            i=1;  do until i==nn+1;
                             mala[i,2]=sqrt((Xt[i,.]-ym')*vc*(Xt[i,.]'-ym));
                                 i=i+1;
                             endo;
                zs=sortc(mala,2);

                        @ the dimension of the clean data set increases
                          by one unit. Vector bsb contains the
                          indexes corresponding to the units of
                          the new basic subset @
                                if mm<nn;
                                   bsb=zs[1:mm+1,1];
                                 else;
                                 endif;
             mm=mm+1;
          endo;
          
          finale:
  retp(Plik);
endp;

