/*
**  FWDMLELD
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {mL}=fwdmleLD(bsb,sz,X,vsz,dd,ch,mis);
**                                     To create a matrix which
**                                     contains MLE of transformation
**                                     parameters at each step
**                                     of the forward search when
**                                     there are different groups
**                                     of units (We assume equality
**                                     among the different var cov
**                                     matrices among the groups)
**
*/

/*
**
** FWDMLEQD
**
** PURPOSE:  to create a matrix which contains Max. Lik. Est.
**           of transformation parameters at each step of
**           the forward search when there are different
**           groups of units. A pooled var cov matrix
**           is used
**
** FORMAT:    {mL}=fwdmleLD(bsb,sz,X,vsz,dd,ch,mis);
**
** INPUT:    bsb = mm x 1 vector containing the units
**                belonging to the initial subset
**            sz = g x 1 vector which contains
**                the dimension of the g subgroups forming
**                the initial basic subset
**                   e.g.: if sz =50|40|30
**                   then size of 1st gr. is 50
**                   then size of 2nd gr. is 40
**                   then size of 3rd gr. is 30
**            X = nn x p. Full matrix
**          vsz = g x 1 vector which contains sizes of the groups
**          dd  = 2 x 1 vector specifying which kind of forward search
**                you want
**                if dd[1]=0 inclusion is based on MD without constraints
**                if dd[1]=1 inclusion is based on the difference
**                           among the sizes of the groups
**                  else     inclusion is based on the ratio
**                           among the sizes of the groups
**                element dd[2] is a suboption when dd[1] /= 0
**                if dd[2]=0  when you include a unit from
**                            universe j basic subset for
**                            universe k (k ne q) can change
**                 else       when you include a unit from
**                            universe j basic subset for
**                            universe k (k ne q) cannot change
**          ch  = c x 2 vector
**                1st column = indexes of the variables (VAR)
**                             whose MLE of transformation
**                             parameter is required
**                2nd column = values of the transformation
**                             parameters to be used for ordering MD
**                REMARK: if ch=0 (scalar zero) maximization
**                         concerns all the variables and the fwd
**                         search is done using untranformed values
**                EXAMPLE: if ch=(1~0)|
**                               (4~0.5)
**                         Maximation concerns VAR 1 and 4
**                         Ordering of MD in each step of the fwd
**                         search uses log transformation for VAR 1
**                         and sqrt transformation for VAR 4
**          mis = scalar.
**                If mis=1 we search the transformation parameters
**                which minimize the overall probability of
**                misallocation
**                else  we search the transformation parameters
**                which maximize the likelihood.
**
** OUTPUT:    mL    = (nn-mm+1)  x (p+1) matrix. In the first column
**                    it contains the dimension of the basic subset used
**                    in the forard search. From column 2 to (p+1) we
**                    can find max. lik. estimates of the transformation
**                    parameters in each step of the forward search
**
**
** REMARK:  the input vector f can be conveniently created by
**          program UNIBIV.G
**          The transformtion used in the maximization is 2.5*sin(\lambda)
**          The initial \lambda values are =1 (no transformation)
**
** WARNING: 3 global variables named xglobal, chglobal and
**          szglobal are created
**
*/

 clearg Xglobal,chglobal,szglobal,kglobal;

proc (1)=assLD(la0);    /* procedure which transforms the parameters */
     local j,nla,la,k;
         nla=rows(la0);
         la=zeros(nla,1);
          @  if abs(la0)>1e+10;
              k=k+1;
            endif;
             if kglobal==1;   @
                    j=1; do until j==nla+1;
                                 la[j]=  4*sin(  la0[j]  )  ;
                    j=j+1;
                    endo;
          @   endif;  @
    retp(la);
endp;


proc (1)=likLD(la);   @ procedure which computes max. lik. @
     local j,nla,G,z,nXc,dZ,p,i,sz;
             p=cols(Xglobal);
             nXc=rows(Xglobal);
             nla=rows(chglobal);
             Z=Xglobal;  sz=szglobal;
             dZ=0; @ Initialize dZ @
             {la}=assLD(la);

                  G=exp(meanc(ln(Xglobal[.,chglobal])));
                   G=G'; @ Vector of geometric means @
          Z=(Z-1);  @ just Z[.,chglobal] is tested @
 Z[.,chglobal]=(Xglobal[.,chglobal].^(la')-1)./((la').*G.^((la')-1));
    i=1; do until i==rows(sz);
        dZ= dZ+
        sz[i+1]*(vcxm(Z[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]));
          i=i+1;
     endo;
       dZ=100*det(dZ);
   retp(dZ);
endp;


@ miscl= proc. which computes the overall prob. of misclassification @
proc (1)=misclaLD(la);
     local j,nla,G,z,nXc,dZ,p,i,sz,post,spost,gr;
             p=cols(Xglobal);
             nXc=rows(Xglobal);
             nla=rows(chglobal);
             Z=Xglobal;  sz=szglobal;
             gr=rows(sz)-1;
             dZ=0; @ Initialize dZ @
             {la}=assLD(la);

                  G=exp(meanc(ln(Xglobal[.,chglobal])));
                   G=G'; @ Vector of geometric means @
          Z=(Z-1);  @ just Z[.,chglobal] is tested @
  Z[.,chglobal]=(Xglobal[.,chglobal].^(la')-1)./((la').*G.^((la')-1));

  {post}=ldasimpl(Z,sz[2:gr+1],ones(gr,1)/gr);
    i=1; do until i==gr+1;
        @ mis matrix which contains the overall prob. of misclass.
          for the various groups @
         spost=post[sumc(sz[1:i])+1:sumc(sz[1:i+1]),1];
         dZ= dZ+ sumc(1-spost);
          i=i+1;
     endo;
       dZ= 100*dZ;
   retp(dZ);
endp;


proc (1)=fwdmleLD(bsb,sz,X,vsz,dd,ch,mis);
       local nn,seq,Plik,mm,la0,zs,mala,min,dZ,
             psi,fvalue,grad,retcode,i,ym,psin,p,geom,prov,
             g,eqsz,Le1,m,Z,vZ,iVZ,xr,MD,k,a,de,Ident,j,Te,
             ord,ab1,nbsb,id,Teall,sel,Xt,la;
          nn=rows(X);
          p=cols(X);
          seq=SEQA(1,1,nn);
          mm= rows(bsb);
          min=mm;

g=rows(sz);    @ g= number of groups @
eqsz=seqa(1,1,g)~sz;    @ For the iris data for example
                                 eqsz= 1  37
                                       2  39
                                       3  43
                        Later matrix eqsz will bew changed @

    @  Matrix M contains (in the rows) the vector of the means for
       the groups
       first row refers to first group
       ...
       last row refers to the last group  @
    M=zeros(g,p);

    @ Ident = n x 1 vector which contains the labels of the units
      referred to the different groups @
           id=0;
           i=1; do until i==g+1;
           id=id|reshape(i,vsz[i],1);
             i=i+1;
           endo;
          Ident=id[2:nn+1];

   @ chglobal specifies the variables to analyze in the maximization @
       if rows(ch)==1 and ch==0;
           sel=seqa(1,1,p);
           la=ones(p,1);
       else;
           la=ch[.,2];
           sel=ch[.,1];
       endif;
           chglobal=sel;

          Plik=seqa(mm,1,nn-mm+1)~ones(nn-mm+1,p);

       @ Divide each each column by its standard deviation @
     @    X=X./(stdc(X)'); @

          @ Create Xt=matrix of transformed values to be used
             in the forward search (for ordering MD) @
          Xt=X;
        j=1; do until j==rows(sel)+1;
           if la[j] NE 0;
                  Xt[.,sel[j]]= X[.,sel[j]]^la[j];
              else;
                  Xt[.,sel[j]]=ln(X[.,sel[j]]);
             endif;
           j=j+1;
        endo;


    @ Mala= nn x 2 matrix whose 2nd col contains MD of
      unit from its own centroid @
     Mala=seq~zeros(nn,1);
     la0=sin(ones(rows(chglobal),1));

    Le1=eqsz[.,1]~zeros(g,1);
     @ e.g. for the iris
           1   0
    Le1=   2   0
           3   0   @

          vsz=0|vsz;
          sz=0|sz;
 @        0                  0
    vsz = 50           sz=   36
          50                 37
          50                 36   @

       mm=mm; do until mm==nn+1;  @ nn= numero tot. di unita' stat. @
 @ "fwd";;    vcx(X[bsb,.]); wait;   @
       if mm>min;
          la0=psi;
       else;
       endif;

       @ la0=psiglobal;    @

                  szglobal=sz; @ sz= vector containing sizes of groups @
                  Xglobal=X[bsb,.];
                  Z=Xt[bsb,.];

                  if mis==1;
                    {dZ}=misclaLD(la0);
                    {psi,fvalue,grad,retcode}=OPTMUM(&misclaLD,la0);
                  else;
                    {dZ}=likLD(la0);
                    {psi,fvalue,grad,retcode}=OPTMUM(&likLD,la0);
                  endif;
               psiglobal=psi;
                    @ Retransform the (MLE) parameters @
                            {psin}=assLD(psi);

                    Plik[mm-min+1, (chglobal)+1]=(psin');
    @ Compute mean, pooled varcov, and
       (inv(varcov)) for each population  @
        Vz=0;
    i=1; do until i==g+1;
      M[i,.]=(meanc(Z[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]))';
      VZ= Vz + vcxm(Z[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]) * sz[i+1] ;
       @  VZ; wait;    @
          i=i+1;
     endo;
      VZ=VZ/(mm-g);  @ Pooled var cov matrix @
      iVZ=invpd(VZ);

                  @ Compute Vector of MD from each appropriate
                    population @

      i=1;  do until i==nn+1;
                      @ MD of each unit from every population @
         Xr=(Xt[i,.]-M[ident[i],.]);
         Mala[i,2]=sqrt(Xr*iVz*(Xr'));
            i=i+1;
      endo;
      @  "vcxmala";   vcx(Mala); wait;    @
        zs=sortc(Mala,2);

  if mm<nn;
    @ if dd[1] /= 0 Include alternatively a unit from each group
          according to difference (if dd[1]=1)
          or according to ratio (if dd[1] ne 1) @
    if dd[1] /= 0;
       @ Matrix Teall contains sorted MD for each of the g populations @
       j=1; do until j==g+1;
            @ Te is a matrix whose 2nd col contains MD for
              universe j ordered @
         if j==1;
           Teall=sortc(mala[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),.],2);
         else;
           Teall=Teall|sortc(mala[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),.],2);
         endif;
         j=j+1;
       endo;
       @ Second col. of eqsz contains dimension of basic subsets @
       @ Vector de contains the index of the universes
         whose size must be increased @
      if dd[1]==1;
       a=vsz[2:g+1]-eqsz[.,2];
       de=delif(eqsz[.,1],a.<maxc(a));
      else;
       a=eqsz[.,2]./vsz[2:g+1];
       de=delif(eqsz[.,1],a.>minc(a));
      endif;
                               @ vsz[2:g+1]~eqsz[.,2];
                                de;  wait;      @


        Le1=de~zeros(rows(de),1); @ 1 \leq rows(Le1) \leq g
                                 where g is the number of the groups @
       @ Among the subsets which have comparable size insert the unit
        with the smallest Mahalanobis distance @
       i=1; do until i==rows(de)+1;
           j=de[i];
            @ Te is a matrix whose 2nd col contains MD for
              universe j ordered @
      @ Te=sortc(mala[sumc(vsz[1:j])+1:sumc(vsz[1:j+1]),.],2); @
            @ second col. of matrix Le1 will contain MD which occupy
              position mm+1 in the groups of equal sizes.
              sz is a vector which contains the sizes of the BSBs
              REMARK: sz has a 0 in position [1] this is the reason
              why in the next line we write sz[j+1] @
           Le1[i,2]=Teall[sumc(vsz[1:j])+sz[j+1]+1,2];
           i=i+1;
       endo;
      ord=sortc(Le1,2);
      ab1=ord[1,1];  @ ab1 is the group with the smallest MD
                     among the groups of comparable size @


       eqsz[ab1,2]=eqsz[ab1,2]+1; @ dim. of basic subset for univ.
                                       ab1 is increased by one unit @

             @ Store in matrix Un the unit(s) included in this step @
     Te=Teall[sumc(vsz[1:ab1])+1:sumc(vsz[1:ab1])+sz[ab1+1]+1,1];
         if dd[2]==0;
           @ select all groups apart the ab1-th @
           sel=setdif(eqsz[.,1],ab1,1);
           nbsb=Te;
            i=1; do until i==g;
              j=sel[i];
              nbsb=nbsb|Teall[sumc(vsz[1:j])+1:sumc(vsz[1:j])+sz[j+1],1];
             i=i+1;
            endo;
         else;
             nbsb=delif(bsb,bsb.>= sumc(vsz[1:ab1])+1
                       .and bsb .<=sumc(vsz[1:ab1+1]));
             @ nbsb=nbsb|Te[1:sz[ab1+1]+1,1];  @
              nbsb=nbsb|Te;
          endif;
           @  Te=Te[.,1]|miss(zeros(10,1),0);    @
            sz[2:g+1]=eqsz[.,2];   @ new sizes for new basic subset @

    else;
             zs=zs[.,1]|miss(zeros(10,1),0);
             nbsb=zs[1:mm+1,1];
              @ Create  sz[2:g+1]= vector which contains
              the new sizes of the groups  @
              j=1; do until j==g+1;
                sz[j+1]=rows(selif(nbsb,
                nbsb.>= sumc(vsz[1:j])+1 .and nbsb .<= sumc(vsz[1:j+1])));
               j=j+1;
              endo;
    endif;
       @ format 9,7;        vcx(bsb); vcx(nbsb); wait;  @
       @  "setdif";;    setdif(nbsb,bsb,1);  wait;  @
    @ Z=X[sortc(nbsb,1),.];     @
    bsb=sortc(nbsb,1);
   else;
 endif;
                    mm=mm+1;
          endo;
  retp(Plik);
endp;

