/*
**  FWDOLS
**  written by MARCO RIANI
**
**  FORMAT                                    PURPOSE               Line
** =======================================================================
**
**  {BB,TT,pvt,as,ku,nor,ram,lik,sco,s2,rst,Coo,dres,
**        cc,rres,un,tout,covr,fr,lev,mIn,diff,start}
**                 = fwdols(X,y,la,Rou,q,sh,sub);
**
**                                            To perform a fwd search
**                                            for linear regression models
**
**
*/


/*
**
** FWDOLS
**
** PURPOSE:  to carry out a fwd search for
**           linear regression models
**
**
**
** FORMAT:
**
**  {BB,TT,pvt,as,ku,nor,ram,lik,sco,s2,rst,Coo,dres,
**        cc,rres,un,tout,covr,fr,lev,mIn,diff,start}
**                 = fwdols(X,y,la,Rou,q,sh,sub);
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             la = scalar which defines the
**                  transformation parameter \lambda
**                  for the response vector
**            Rou = Rou is a scalar which speficifies how to start
**                  the fwd search
**                  If Rou=1 the best subset is extracted according to
**                  lms else we choose other robust methods specified in scalar q @
**              q = scalar which specifies the kind of residuals
**                     if Rou=0 and q=1 we use least abs. dev.
**                     if Rou=0 and q=2 we use lts [n/2]+1 res.
**                     if Rou=0 and q=3 we use lts [n/2]+[(p+1)/2] res.
**                     if Rou=0 and q ne 1,2 or 3 regr. coeff. are
**                        solved using an iterative weighted regression
**             sh = scalar which specifies the ordering
**                  of the residuals in the fwd search
**                  if sh=1      r_i^2= e_i^2             if i \in  BSB
**                               r_i^2= e_(i)^2/(1+h_(i)) if i \in  NCL
**                  if sh=2      r_i^2= e_i^2/(1-h_i)      if i \in  BSB
**                               r_i^2= e_(i)^2/(1+h_(i))  if i \in  NCL
**                  else         r_i^2= e_i^2     if i \in  BSB
**                               r_i^2= e_(i)^2   if i \in  NCL
**              sub = scalar which specifies the number or sumsamples
**                    REMARK: is sub=1 all subsamples are extracted
**            start = specifies the starting point for fwd search
**                    fwd search start with an initial subset
**                    size equal to p+start
**
** OUTPUT:  BB = n-p+1 x p matrix
**               beta coefficients in each step of the fwd search
**          tt = n-p x p matrix
**               t test in each step of the fwd seach
**         pvt = n-p x p p-values for ttest of beta coeff
**         as  = n-p x 3 matrix which contains asymmetry test
**               and p values in each step of the fwd search
**         ku  = n-p x 3 matrix which contains kurtosis test
**               and p values in each step of the fwd search
**         nor = n-p x 3 matrix which contains normality test
**               and p values in each step of the fwd search
**         ram = n-p x 3 matrix which contains Ramsey test
**               and p values in each step of the fwd search
**         lik = n-p x 2 matrix which contains likelihood
**               in each step of the fwd search
**         sco = n-p-1 x 2 matrix which contains score test
**               in each step of the fwd search
**         S2  = n-p+1 x 3 matrix which
**               1st col = fwd search index
**               2nd col = var. of residuals
**                         = \sum e_i^2 /(m-p)
**               3rd col = r^2
**         RST = n-p x 2 matrix which contains max standardized
**               residual among BSB in each step of the fwd search
**         Coo = n-p x 2 matrix which contains
**               distance from old and new basic
**               subset in each step of the fwd search  eq. 3.2.2 p. 24
**               \beta_{(i)}= old basic subset
**               \beta= new basic subset
**        DRES = (n-p-1) x 2 matrix which will contain min deletion
**               residual among nobsb  r_i^*  eq. 3.1.1
**         CC  = (n-p-1) x g+1 matrix which will contain modified
**               deletion residual
**               C_i= \sqrt( \{ (n-p)/p hi/(1-hi) \})|r_i^*|
**               for the units which have been included into the basic
**               subset
**               Remark: It starts with p+2 because it uses s_{(i)} @
**        RRES= (n-p) x 4 matrix which will contain recursive
**               residuals  v_t=(y_k -x_k' \beta_{k-1})/sqrt(1+hi)
**               hi=x_k'* inv(X_{k-1}'X_{k-1}) x_k   k=p+1, ..., n
**               REMARK: Rres has more than two cols just in case that
**               at a certain step the new basic subset differs by
**               more than one unit
**         UN  = n-p x 13
**               1st col = fwd search index
**               2nd-12th = unit(s) included
**        TOUT = (n-p-1) x 2 matrix which contains an F test that the
**               units which belong to ncl are outliers. It is done in
**               every step of the forward search
**               eq. 5.5.7 with \sigma^2 replaced by s^2_{(I)} @
**        COVR =  (n-p-1) x 2 matrix which contains the log of the cov
**                ratio
**                cov ratio= (s^{2p}/|X'X|)^(1/2)   /
**                (s_(I)^{2p}/|X_(I)'X_(I)|)^(1/2)  p.227
**         FR  =  n x (n-p+1) matrix every row refers to a unit. Every
**                column to a step in the fwd search. This matrix will
**                contain residuals for
**                every unit.
**                REMARK: The kind of residuals used is specified in
**                scalar sh
**         LEV =  n x (n-p+1) matrix every row refers to a unit. Every
**                column to a step in the fwd search. This matrix will
**                contain leverage values  for the units belonging to
**                BSB.
**         mIn =  matrix which contains the units forming ncl
**                in each step
**        diff =  matrix with 4 columns 
**                1st col = fwd search index
**                2nd col = interchange index 
**                3rd col = proportion of units with positive
**                          residual inside subset 
**                4th col = proportion of units with positive
**                          residual in the sample
**                5th col = Difference between positive
**                          and negative residuals inside subset 
**                6th col = Difference between positive
**                          and negative residuals in the sample
**
**
**  REMARK1    this program does not add a column of 1 in matrix X to
**             create the intercept.
**
*/

proc   (22) =  fwdols(X,y,la,Rou,q,sh,sub,start);
   local  as,ku,nor,ram,s2,sco,dres,
          cc,rres,un,tout,covr,fr,lev,d,
          r,n,p,i,j,tp,mIn,BB,TT,PVT,LIK,RST,Coo,
          diff,mt,seq,HH,IH,mE,lm,b1,r2,rs,jk,unit,bs,me1,mm,
          bsb,mX,ncl,vY,vYt,Yt,EE,mmX,oldmmX,mAm,b,sres,rsq,NO,
          RA,sc,Hi,sor,bib,LI,e,k,ord,oldbsb,pr,se,Unitout,posress,
          posresU,selDres;
  n=rows(X);
  p=cols(X);
  tp=start;
d=0; @ d=0 implies not a time series @
pr=0; @ no print output from procedure regressi @
seq=seqa(1,1,n);
R=seq~zeros(n,1); @ The second column of matrix R will contain the
                    OLS residuals at each step of the forward search   @
  @ Lik= (n-p) x 2 matrix. LIk.  @
    Lik=seqa(p+1,1,n-p)~zeros(n-p,1);
  @ Sco = (n-p-1) x (nla+1) matrix. Score test @
    Sco=seqa(p+2,1,n-p-1)~miss(zeros(n-p-1,1),0);

  @ Diff = (n-p-tp) x 4 matrix. It says how many new units join the basic
    subset in every step of the forward search. This index is called
    interchange index @
  Diff=seqa(p+tp,1,n-p-tp+1)~zeros(n-p-tp+1,5);

  mIn=seqa(p+1,1,n-p)~miss(zeros(n-p,8),0); @ units removed in each step of the fwd search 
                         @

     @ IH=Projection matrix for the overall sample IH=(I-H) @
           mT=invpd(X'X);
           HH=X*mT*X';
           IH=diagrv(-HH,(ones(n,1)-diag(HH)));

 if Rou==1;
     {mE,lm,BS}=lms(X,y,la,sub);
       @ "lm";; lm;   @
  else;
     {mE,b1}=rob(X,y,la,q);
  endif;

  @ Matrix BB will contain the beta coeff. in each step of the fwd search @
  BB=seqa(p,1,n-p+1)~zeros(n-p+1,p);

  @ TT= (n-p) x (p+1) Matrix which will contain t tests for
    beta coeff. in each step of the fwd search @
  TT=miss(seqa(p+1,1,n-p)~zeros(n-p,p),0);

  @ Matrix PVT will contain pvalues for ttest of beta coeff. @
  PVT=TT;

  @ AS= (n-p) x 3 Matrix which will contain asymmetry test and p values
    in each step of the fwd search @
  AS=seqa(p+1,1,n-p)~zeros(n-p,2);

  @ KU= (n-p) x 3 Matrix which will contain kurtosis test and p values
    in each step of the fwd search @
  KU=AS;

  @ NOR= (n-p) x 3 Matrix which will contain normality test and p values
    in each step of the fwd search @
  NOR=AS;

  @ RAM= (n-p) x 3 Matrix which will contain Ramsey test and p values
    in each step of the fwd search @
  RAM=AS;

  @ S2=(n-p+1) x 3  matrix which will contain
    1st col = fwd search index
    2nd col = S2= \sum e_i^2 / (n-p)
    3rd col = R^2 @
  S2=seqa(p,1,n-p+1)~zeros(n-p+1,2);

  @ Matrix RST will contain max standardized residual among BSB
    in each step of the fwd search @
  RST=seqa(p+1,1,n-p)~miss(zeros(n-p,1),0);

  @ Coo= (n-p) x 2 Matrix which contains Cook distance from old and new basic
    subset in each step of the fwd search  eq. 3.2.2 p. 24
    \beta_{(i)}= old basic subset
    \beta= new basic subset  @
  COO=seqa(p+1,1,n-p)~miss(zeros(n-p,1),0);

  @ DRES= (n-p-1) x 2 matrix which will contain min deletion residual
    among nobsb  r_i^*  eq. 3.1.1 @
  DRES=seqa(p+1,1,n-p-1)~miss(zeros(n-p-1,1),0);

  @ CC= (n-p-1) x g+1 matrix which will contain modified
     deletion residual  C_i= \sqrt( \{ (n-p)/p hi/(1-hi) \} ) |r_i^*|
     for the units which have been included into the basic subset
     Remark: It starts with p+2 because it uses s_{(i)} @
  CC=MISS(seqa(p+2,1,n-p-1)~zeros(n-p-1,10),0);

  @ RRES= (n-p-1) x 4 matrix which will contain recursive
     residuals  v_t=(y_k -x_k' \beta_{k-1})/sqrt(1+hi)
     hi=x_k'* inv(X_{k-1}'X_{k-1}) x_k   k=p+1, ..., n
     REMARK: Rres has more than two cols just in case that at a certain
         step the new basic subset differs by more than one unit @
  RRES=seqa(p+1,1,n-p)~zeros(n-p,12);

  UN=seqa(p+1,1,n-p)~miss(zeros(n-p,12),0);


  @ Tout= (n-p-1) x 2 matrix which will contain an F test that the
    units which belong to ncl are outliers. It is done in every step
    of the forward search
    eq. 5.5.7 with \sigma^2 replaced by s^2_{(I)} @
  Tout=seqa(p+1,1,n-p-1)~miss(zeros(n-p-1,1),0);

  @ CovR= (n-p-1) x 2 matrix which will contain the log of the cov ratio
  cov ratio= (s^{2p}/|X'X|)^(1/2)   /  (s_(I)^{2p}/|X_(I)'X_(I)|)^(1/2)
  It is done in every step of the forward search
    pag. 227  @
  CovR=seqa(p+1,1,n-p-1)~miss(zeros(n-p-1,1),0);

  @ FR =n x (n-p+1) matrix every row refers to a unit. Every column
    to a step in the fwd search. This matrix will contain residuals for
    every unit.
    REMARK: The kind of residuals used is specified in scalar search @
  FR=zeros(n,n-p+1);

  @ LEV =n x (n-p+1) matrix every row refers to a unit. Every column
    to a step in the fwd search. This matrix will contain leverage values
    for the units belonging to BSB. @
  LEV=miss(zeros(n,n-p+1),0);


   jk=0; @ this is a dummy index linked to the columns of matrix mIn @
   unit=0;



            mE1=sortc(seq~abs(mE[.,1]),2);
                               start:
                 mm=p+tp;
                  @ bsb contains the indexes of the units forming BSB @
                      if Rou==1;
                         Bsb=BS;
                      else;
                         bsb=ME1[1:mm,1];
                      endif;
                      @ bsbglobal=bsb;  @
 
                       mX=X[bsb,.]; @ initial basic subset for X @
                       
                      @ "initial mX inside fwdols";     mX; wait;  @ 
              
                   if rank(mX)<p;
                     "rank less than p";; wait;
                      tp=tp+1;
                      goto start;
                   else;
                   endif;
                   ncl=setdif(seq,bsb,1);
                @  ncl=ME1[mm+1:n,1];   @
                  vY=y[bsb];   @ inital basic subset for y @
                  @ Transform all the units of the response var. @
                     if la==0;
                         vYt=ln(vY);
                         Yt=ln(Y);
                     else;
                         vYt=vY^(la);      @ transform the subset @
                         Yt=(Y^la);        @ transform all the units @
                     endif;

     @ EE=n x 1 vector containing residuals for the overall sample
           using transformed response and beta based on all the
           observations @
           EE=IH*Yt;
     oldmmX=0;  @ initialize oldmmX. This quantity will be used in the
                  computation of recursive residuals @
             @  forward search @
          mm=p+tp; do until mm==n+1;
         @ "mm";; mm;  @
                    mAm=mX'mX;
                    mmX=invpd(mAm);

              @ ols on basic subset using transf. response @
                  {B,sres,Rsq,NO,RA,SC}=regressi(vYt,mX,d,pr);

                  BB[mm-p+1,2:p+1]=(B[.,1]'); @ store beta coeff. @
                      @store res. sum of squares/(mm-k)   @
                      S2[mm-p+1,2]=Rsq[1,2];
                      S2[mm-p+1,3]=Rsq[1,1];    @ Store R^2 @

                  if mm>p+tp;
                      AS[mm-p,2:3]=NO[1,1:2];  @ Store Asimmetry test @
                      KU[mm-p,2:3]=NO[2,1:2];  @ Store Kurtosis test  @
                      NOR[mm-p,2:3]=NO[3,1:2]; @ Store Normality test @
                      RAM[mm-p,2:3]=RA[1,.]; @ Store Ramsey test @
                      @ Store recursive residuals @

                      i=1; do until i==rows(unit)+1;
                           @ hi= x_i' inv(X'_(i)X_(i)) x_i @
                           hi=X[unit[i],.]*oldmmX*(X[unit[i],.]');
                           RRES[mm-p,i+1]=
                       (Yt[unit[i]]-X[unit[i],.]*(BB[mm-p,2:p+1]'))/sqrt(1+hi);
                        i=i+1;
                      endo;


                      @ Store max standardized residual among BSB @
                       if ismiss(sres) ne 1;
                         sor=sortc(abs(sres),1);
                         RST[mm-p,2]=sor[mm];
                       else;
                       endif;

                      TT[mm-p,2:p+1]=(B[.,3]'); @ store t test  @
                      PVT[mm-p,2:p+1]=(B[.,4]'); @ store p values of t test  @
                      @ Compute Cook distance @
                      bib=BB[mm-p+1,2:p+1]-BB[mm-p,2:p+1];
                       if S2[mm-p+1,2]>0;
                           Coo[mm-p,2]=bib*mAm*(bib')/(p*S2[mm-p+1,2]);
                       else;
                       endif;
                       @ Score test and lik. on original data @
                       {Sc,Li}=likla(mX,vY,la);
                       Lik[mm-p,2]=Li;
                 else;
                 endif;


            @ res. for the overall sample using transf. response
              and beta based on units forming BSB @
            e=Yt-X*B[.,1];

            r[.,2]=e.^2;    @ r_i =e_i^2 @

       @  "Unit";;  unit; wait;   @
          if mm>p+tp+1 and S2[mm-p,2]>0;
              @ Score test coming out from procedure likla @
              Sco[mm-p-1,2]=Sc;

            @ Construct modified Cook dist. for the units included in
              bsb @
              i=1; do until i==rows(unit)+1;
               @ hi= x_i' inv(X'X) x_i @
               hi=X[unit[i],.]*mmX*(X[unit[i],.]');
               CC[mm-p-1,i+1]= 1/(1-hi)* sqrt((mm-p)/p *hi *
                               r[unit[i],2]/S2[mm-p,2]);
                   i=i+1;
                endo;
              else;
          endif;

                         @ Units inside the basic subset are rettified @
                          k=1; do until k==rows(bsb)+1;
                              hi=X[bsb[k],.]*mmX*(X[bsb[k],.]');
                              LEV[bsb[k],mm-p+1]= @ (mm/p)* @ hi;
                                 if sh==2;
                                    if hi==1;
                                    hi=0.99999999999;
                                    else; endif;
                                   r[bsb[k],2]=r[bsb[k],2]/(1-hi);
                                  else;
                                 endif;
                              k=k+1;
                          endo;


         if mm<n;
                        if sh==1 or sh==2;
                          k=1; do until k==rows(ncl)+1;
                                @  if a unit is not in bsb then
                               e_i^2 = e_i^2 /(1+h_{(i)}
                               where
                               h_{(i)}=x_i' inv(X'_{(i)} X_{(i)}) x_i @
                            hi=X[ncl[k],.]*mmX*(X[ncl[k],.]');
                            r[ncl[k],2]=r[ncl[k],2]/(1+hi);
                            k=k+1;
                           endo;
                         else;
                         endif;


                   @ initialize ord @
                   ord=10000~0;
            if mm>p+tp;
                 k=1; do until k==rows(ncl)+1;
                   @ Take minimum deletion residual among noBSB @
                     hi=X[ncl[k],.]*mmX*(X[ncl[k],.]');
                   @ ord= vector which contains deletion residuals @
                   @ "hi";; hi; wait; @
                     ord=ord|(r[ncl[k],2]/(1+hi)~e[ncl[k]]);
                    k=k+1;
                  endo;


                  @ Matrix Dres will contain min. among deletion residuals @
                     if S2[mm-p+1,2]>0;
                       selDres=sortc(ord,1);
                     
                     
                     @
                      seldres[1,1];
                      minc(ABS(ord[.,1])); wait;
                     @
                      if selDres[1,2]>=0;
                       Dres[mm-p,2]=sqrt(seldres[1,1]/S2[mm-p+1,2]);
                      else;
                       Dres[mm-p,2]=-sqrt(seldres[1,1]/S2[mm-p+1,2]);
                      endif;
                        
                     else;
                     endif;
                 else;
                 endif;

                    @ Include ncl in matrix mIn 
                        if mm>=p;
                           jk=jk+1;
                        mIn[1:rows(ncl),jk]=ncl;
                          else;
                        endif; @

                 @ Tout will contain a F test that the units belonging to
                    ncl are outliers @
                    if mm>p and S2[mm-p+1,2]>0;
                       Tout[mm-p,2]= 1/(rows(ncl)) *
                         EE[ncl]'invpd(IH[ncl,ncl])*EE[ncl]/S2[mm-p+1,2];
                         @ Numerator of Cov ratio @
                         CovR[mm-p,2]=S2[mm-p+1,2]^(2*p)*det(mmX);
                    else;
                    endif;

              oldbsb=bsb;
              oldmmX=mmX; @ store inv(X'X) with old basic subset for
                            computing recursive residuals @
              @ order the r_i and include the smallest among the units
                forming the group of potential outliers @
                   ord=sortc(R,2);
                   @ bsb= units forming the new basic subset @
                   bsb=ord[1:(mm+1),1];
                   mX=X[bsb[.,1],.];  @ subset of X @
                   vY=Y[bsb[.,1]];    @ subset of untransformed Y @
                   vYt=Yt[bsb[.,1]];  @ subset of transformed Y @
                     if mm+1 < n;
                        @ ncl= units forming the new noclean @
                        ncl=ord[mm+2:n,1];
                     else;
                     endif;

                                               
                   Unit=setdif(bsb,oldbsb,1);
                   Unitout=setdif(oldbsb,bsb,1); 
                   diff[mm-p-tp+1,2]=rows(Unit);
                   if diff[mm-p-tp+1,2]<13;
                    Un[mm-p+1,2:rows(unit)+1]=(Unit');
                    mIn[mm-p+1,2:rows(unitout)+1]=(Unitout');
                    if e[unit[1]]>0;
                      Un[mm-p+1,cols(Un)]=1;
                    else;
                      Un[mm-p+1,cols(Un)]=-1;
                    endif;
                   endif;
                                 
                   
          else;
         endif;
             posress=rows(selif(e[oldbsb],e[oldbsb].>=0));
             posresU=rows(selif(e,e.>=0));
                   
             diff[mm-p-tp+1,3]=posress/mm;
             diff[mm-p-tp+1,4]=posresU/nn;
             diff[mm-p-tp+1,5]=posress-(mm-posress);
             diff[mm-p-tp+1,6]=posresU-(nn-posresU);


                 @ FR= n x (nn-m+1) Matrix. It contains fwd search
                   residuals in every step
                   Remark: residuals are specified in sclar sh @
                   if sh==0;
                     FR[.,mm-p+1]=e;
                   else;
                      se=selif(r[.,1],e.<0);
                     FR[.,mm-p+1]= real(sqrt(r[.,2]));
                     FR[se,mm-p+1]= -FR[se,mm-p+1];
                   endif;

                mm=mm+1;
          endo;
   fr=fr/sqrt(rsq[1,2]);



 retp(BB,TT,pvt,as,ku,nor,ram,lik,sco,s2,rst,Coo,dres,
      cc,rres,un,tout,covr,fr,lev,mIn,diff);
endp;

