/*
**  FWDOLSST
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {Sco,S21,Co}=fwdolsst(X,y,w,la,sub,init)
**
**                                To extract the t statistics for
**                                variable w during the fwd search when variable w is not
**                                included in the model
**
*/

/*
** FWDOLSST
**
** PURPOSE:  to extract the t stat for variable w when
**           variable w is not included in the model (see Atkinson and Riani, 2002 Biometrika)
**
** FORMAT:   {Sco,S21,Co}=fwdolsst(X,y,w,la,sub,init)
**
** INPUT:     X = n x p data matrix of expl. variables;
**                n observations and p variables
**            y = n x 1 vector (response variable)
**            w = extra variable whose t test must be computed
**           la = transformation parameter for the response
**          sub = number of subsamples to extract
**                if sub=1 all subsamples are extracted
**         init = point in which the fwd search must start
**
**
** OUTPUT: Sco = n-p-1 x 2 matrix
**               1st col = fwd search index
**               2nd col = t test for w
**        S21  = n-p+1 x 3 matrix
**               1st col = fwd search index
**               2nd col = S2
**               2nd col = R2
**         CO  = n-p+1 x 2 matrix
**               1st col = fwd search index
**               2nd col = correlation between
**                         predicted values of X and w
**
**      REMARK: Program does not add constants
**
*/



proc (3)=fwdolsst(X,y,w,la,sub,init);

          local n,p,SEQ,mE,mE1,lm,BS,Sco,S21,CO,mm,mX,start,vY,Yt,
                vYt,vw,B,sres,Rsq,NO,RA,SC,e,R,ord,bsb,Tw,cor;

     n=rows(X);
     p=cols(X);
     seq=seqa(1,1,n);
     R=seq~zeros(n,1); @  The second column of matrix R will
                                contain the OLS residuals at
                                each step of the forward search   @

     {mE,lm,BS}=lms(X,y,la,sub);

  @ Sco = (n-p-2) x 2 matrix. Score test @
    Sco=seqa(p+2,1,n-p-1)~miss(zeros(n-p-1,1),0);

  @ S21 = (n-p+1) x (3) matrix.
          1st col = fwd search index
          2nd col = S2
          3rd col = R2  @
    S21=seqa(p,1,n-p+1)~miss(zeros(n-p+1,2),0);

  @ CO = (n-p+1) x (2) matrix.
          1st col = fwd search index
          2nd col = Correlation between predicted and extra var. @
    CO=seqa(p,1,n-p+1)~miss(zeros(n-p+1,2),0);

            mE1=sortc(seq~abs(mE[.,1]),2);
                               start:
                 mm=p+init;
                  @ bsb contains the indexes of the units forming BSB @
                         Bsb=BS;

                  mX=X[bsb,.]; @ initial basic subset for X @
                 @  if rank(mX)<p;
                     "rank less than p";; wait;
                      init=init+1;
                      goto start;
                   else;
                   endif;  @

                   vY=y[bsb];   @ inital basic subset for y @
                   vw=w[bsb];

                  @ Transform all the units of the response var. @
                     if la==0;
                         vYt=ln(vY);
                         Yt=ln(Y);
                     else;
                         vYt=vY^(la);      @ transform the subset @
                         Yt=(Y^la);        @ transform all the units @
                     endif;

       @  forward search @
        mm=p+init; do until mm==n+1;
          "mm";; mm;

               @    if rank(mX)<p;
                     "WARNING: Matrix mX is not full rank when mm=";; mm;
                     "rank less than p";; wait;
                      mm=mm+1;
                      goto update;
                   else;
                   endif;   @


                  @ ols on basic subset using transf. response @
                  {B,sres,Rsq,NO,RA,SC}=regressi(vYt,mX,0,0);

                        @ Store correlation between predicted values
                          and additional variable @
                         cor=corrx(mX*B[.,1]~vw);
                         Co[mm-p+1,2]=cor[1,2];
                        @ Store S2 and R2 @
                        S21[mm-p+1,2:3]=Rsq[1,2]~Rsq[1,1];

                  if mm>p+init+1;

                        {Tw}=addt(vy,mX,vw,la);
                        @ Store t-test for additional explanatory
                          variable  @
                        Sco[mm-p-1,2]=Tw;

                 else;
                 endif;


            @ res. for the overall sample using transf. response
              and beta based on units forming BSB @
            e=Yt-X*B[.,1];

            r[.,2]=e.^2;    @ r_i =e_i^2 @

           if mm<n;

              @ order the r_i and include the smallest among the units
                forming the group of potential outliers @
                   ord=sortc(R,2);

                   update:

                   @ bsb= units forming the new basic subset @
                   bsb=ord[1:(mm+1),1];
                   mX=X[bsb,.];  @ subset of X @
                   vY=Y[bsb];    @ subset of untransformed Y @
                   vYt=Yt[bsb];  @ subset of transformed Y @
                   vw=w[bsb];    @ subset for additional expl. variable @

            else;
            endif;
         mm=mm+1;
       endo;

   retp(Sco,S21,Co);
 endp;

