/*
**  FWDPCA
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {pos,pot,malaF,In,Un,rdet,varex,avec,sComp,sComp2}=fwdpca(X,bsb,k,st);
**
**                                To perform a fwd search using principal component analysis
**                                A series of quantities are monitored
**                                in each step of the forward
**                                procedure
*/

/*
** FWDPCA
**
** PURPOSE:  to perform robust principal component analysis
**           plot technique. It also produces min, average,
**           maximum of Mahalanobis distances in
**           each step of the forward procedure
**
** FORMAT: {pos,pot,malaF,In,Un,rdet,varex,avec,sComp,sComp2}
**                                   =fwdpca(X,bsb,k,st)
**
** INPUT:     X = n x p data matrix; n observations
**                and p variables
**          bsb = m x 1 vector containing the
**                the indexes of the m units forming
**                the initial subset
**            k = scalar specifying the number of principal
**                components to be considered in the fwd search
**           st = scalar
**                if st==1 the variables are standardized
**                else the variables are simply centerd
**                removing the arithmetic means
**
** OUTPUT:  pos =  (n-m+1) x 4 matrix.  The first column of pos
**                contains the
**                dimension  of the sample used to compute Mahalanobis
**                distances Columns 2,3 respectively contain
**                Max, Average and trimmed average of Mahalanobis
**                distances of the units
**                trimmed average = average exluding the biggest
**                                  MD of the subset
**                forming the basic subset
**          pot =  (n-m-1) x 2.The first column contains the
**                 dimension of the sample used in order to compute
**                 Mahalanobis distances, the second
**                 reports the minimum of Mahalanobis distances among
**                 the units which form the group of potential
**                 outliers
**        malaF =  n x (n-m+1) matrix of MD in
**                 each step
**                 of the fwd search
**                 The first row is a forward search for the first unit
**                 First column is a seqa ffrom 1 to n
**            IN = (nn-mm) x 11 matrix.
**                 1st col = forward search index
**                 2nd col = unit with max MD in bsb (mm-th MD)
**                 3rd col = unit with (mm+1)-th MD
**                    ...
**                 11th col = unit with (mm+10)-th MD
**            Un = (nn-mm) x 11 Matrix which contains the unit(s)
**                 included in bsb at each step of the fwd search
**                 REMARK: in every step the new subset is compared
**                 with the old subset. Un contains the unit present
**                 in the new subset but not in the old one
**         rdet =  (n-m) x 3  matrix
**                 1st col = forward search index
**                 2nd col = ratio between |S_(i)| (determinant of
**                 the covariance matrix excluding the units of noclean)
**                 and |S| (determinant of covariance matrix)
**                 3rd col = trace of S_(i)
**        varex =  (n-m+1) x (k+1) matrix.
**                 1st col = fwd search index
**                 2nd col = % of var explained by the first PC
**                       ...
**                 (k+1)-th col = % of var explained by the k-th PC
**         avec =  (nn-mm+1) x k*p+1 matrix.
**                 1st col = index related to the steps of the forward search
**                 2nd col : (p+1) col  = first eigenvector
**                   ...
**                 p*(k-1)+2 : p*k +1   = kth eigenvector   @
**        sComp =  n x (n-m+1) matrix of scores of the first PC
**                 in each step
**                 of the fwd search
**                 The first row is a forward search for the first unit
**                 First column is a seqa ffrom 1 to n
**       sComp2 =  n x (n-m+1) matrix of scores of the second PC
**                 in each step
**                 of the fwd search
**                 The first row is a forward search for the first unit
**                 First column is a seqa ffrom 1 to n
**
**
*/


 proc (10)=fwdpca(X,bsb,k,st);

   local seq,ym,vc,malaf,in,ds,rdet,va,varX,tr,ml1,chan,
       nn,n,mm,i,gg,pos,pot,zs,min,mala,y,ii,p,maxind,j,Xyst,
       Xb,varex,avec,val,vect,cor,comp,sComp,sComp2,Un,Unit,nbsb,Te;
nn=rows(X); p=cols(X);
seq=SEQA(1,1,nn);

mm=rows(bsb);
@   The first column of pos contains the dimension
    of the sample used to compute Mahalanobis distances
    Columns 2,3 respectively contain
    Max and Average of Mahalanobis distances of the units
    forming the basic subset
    Column 4 trimmed average of Mahalanobis distances of the
    the units forming the subset @
  pos=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,3);

@   pot has two columns the first contains the dimension of the
    sample used in order to compute Mahalanobis distances, the second
    reports the minimum of Mahalanobis distances among the units
    which form the group of potential outliers @
  pot=seqa(mm,1,nn-mm)~zeros(nn-mm,1);

   @ Malaf= nn x (nn-mm) matrix of Mahalanobis distances
       in each step of the forward search @
   Malaf=seq~zeros(nn,nn-mm+1);

   @    In= Matrix whose 2nd column contains the unit in mm-th position
        at each step of the fwd search. The other cols refer to the
        units ranked (mm+1)-th (mm+10)-th in non dexreasing order @
   In=seqa(mm+1,1,nn-mm)~miss(zeros(nn-mm,10),0);

   @    Un= Matrix which contains the unit(s) included in
            bsb at each step of the fwd search @
   UN=In;
   @   rdet= (nn-mm) x 2 matrix.
       1st col = index related to the steps of the forward search
       2nd col = |S_(i)| / |S|
                  where S_(i) is the determinant of the cov.
                  matrix of n-i observations with centroid based on
                  (n-i) observations
       3rd col = trace(S_(i))/ trace(S) @
   rdet=seqa(mm,1,nn-mm)~zeros(nn-mm,2);

   @   varex= (nn-mm+1) x k+1 matrix.
       1st col = index related to the steps of the forward search
       2nd col : (k+1) col  = percentage
                  of variance explained by the first k prin. comp. @
   varex=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,k);
   @   avec= (nn-mm+1) x k*p+1 matrix.
       1st col = index related to the steps of the forward search
       2nd col : (p+1) col  = first eigenvector
         ...
       p*(k-1)+2 : p*k +1   = kth eigenvector   @
   avec=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,k*p);
   @   sComp= nn*(nn-mm) matrix.
       Scores of the first principal comp.
       in each step of the fwd search @
   sComp=seq~miss(zeros(nn,nn-mm+1),0);

   @   sComp2= nn*(nn-mm) matrix.
       Scores of the first principal comp.
       in each step of the fwd search @
   sComp2=sComp;

   varX=vcx(X);
   ds=det(varX);
   tr=sumc(diag(varX));


    mm=rows(bsb);
    min=rows(bsb);

    /*** The 2nd col. of mala will contain MD for all the units
         in each step of the forward search  ***/
     mala=seq~zeros(nn,1);

     y=bsb;

 do until mm==nn+1;  @ nn= sample size @
      mm;
      Xb=X[y,.];
      ym=meanc(Xb);
      va=vcx(Xb);
     vc=invpd(va);

     {val,vect,cor,Comp}=pca(Xb,k,st,0);
     @    if vect[1,1]*vect[2,1]<0;
           vect=vect[., 2 1];
         else;
         endif;
           vect[.,1]=abs(vect[.,1]);

         vect;wait;   @

   @   varex= (nn-m+1) x k+1 matrix.  @
   varex[mm-min+1,2:k+1]=(val[.,2]');
   @   avec= (nn-mm+1) x k*p+1 matrix.  @
                @   Store correlations    @
    avec[mm-min+1,2:(k*p+1)]=(vec(cor))';
       @  Store eigenvectors
      avec[mm-min+1,2:(k*p+1)]=vec(vect)';  @
  @ The following loop controls that the sign of the elements
    of the eigenvectors are coherent in every step of the fwd search @

  if mm>min;
      j=1; do until j==k+1;
        maxind=maxindc(abs(avec[mm-min+1,(j-1)*p+2:j*p+1]'));
     @    vect;
        maxind=maxindc(vect[.,j]);
        "jjj";  j;    @
       @  avec[mm-min+1,(j-1)*p+2:j*p+1]';
         vect[.,j];      @

       @  "maxind";  maxind;   wait;  @
         if avec[mm-min+1,(j-1)*p+maxind+1]*
                avec[mm-min,(J-1)*p+maxind+1]<0;
           avec[mm-min+1,(j-1)*p+2:j*p+1]=-avec[mm-min+1,(J-1)*p+2:j*p+1];
           vect[.,j]= -vect[.,j];
           Comp[.,j]= -Comp[.,j];
          endif;
       j=j+1;
      endo;
   endif;
        @  "Eigenvector";   @
       @   vect[.,.]; wait;  @
      chan=rows(bsb);
     if mm==chan;
       ml1=(X-ym')*vect[.,1];
       @   j=2; do until j==mm-min+2;

                i=1; do until i==nn+1;
                  if ml1[i]<0;
                       malaf[i,j]= -malaf[i,j];
                 else;
                 endif;
                 i=i+1;
                endo;
             j=j+1;
           endo;    @
     else;
     endif;

     @ 2nd col of Matrix Mala contains Mahalanobis distances @
     i=1;  do until i==nn+1;
           mala[i,2]=sqrt((X[i,.]-ym')*vc*(X[i,.]'-ym));
            @  if mm>=chan;
                  if ml1[i]<0;
                       mala[i,2]= -mala[i,2];
                 else;
                 endif;
              else;
              endif;   @
            i=i+1;
         endo;

      Malaf[.,mm-min+2]=mala[.,2];


   @   sComp= n x (nn-mm+1) matrix.  @
   sComp[y,mm-min+2]=Comp[.,1];      @ Store 1st PC scores @

   Xyst=(X-ym')./(sqrt(diag(va))'); @  Store all 1st PC scores  @
   sComp[.,mm-min+2]=Xyst*vect[.,1];

    @ Store all 1st PC scores @
   if k >1;
       @  comp; wait; @
   sComp2[y,mm-min+2]=Comp[.,2];     @ Store 2nd PC scores @
   @ scomp2[y[1],mm-min+2];  @

   sComp2[.,mm-min+2]=Xyst*vect[.,2];    @  Store all 2nd PC scores @
   @  scomp2[y[1],mm-min+2];    wait;   @

   endif;
      @ The 2nd col. of zs contains sorted MD @
      zs=sortc(abs(mala),2);
      @ pos has 3 columns: the first contains the dimension of bsb
       set, Columns 2, and 3 respectively contain
       Max, Average, and trimmed average forming the basic subset @

                         @   maximum       average  @
    pos[mm-min+1,2:4]=zs[mm,2]~meanc(zs[1:mm,2])~meanc(zs[1:(mm-1),2]);

     if mm<nn;

     @ rdet contains in the 2nd and 3rd col. functions of Wilks ratio @
       rdet[mm-min+1,2]=det(va)/ds;
       rdet[mm-min+1,3]=sumc(diag(va))/tr;
     @ pot contains minimum of Mahalanobis distances among the units
       which form the group of potential outliers @
       pot[mm-min+1,2]=zs[mm+1,2];

     @ the dimension of the clean data set increases of one unit.
       vector y contains the indexes corresponding to the units of
       the new basic subset @
       y=zs[1:mm+1,1];

             nbsb=y;
             unit=setdif(nbsb,bsb,1);
             if rows(unit)>=cols(un);
                  unit=unit[1:cols(un)-1];
                 "Warning: interchange";
             else;
             endif;
             @     rows(unit);
                  cols(Un);
                  wait;  @
             UN[mm-min+1,2:rows(unit)+1]=(unit');

             @ Store in matrix In unit ranked mm_j-th (mm_j+10)-th
             in terms of MD @
             Te=zs[.,1]|miss(zeros(10,1),0);
             In[mm-min+1,2:11]=Te[mm+1:mm+10,1]';
            bsb=nbsb;
       else;
   endif;


                    mm=mm+1;
                  endo;
   retp(pos,pot,Malaf,In,Un,rdet,varex,avec,sComp,sComp2);
endp;

