/*
**  GLM
**  written by MARCO RIANI
**
**  FORMAT                                    PURPOSE               Line
** =======================================================================
**
**  {B,RS,Hi,WW,R,notcv}=glm(X,y,q,fr,mint,tol,b0,off) To fit a
**                                                generalized
**                                                linear model
**
**
*/


/*
**
** GLM
**
** PURPOSE:  to fit a generalized linear model, estimate
**           the parameters and produce diagnostics
**
**
**
** FORMAT:
**
** {B,RS,Hi,WW,R,notcv} = glm(X,y,q,fr,mint,tol,b0,off)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             q  = scalar which specifies the kind of GLM
**                  q=1 ==> Poisson link
**                  q=2 ==> Logit link
**                  q=3 ==> Complementary Log log link
**                  q=3.05 ==> Gamma link (Box Cox la=2)
**                  q=3.1 ==> Gamma link (Box Cox la=1)
**                  q=3.2 ==> Gamma link (Box Cox la=0.5)
**                  q=3.3 ==> Gamma link (Box Cox la=0)
**                  q=3.4 ==> Gamma link (Box Cox la=-0.5)
**                  q=3.5 ==> Gamma link (Box Cox la=-1)
**                  q=3.6 ==> Gamma link (Power family la=3)
**                  q=3.7 ==> Gamma link (Power family la=2)
**                  q=3.8 ==> Gamma link (Power family la=1)
**                  q=3.9 ==> Gamma link (Power family log)
**                  q=4   ==> Gamma link (Power family la=-1)
**                            (inverse)
**                  q>4  and q<5 Gamma link all values of lambda
**                       between 0 and -1
**                       For example q=4.25 ==> la=-0.25
**                       For example q=4.5 ==> la=-0.5
**                  q=5 ==> Inverse Gaussian link
**                  q=6 ==> Probit link
**                  q=7 ==> Arc sin link
**                  q=8 ==> Modified logit link
**             fr = n x 1 vector of frequencies
**                  if fr = scalar 0 a n x 1 vector of 1s created
**           mint = scalar. If mint is equal to 1
**                  the null model is also considered
**            tol = 2 x 1 vector
**                 tol[1] = tolerance level for parameter estimation
**                 tol[2] = specifies the stopping criterion for conv.
**                    tol[2]=1 implies stopping rule based on deviance
**                     else    implies stopping rule base on beta coeff.
**           b0 =  p x 1 vector containing starting values for
**                 beta coefficients
**                 if b0=0 program itsself
**                 computes initial starting values
**           off = n x 1 vector which defines an offset
**                 REMARK: off = scalar 0 implies no offset
**
**
**
** OUTPUT:  B =  p x 4 matrix
**               1st col = b coefficients
**               2nd col = Std. error for each b
**               3rd col = t-stat for beta coeff.
**               4th col = P-value of t-stat
**         RS =  n x 5 matrix
**               1st col. simple residuals  y_i - \mu_i
**               2nd col. Pearson residuals  McN p. 37 eq. (2.11)
**               3rd col. deviance residuals  rD McN p. 30
**               4th col. Studentized Pearson residuals
**                        McN p. 397 eq. (12.4)
**               5th col. Studentized Deviance residuals
**                        McN p. 397  eq. (12.5)
**         Hi  = n x 1 vector containing
**               diagonal elements of generalized Hat matrix
**               McN p. 397 eq. (12.3)
**         WW  = n x 1 vector containing
**               the inverse of the weights associated to the n units
**         R   = 2 x 2 matrix which contains:
**               R[1,1]= pseudo R^2
**               R[2,1]= dispersion parameter (phi)
**                       moment estimator is used
**                       phi= (sum of squares of Pearson residuals)/(n-p)
**               R[1,2]= deviance for the current model
**               R[2,2]= deviance for the null model
**               pseudo R^2 = 1 - (current deviance)/(null deviance)
**       notcv = scalar  which informs about convergence
**               notcv=0 implies convergence
**               notcv=1 implies non convergence
**
**  REMARK1    this program does not add a column of 1 in matrix X to
**             create the intercept.
**
**  REMARK2    The iterative least squares method is used for
**             parameters estimation
**
**  REMARK3    gauss function trap is set equal to 1
**
**  REMARK4   The estimate of the dispersion parameter is always
**            computed using the moment estimator. Standard
**            errors use this estimate only when the link
**            is GAMMA or INVERSE GAUSSIAN. In all other
**            cases (poisson, binomial, arc sin) we use
**            a value equal to 1
**
**
*/


proc (6)=glm(X,y,q,fr,mint,tol,b0,off);
   local    n,p,i,j,mu0,eta0,dm,v0,z0,w0,Xb,b_new,b_old,b1,one,fin,nulm,
            start,th,mu,V,z,W,WW,e,rP,Cb,H,Hi,Xb0,notcv,dev_old,dev_new,
            rd,phi,rSDR,rSPR,B,df,dev,RS,R,cnt,count,phi1,lam;

   trap 1;
   n=rows(X);
   p=cols(X);
   df=n-p;
   one=ones(n,1);
   nulm=0; @ nulm = constant for calculating the deviance
                    under the null model @
   count=0;
   @ B = p x 4 matrix which will contain  beta coeff. std err. t stat
     and p values of tstat. @
     B=zeros(p,4);
   @ mu0 = n x 1 vector containing initial estimate of mu @
     mu0=zeros(n,1);
     mu=mu0; @ Initialize mu @
     dm=mu0; @ Initialize dm @
     V=mu0; @ Initialize V @

   @ R = 2 x 2 matrix which will contain
       [1.1] =pseudo R^2,            [2,1] = disp. parameter,
       [1,2]= residual deviance,     [2.2] = null deviance @
     R=zeros(2,2);
  cnt=0;
   if q==3.05;
       lam=2;
   elseif q==3.1;
       lam=1;
   elseif q==3.2;
       lam=0.5;
   elseif q==3.4;
       lam=-0.5;
   elseif q==3.5;
       lam=-1;
   else;
   endif;

   if fr==0;
    fr=one;
   else;
   endif;
         @ q=1 ==> Poisson link
           q=2 ==> Logit link
           q=3 ==> Complementary Log log link
           q=4 ==> Gamma link
           q=5 ==> Inverse Gaussian link
           q=6 ==> Probit link
           q=7 ==> Arc sin  link
           q=8 ==> Modified Logit link  @

    j=0;    @ j counts the iterations in the estimation procedure @
    notcv=0; @ notcv = dummy variables linked to convergence @

                   @  start:   @

    if b0==0 and rows(b0)==1;

                  mu0=y+0.3;  @ 0.3=old value init estimate of \mu0 @
             if maxc(fr)==1;
                 i=1; do until i==n+1;
                        if y[i] /= 0 and y[i] /=1;
                           mu0[i]=y[i];
                        elseif y[i]==0;
                           mu0[i]=y[i]+0.1;
                        else;
                           mu0[i]=y[i]-0.1;
                        endif;
                    i=i+1;
                 endo;
             else;
                 i=1; do until i==n+1;
                        if y[i] /= 0 and y[i] /=1;
                           mu0[i]=y[i];
                        elseif y[i]==0;
                           mu0[i]=y[i]+0.01;
                        else;
                           mu0[i]=y[i]-0.01;
                        endif;
                    i=i+1;
                 endo;
             endif;

                 if q==1;     @ Poisson link @
                    eta0=ln(mu0); @ initial estimate of \hat{\eta}_0 @
                    dm=1/(mu0);  @ Derivative (d \eta) /(d \mu)_0  @
                    v0=(mu0+1e-8);
                     @ v0= variance function evalueted at  \mu_0 @

                 elseif q==2 or q==8; @ Logit  link @
                  @  mu0=(y+0.5)./(fr+1); do not write fr
                     see for example the Efron data  @
                    eta0=ln(mu0./(1-mu0));
                    dm=1/(mu0.*(1-mu0));
                    v0=mu0.*(1.-mu0)./fr;

                 elseif q==3; @ Complementary log log link @
                   eta0= ln(1+ln(1-mu0));
                   dm= 1/( (1-mu0).*ln(1-mu0) );
                   v0=mu0.*(1-mu0) ./fr;

                 elseif q>3 and q <=3.5;  @ GAMMA LINK (Box Cox) @
                   if q ne 3.3;
                    eta0= (mu0^lam-1)/lam;
                    dm=  mu0^(lam-1);
                   else;
                    eta0 = ln(mu0);
                    dm= 1/mu0;
                   endif;
                    V0=(mu0^2)./fr;

                 elseif q==3.6;  @ GAMMA LINK (third power) @
                   eta0=  mu0^(1/3);
                   dm=  3*(mu0^2);
                   V0=(mu0^2)./fr;

                 elseif q==3.7;  @ GAMMA LINK (square) @
                   eta0=  sqrt(mu0);
                   dm=  2*mu0;
                   V0=(mu0^2)./fr;

                 elseif q==3.8;  @ GAMMA LINK (identity) @
                   eta0=  mu0;
                   dm=   1;
                   V0=(mu0^2)./fr;

                 elseif q==3.9; @ GAMMA LINK (log) @
                   eta0=  ln(mu0);
                   dm=  1/(mu0);
                   V0=(mu0^2)./fr;

                 elseif q==4; @ GAMMA LINK (inverse) @
                   eta0=  1/(mu0+0.01);
                   dm=  -1/((mu0+0.001)^2);
                   v0=(mu0+0.001)^2./fr;
               /****

                 elseif q==4.2; @ GAMMA LINK (-0.5) @
                   eta0=  1/(mu0^2);
                   dm=   -0.5*(mu0^(-1.5));
                   V0=(mu0^2)./fr;
                ****/
                 elseif q>4 and q<5; @ GAMMA LINK (4+lambda) @
                   eta0= 1/((mu0+0)^(1/(q-4)));
                   dm= (4-q)*((mu0+0)^(3-q));
                   V0=((mu0)^2)./fr;

                 elseif q==5;  @ INVERSE GAUSSIAN @
                   eta0=  1/sqrt(-2*mu0) @     1/sqrt(mu0)   @ ;
                   dm=    mu0^(-3)     @       -2/(mu0^3)  @  ;
                   V0= mu0^3;

                 elseif q==6; @ PROBIT LINK @
                   eta0=cdfn(mu0);
                   dm=1/pdfn(mu0);
                   v0=mu0.*(1.-mu0)./fr;

                 elseif q==7; @ Arc sin LINK @
                      eta0=arcsin(2*mu0-1);
                      dm= 2/(sqrt(1-(2*mu0-1)^2));
                      v0=mu0.*(1.-mu0)./fr;

                 else;
                 endif;

                z0=eta0+(y-mu0).*dm-off;
                W0=sqrt((dm.^2).*v0);   @ vector n x 1 @
                Xb0=X./W0;                 @ matrix n x p @
              if iscplx(Xb0)==1 or iscplx(z0)==1 OR COLS(x)==1;
                  b_new=invpd(X'X)*X'y;
              else;
                   if rank(Xb0)==p;
                         b_new=invpd(Xb0'Xb0)*Xb0'(z0./W0);
                   elseif rank(X)==p;
                         b_new=invpd(X'X)*X'y;
                   else;
                       notcv=1;
                       b1=zeros(p,1);
                       b_new=b1;
                       "rank less than k"; wait;
                      goto notconv;
                   endif;
                endif;

                   @  b_new={0.434,0.007,0.00036};    @
                     @ b_new={847.918,-31.2688,-2.27368};   @
    else;
      b_new=b0;
    endif;
                b1=b_new;
                b_old=b_new+1e+10;
                     start:
                dev_new=1e+2; dev_old=dev_new+1e+7;
        @ PARAMETERS ESTIMATION @
        @  dev_old and dev_new depend on tol[2] @
        do until  abs(dev_old - dev_new) < tol[1];

                   j=j+1; @     j;  wait;  @
                    @ "dev_old"; DEV_OLD;     @
                   dev_old=dev_new;
                  @  "DEV_OLD AFTER"; DEV_OLD;  @
                   b_old = b_new;
                   th=X*b_new+off;


                 if q==1;     @ POISSON LINK @
                        i=1; do until i==n+1;
                           if th[i]>25;
                             mu[i]=1e+10;
                             dm[i]=1e-10;
                             V[i]=1e+10;
                             count=count+1;
                           elseif th[i]<-25;
                             mu[i]=1e-10;
                             dm[i]=1e+10;
                             V[i]=1e-10;
                             count=count+1;
                           else;
                             mu[i]=exp(th[i]);
                             dm[i]=1/mu[i];
                             V[i]=mu[i];
                           endif;
                          i=i+1;
                         endo;

                 elseif q==2; @ LOGIT LINK @
                        i=1; do until i==n+1;
                           if th[i]>25;
                             mu[i]=1-1e-12;
                             dm[i]=1e+8;
                             V[i]=1e-8;
                             count=count+1;
                           elseif th[i]<-25;
                             mu[i]=1e-12;
                             dm[i]=1e+8;
                             V[i]=1e-8;
                             count=count+1;
                           else;
                             mu[i]=exp(th[i])/(1+exp(th[i]));
                             dm[i]=1/((mu[i]).*(1-mu[i]));
                             V[i]=mu[i].*(1-mu[i]) ./fr[i];
                           endif;
                          i=i+1;
                         endo;

                 elseif q==3; @ COMPLEMENTARY LOG LOG LINK @
                        i=1; do until i==n+1;
                           if th[i]>5;
                             mu[i]=1-1e-10;
                           elseif th[i]< -18;
                             mu[i]= 1e-10;
                           else;
                             mu[i]=1-exp(-exp(th[i]));
                           endif;
                          i=i+1;
                         endo;
                   dm= -1/( (1-mu+1e-12).*ln(1-mu+1e-12) );
                   V=mu.*(1-mu) ./fr+ 1e-12;


                 elseif q>3 and q <=3.5;  @ GAMMA LINK (Box Cox) @
                   if q ne 3.3;
                    mu=  (1+lam*th)^(1/lam);
                    dm=  mu^(lam-1);
                   else;
                    mu=  exp(th);
                    dm= 1/mu;
                   endif;
                    V=(mu^2)./fr;

                 elseif q==3.6;  @ GAMMA LINK (third power) @
                   mu=  th^(1/3);
                   dm=  3*(mu^2);
                   V=(mu^2)./fr;

                 elseif q==3.7;  @ GAMMA LINK (square) @
                   mu=  sqrt(abs(th));
                   dm=  2*mu;
                   V=(mu^2)./fr;

                 elseif q==3.8;  @ GAMMA LINK (identity) @
                   mu=  th;
                   dm=   1;
                   V=(mu^2)./fr;

                 elseif q==3.9;  @ GAMMA LINK (log) @
                   mu=  exp(th);
                   dm=   1/(mu+1e-8);
                   V=(mu^2)./fr;

                 elseif q==4; @ GAMMA LINK (inverse) @
                   mu=  1/th;
                   dm= -  1/(mu^2);
                   V=(mu^2)./fr;
            /*****
                 elseif q==4.2; @ GAMMA LINK (-0.5) @
                   mu= 1/(th^2);
                   dm= -0.5*(mu^(-1.5));
                   V=(mu^2)./fr;
            ****/
                 elseif q>4 and q<5; @ GAMMA LINK (4+lambda) @
                   mu= 1/(th^((1/(q-4))));
                   dm= (4-q)*(mu^(3-q));
                   V=(mu^2)./fr;

                 elseif q==5;  @ INVERSE GAUSSIAN LINK @
                       if iscplx(th)==1;
                           b_new=rndn(rows(b_new),1)/100;
                           goto start;
                       else;
                       endif;
                   mu=  1/sqrt(th)   @   1/sqrt(-2*th)  @    ;
                   dm=  -2/(mu^3)    @  mu^(-3)        @   ;
                   V= mu^3;

                 elseif q==6; @ PROBIT LINK @
                   mu= cdfn(th);  @ CDF evaluated in th @
                   dm=1/pdfn(TH);  @ inv. of density evaluated at th @
                   V=mu.*(1-mu)./fr;

                 elseif q==7; @ ARC SIN LINK @
                  i=1; do until i==n+1;
                    if th[i]>= pi/2;
                     mu[i]=1 -1e-8;
                     dm[i]=1e+6;
                    elseif th[i]<= - pi/2;
                     mu[i]=0 +1e-8;
                     dm[i]=1e+6;
                    else;
                     mu[i]=  0.5 * (1+sin(th[i]));
                     dm[i]=  2/sqrt(1-((2*mu[i]-1)^2));
                    endif;
                   i=i+1;
                  endo;
                  V= 1e-6+ mu.*(1.-mu)./fr;

                 elseif q==8; @ MODIFIED LOGIT LINK @
                        i=1; do until i==n+1;
                           if th[i]>13.8;
                             mu[i]=1-1e-6;
                             dm[i]=1e+6;
                             V[i]=1e-6;
                             count=count+1;
                           elseif th[i]< -13.8;
                             mu[i]=1e-6;
                             dm[i]=1e+6;
                             V[i]=1e-6;
                             count=count+1;
                           else;
                             mu[i]=exp(th[i])/(1+exp(th[i]));
                             dm[i]=1/((mu[i]).*(1-mu[i]));
                             V[i]=mu[i].*(1-mu[i]) ./fr[i];
                           endif;
                          i=i+1;
                         endo;
                 else;
                 endif;

               @  rD  = Deviance Residuals @
               @  phi = Estimate of disperion parameter phi:  McN p. 296 @
              @ RESIDUAL DEVIANCE (SPLUS) = SCALED DEVIANCE (mcN p. 34)
                = sumc(rD^2)  = -2 lik   @
               if q==1;     @ POISSON LINK @
                     rD=zeros(n,1);
                  i=1; do until i==n+1;
                     if y[i]>0;
                      rD[i]=2*(y[i].* ln(y[i]./mu[i]) - (y[i]-mu[i])  );
                     else;
                      rD[i]= 2* mu[i];
                     endif;
                    i=i+1;
                   endo;

               elseif q==2 or q==3 or q==6 or q==7 or q==8;
                @ BINOMIAL  LINK @
                     rd=zeros(n,1);
                 if maxc(fr)>1;
                    i=1; do until i==n+1;
                      if y[i]==mu[i];
                        rd[i]=0;
                      elseif y[i]==0;
                        rd[i]=2*(fr[i]*ln(1/(1-mu[i]-1e-12)));
                      else;
                    rd[i]=2*(fr[i]*y[i]*ln(y[i]/(mu[i]+1e-12))+
                    (fr[i]-fr[i]*y[i])*ln((1-y[i]+1e-12)/
                    (1-mu[i]+1e-12)));
                      endif;
                    i=i+1;
                    endo;
                 else;
                    i=1; do until i==n+1;
                      if mu[i]==1;
                        rd[i]= -2*(y[i]*ln(mu[i])+
                        (1-y[i])*ln(1-mu[i]+1e-12));
                      elseif mu[i]==0;
                        rd[i]= -2*(y[i]*ln(mu[i]+1e-12)+
                        (1-y[i])*ln(1-mu[i]));
                      else;
                        rd[i]= -2*(y[i]*ln(mu[i])+
                        (1-y[i])*ln(1-mu[i]));
                      endif;
                    i=i+1;
                    endo;
                 endif;

               elseif q>3 and q<5; @ GAMMA LINK @
                 if minc(abs(mu))==0;
                      "Convergence not obtained";
                      notcv=1;
                     goto notconv;
                  else;
                 endif;

                     rD=zeros(n,1);
                 i=1; do until i==n+1;
                   if y[i]>0;
                    rD[i]=2*(fr[i])*(
                    - ln(y[i]/mu[i]) + (y[i]-mu[i])/mu[i]);
                   else;
                    rD[i]=2*(ln(mu[i]) -1);  @p. 290 McN @
                   endif;
                  i=i+1;
                 endo;
               elseif q==5;  @ INVERSE GAUSSIAN LINK @
                  rD= ((y-mu).^2) ./ ((mu.^2).*y) ;
               else;
               endif;

               @  Dev= SCALED (RESIDUAL) deviance for current model @
                  Dev=sumc(rD);

                   z=th+(y-mu).*dm - off;      @ Taylor expansion @
                   W=sqrt((dm.^2).*V);   @ sqrt of weight vector n x 1 @
                   Xb=X./W;              @ weighted regressors n x p @
                        @ trap=1 therefore Cb is equal to missing
                                 or is equal to inverse @

                        Cb=invpd(Xb'Xb);

                    if ismiss(Cb)==1;
                        "Convergence not obtained";
                       notcv=1;
                      goto notconv;
                     else;
                   b_new=Cb*Xb'(z./W); @ new estimate of \beta @
                    endif;

                    if count==n;
                       notcv=1;
                      goto notconv;
                    else;
                      count=0;
                    endif;

                    if j>200 @ and cols(Xb)>1 @ ;
                       notcv=1;
                           cnt=cnt+1;       @ changed @
                         goto notconv;
                    else;
                    endif;

             if tol[2]==1;
                dev_new=dev;
             else;
                dev_new=b_new;
             endif;
  endo;
        notconv:

        /* MODEL CHECKING THROUGH RESIDUALS */
                  e=y-mu;    @ simple residuals @

          if nulm==1;
           @ nulm=1 implies a model with just the intercept @
               R[2,2]=sumc(rD);
               R[1,1]=1-R[1,2]/R[2,2];  @ R[1,1]= pseudo R^2 @
             goto fin;
          else;
          endif;

               @  Dev= SCALED (RESIDUAL) deviance for current model @
                  Dev=sumc(rD);
               @ Now we take the sign square root of the elements of rD @
                  i=1;  do until i==n+1;
                    if e[i] > 0;
                      rD[i]=sqrt(rd[i]);
                    else;
                      rD[i]=-sqrt(rd[i]);
                    endif;
                     i=i+1;
                   endo;

               @  rP= PEARSON RESIDUALS  eq. 2.11 p. 37 McN @
                  rP=e./sqrt(V);

                   @ Estimate of dispersion parameter
                     using moment estimator @
                     if df>0;
                              phi1=sumc(rP^2)/df;
                          if q>3 and q<=5;
                              phi=phi1;
                          else;
                              phi=1;
                              phi1=1;
                          endif;
                     else;
                        phi=miss(0,0);
                        phi1=phi;
                     endif;
               @ HAT MATRIX @
                  H=Xb*Cb*Xb';
                  Hi=diag(H);  @ Hi contains generalized leverage values @
             if maxc(Hi)<1;
               @  rSPR= Studentized Pearson Residuals ( 12.4 p. 397 McN) @
                  rSPR= e./  sqrt(phi * V.*(1-Hi));
               @  rSDR= Studentized Deviance Residuals (12.5 p. 397 McN) @
                  rSDR=rD./sqrt(phi *(1-Hi));
             else;
                 rSPR=miss(one,1);
                 rSDR=rSPR;
             endif;
               /* Store b_new in the first column of matrix B */
               B[.,1]=b_new;
                  if ismiss(Cb) /=1;
                    /* Standard error and t statistics
                    stored in matrix B */
                  @  Cb;
                    corrvc(Cb); wait;  @

                    B[.,2]=sqrt(phi*diag(Cb));
                    B[.,3]=b_new./B[.,2];
                    B[.,4]=2*cdftc(abs(B[.,3]),df);
                    /* pvalue of t ratios (2 sided test)  */
                   else;
                   endif;

                /* Store all the residuals  inside a matrix
                 1st col. simple residuals
                 2nd col. Pearson residuals
                 3rd col. deviance residuals
                 4th col. Studentized Pearson residuals
                 5th col. Studentized Deviance residuals   */
                 RS=real(e~rP~rD~rSPR~rSDR);
                 /* Store the deviance of the model
                 and the dispersion parameter phi1 in Matrix R */
                 R[1,2]=dev;
                 R[2,1]=phi1;
                 /* Store sqrt of weight vector */
                 WW=W;

            if nulm==0 and mint==1;
            @ Find the deviance of the NULL MODEL
              NULL MODEL contains only the intercept @
                X=X[.,1];
           @   one   /    sqrt(rows(X));    TO BE CHANGED    @
              b_new=b_new[1];
              b_old=b_new+1e+10;
              nulm=nulm+1;
              j=0;
              goto start;
            else;
            endif;
        fin:



   retp(B,RS,Hi,WW,R,notcv);
endp;

