/*
**  LDASIMPL.G  (simplified version of LDA.G)
**  written by MARCO RIANI
**
**  FORMAT                                 PURPOSE       Line
** =======================================================================
** {post1}=ldasimpl(Z,sz,prior);
**                                      To compute posterior
**                                      probabilities in
**                                      linear discr. analysis
**
*/


/*
**
** LDASIMPL.G
**
** PURPOSE:  to compute posterior probabilities in linear
**           discriminant analysis
**
** FORMAT:   {post1}=qdasimpl(Z,sz,prior);
**
** INPUT:     Z    = (n1+n2+...+ng) x p matrix containing
**                    observations from the g populations.
**                    Units forming matrix Z will be used
**                    to obtain quadratic discriminant rule.
**                    This is the so called "training set".
**            sz   = g x 1 (where g is the number of groups) vector
**                   which contains the sizes of the
**                   groups forming the training set
**                   e.g.: if sz =50|40|30
**                   then size of 1st gr. is 50
**                   then size of 2nd gr. is 40
**                   then size of 3rd gr. is 30
**            prior = g x 1 vector containing prior probabilities
**
** OUTPUT:   post1 = nn x 1 vector of posterior probabilities
**
** REMARK1:   Pooled unbiased variance covariance matrix is used
**
*/


proc (1)=ldasimpl(Z,sz,prior);
    local n,g,M,mdt,Vz,iVz,Lnd,big,ident,
    ln1,ln2,p,dist,edd,s,post,Post1,i,k,Id;
n=rows(Z); p=cols(Z);   @ Z = training set @
g=rows(sz);           @ ng = number of groups @
sz=0|sz;
    @  Matrix M contains (in the rows) the vector of the means for
       the groups
       first row refers to first group
       ...
       last row refers to the last group  @
    M=zeros(g,p);

    @  Matrix VZ contains the pooled varcov matrix @
    VZ=zeros(p,p);


    @ Post1= n x 1 matrix of posterior probabilities of each unit
      from its own universe @
    Post1=zeros(n,1);
           id=0;
           i=1; do until i==g+1;
           id=id|reshape(i,sz[i+1],1);
             i=i+1;
           endo;
          @ Ident Identifier vector  @
          Ident=id[2:n+1];
          @ Big= vector of big numbers @
          BIG=10^5*ones(n,1);

    @ Post=(n x g+2) matrix
      1st col contains the identifier vector
      2st col contains posterior prob. of belonging to gr. 1
      ...
      (g+1)th col contains posterior prob. of belonging to last gr.
      (g+2)th col contains to which group the unit has been assigned @
    Post=Ident~zeros(n,g)~BIG;

   @ Store means, var cov, inv(var cov) and ln(det(var cov))
     for each group @

  i=1; do until i==g+1;
    @ Matrix M contains the vector of the means for each group @
    M[i,.]=(meanc(Z[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]))';
    @ VZ is the pooled variance covariance matrix @
    VZ=VZ+vcx(Z[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.])*
         (sz[i+1]-1);
        i=i+1;
   endo;
     VZ=VZ/(n-g);
     iVZ=invpd(VZ);  @ iVZ= inverse of pooled var cov matrix @
     Lnd=ln(det(VZ)); @ Lnd= ln of the determinant of pooled var cov
                             matrix @
     dist=zeros(n,g);  @ Matrix of sample distances
   @ MDt=dist; @ Matrix of Mahalanobis distances from every population @

   i=1;  do until i==n+1;
          k=1; do until k==g+1;
              @ MD of each unit from every population @
              MDt[i,k]= (Z[i,.]-M[k,.])
             *iVz*((Z[i,.]-M[k,.])');

            dist[i,k]=0.5*(Lnd+ MDt[i,k])-ln(prior[k]);
              if dist[i,k]<post[i,g+2];
               post[i,g+2]=k;
              else;
              endif;
          k=k+1;
         endo;
    i=i+1;
   endo;
      @ Convert distances to posterior probabilities @
     edd=exp(-dist);
     s=sumc(edd');   @ denominator of posterior probabilities @
     post[.,2:(g+1)]=edd./s;
      @  rows correspond to units
         cols correspond to groups
    e.g. post[r,(s+1)]= posterior prob. of unit r of belonging to
                        group s @


i=1; do until i==n+1;
   post1[i]=post[i,ident[i]+1];  @ post1= nn x 1 vector of posterior
                                  prob.
                                 of each
                                 unit from its own universe @
   i=i+1;
endo;

  retp(post1);
 endp;

