/*
**  LIKLA
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
**
**  {Tl,lik} =likla(X,y,la)         To calculate the likelihood and
**                                  the score test
**                                  for different values of the
**                                  transformation parameter lambda
**
*/


/*
**
** LIKLA
**
** PURPOSE:  to calculate the likelihod and the score test for different
**           values of the tranformation parameter la
**
**
**
** FORMAT:
**
** {Tl,lik} = likla(X,y,la)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             la = k x 1  vector which specifies for which values
**                  of the transformation parameter lambda
**                  you want to compute the likelihood and the score
**                  test
**
** OUTPUT:     Tl = k x 1 vector which contains the values of
**                  the score test for each element of vector lambda
**            lik = k x 1 vector which contains the value of the
**                  likelihood for each element of
**                  vector lambda
**
**  REMARK    this program does not add a column of 1 in matrix X to
**            create the intercept.
**            lik is 2* Lmax
**            if Sz (see Atkinson, 1985 pag. 98) is <0.000000001
**            Gauss inserts a missing value for the score test)
**
*/

 proc (2)=likla(X,y,la);
    local n,p,C,G,A,nla,Lik,Convar,Tl,i,z,w,r,zAw,wAw,Sz,fin;

 n=rows(X);
 p=cols(X);
 nla=rows(la);
 C=X/maxc(maxc(X));
  if det(C'C)<1e-50;
   " WARNING: Matrix X is not full rank ";
   " Score and Likelihood are missing values";
   goto fin;
  else;
  endif;
 A=eye(n)-X*invpd(X'X)*X';
 Lik=zeros(nla,1);
  @ L is a vector which contains the likelihoods for diff. values of la @
  @ w= Vector which contains the constructed variable @
 Convar=zeros(n,nla);  @ Convar = Matrix of constructed variables @
 Tl=Lik;  @ Tl= vector which contains the t test for constructed
               variables @
 if n > p and minc(y)>1e-12;
    G=exp(meanc(ln(y)));  @ Geometric mean of the observations @

     i=1; do until i==nla+1;
     if abs(la[i])<0.0001;
        z=G*ln(y);
        w=G*ln(y).*(ln(y)/2-ln(G));
         else;
      z=(y.^la[i]-1)/(la[i]*G^(la[i]-1));
      w=(y^la[i].*ln(y)-(y^la[i]-1)*(1/la[i]+ln(G))) /(la[i]*G^(la[i]-1));
         endif;

            r=z'A*z;
            zAw=z'A*w;   wAw=w'A*w;
            if abs(wAw) <1e-12;
             Tl[i]=miss(0,0);
              goto fin;
            else;
             Sz=sqrt(r-zAw^2/wAw); @ See Atkinson (1985) p. 98 @
            endif;

     if ABS(real(Sz)) < 0.0001;
            Tl[i]=miss(0,0);
       else;
            Tl[i]=-zAw*sqrt(n-p-1)/(Sz*sqrt(wAw));
       endif;

        if r==0;
          Lik[i]=miss(0,0);
        else;
          Lik[i]= -n*ln(r/n);
        endif;
                i=i+1;
           endo;
 else;
  @   "Minimum value of y is smaller than 0 or n=p";
     "Tl and Lik in procedure LIKLA are missing values";  @
          fin:
       Lik=miss(0,0);
       Tl=miss(0,0);
 endif;

 retp(Tl,Lik);
endp;

