/*
**  LIKLABS
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
**
**  {Tl,lik} =liklaBS(X,y,la)       To calculate the likelihood and
**                                  the score test
**                                  for different values of the
**                                  transformation parameter lambda
**                                  Both sides (y and fitted values) 
**                                  are transformed with the same lambda
**                                  (There is just one
**                                  explanatory variable)
**
**
*/


/*
**
** LIKLABS
**
** PURPOSE:  to calculate the likelihood and the score test for
**           different
**           values of the tranformation parameter la
**           Both sides (y and fitted values are transformed)
**
**
** FORMAT:
**
** {Tl,lik} = liklaBS(X,y,la)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             la = k x 1  vector which specifies for which values
**                  of the transformation parameter lambda
**                  you want to compute the likelihood and the score
**                  test
**
** OUTPUT:     Tl = k x 1 vector which contains the values of
**                  the score test for each element of vector lambda
**            lik = k x 1 vector which contains the value of the
**                  likelihood for each element of
**                  vector lambda
**
**  REMARK    this program does not add a column of 1 in matrix X to
**            create the intercept.
**            lik is 2* Lmax
**
*/

 proc (2)=liklabs(X,y,la);
    local n,p,C,G,A,nla,Lik,Tl,i,z,r,convar,x1,x2,
          B,Sres,R1,NO,RA,SC,q,u,w,fin,qhat;

 n=rows(X);
 p=cols(X);
 nla=rows(la);
 C=X/maxc(maxc(X));
  if det(C'C)<1e-50;
   " WARNING: Matrix X is not full rank ";
   " Score and Likelihood are missing values";
   goto fin;
  else;
  endif;
 Lik=zeros(nla,1);
  @ L is a vector which contains the likelihoods for diff. values of la @
  @ w= Vector which contains the constructed variable @
 Convar=zeros(n,nla);  @ Convar = Matrix of constructed variables @
 Tl=Lik;  @ Tl= vector which contains the t test for constructed
               variables @
 if n > p and minc(y)>1e-12;
    G=exp(meanc(ln(y)));  @ Geometric mean of the observations @

     i=1; do until i==nla+1;
     if abs(la[i])<1e-6;
      q=ln(y);
      u=ln(X);
      qhat=meanc(q-u)+u;
      w=0.5*(qhat^2-q^2)-(qhat-q)*ln(G);
      X1=ones(n,1)~w;
          x2=x1/maxc(maxc(x1));
         if det(x2'x2)<1e-10;
          goto fin;
         else;
         endif;

      {B,Sres,R1,NO,RA,SC}=regressi(q-u,X1,0,0);
      r=moment(q-qhat,0)*(G^2);

     else;

      q=y^(la[i]);
      u=X^(la[i]);
      qhat=u*u'*q/(u'u);
      w=(qhat.*ln(qhat)-q.*ln(q))/la[i]-(qhat-q)*(1/la[i]+ln(G));
      X1=u~w;
          x2=x1/maxc(maxc(x1));
         if det(x2'x2)<1e-10;
          goto fin;
         else;
         endif;

      {B,Sres,R1,NO,RA,SC}=regressi(q,X1,0,0);
      r=moment(q-qhat,0)/((la[i]*G^(la[i]-1))^2);

    /****
      qstar=q-qhat;
      wstar=w-u*u'*w/(u'u);
      _plctrl=-1;
      xy(wstar,qstar);
     ***/
     endif;


        if ismiss(B)==1;
            Tl[i]=miss(0,0);
       else;
            Tl[i]=B[2,3];
       endif;

        if r==0;
          Lik[i]=miss(0,0);
        else;
          Lik[i]= -n*ln(r/n);
        endif;

                i=i+1;
           endo;
 else;
     "Minimum value of y is smaller than 0 or n=p";
     "Tl and Lik in procedure LIKLA are missing values";
          fin:
       Lik=miss(0,0);
       Tl=miss(0,0);
 endif;

 retp(Tl,Lik);
endp;

