/*
**  LIKLAG
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
**
**  {Sc,Sc1} =liklag(X,y,la,d)        To calculate the score test
**                                    in correspondence of different
**                                    values of the transformation
**                                    parameter lambda.
**                                    Both response and explanatory
**                                    can be transformed
**
*/


/*
**
** LIKLAG
**
** PURPOSE:  to calculate the score test for different
**           values of the transformation parameter la
**           In this program you can transform both y and
**           one explanatory variable at the same time
**
** FORMAT:
**
** {Sc,Sc1} = liklag(X,y,la,d)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             la = k x 1  vector which specifies for which values
**                  of the transformation parameter \lambda
**                  you want to compute the score test
**              d = dummy variable which specifies the two
**                  variables which potentially need to
**                  be transformed
**                  if d = 2 x 1 vector the program tests for
**                        transformations for the columns specified
**                        in vector d;
**                        Example: d=0|1; response and first column
**                        of matrix X are subject to inquiry
**                        Example: d=2|3 second and third column
**                        of matrix X are analyzed
**
** OUTPUT:      Sc = k x k matrix which contains the values of the
**                  score test for the different values of the
**                  transformation parameter specified in la
**                  Sc refers to the t test for the first
**                  transformed variable
**             Sc1 = k x k matrix which contains the values of
**                  the score test for the transformation
**
**  REMARK    this program does not add a column of 1 in matrix X to
**            create the intercept.
**            if Sz (see Atkinson, 1985 pag. 98) is <0.000000001
**            Gauss inserts a missing value for the score test)
**           Example d=0,2; lambda=0.2|0.8
**                             sc = 2 x 2
**                    X[.,2]^0.2         X[.,2]^0.8
**         y^0.2     {   ccccc               cccc  }
**  sc=              {                             }
**         y^0.8     {   cccc                ccc   }
**
*/

 proc (2)=liklag(X,y,la,d);
    local n,p,pp,q,b,Id,j,C,G,A,nla,la1,X1,X2,
          w1,Xcon,Sc,Sc1,i,jj,jk,z,w,cw,r,zAw,wAw,Sz,dd,tr;

n=rows(X);

G=exp(meanc(ln(y)));  @ Geometric mean of th observations @
C=invpd(X'X)*X';
b=C*y;    @ ols on the original matrix X @
Id=eye(n);
nla=rows(la);
Sc=zeros(nla,nla);
Sc1=Sc;
  @ w= Vector (matrix) which contains the constructed variable(s) @

i=1; do until i==nla+1;

   if d[1]==0; @ if the response must be tranformed @
        pp=1; @ cols of the matrix of constructed variables @
        dd=d[2]; @ scalar which tells which X col. must be transformed @
                          if abs(la[i])<0.0001;
     z=G*ln(y);      @ transform response @
     w=G*ln(y).*(ln(y)/2-ln(G)); @ constructed variables for response @
                           else;
     z=(y.^la[i]-1)/(la[i]*G^(la[i]-1));
     w=(y^la[i].*ln(y)-(y^la[i]-1)*(1/la[i]+ln(G)))/(la[i]*G^(la[i]-1));
                           endif;
   else;
       z=y; @ do not transform the response variable @
       pp=rows(d); @ cols of matrix of constructed variables @
       dd=d; @ vector which tells which X cols. must be transformed @
   endif;

       /* create the constructed variables for matrix X */
 Xcon=zeros(n,pp);
 X1=X;

 j=1; do until j==nla+1;
                   la1=la[j];
       jj=1; do until jj==pp+1;
   Xcon[.,jj]=b[dd[jj]]*(X[.,dd[jj]]).^(la1)  .* ln(X[.,dd[jj]]);
   X1[.,dd[jj]]=X[.,dd[jj]]^la1;  @ modify the cols of X
                                  which correspond to the transf. @
                   la1=la[i];
         jj=jj+1;
       endo;

         if d[1]==0;  @ if the response has been transformed @
    w1=w~Xcon;
   @ the first column of Matrix w contains the constructed
             variable corresponding to y the second column the
             constructed variable corresponding to the jth column of X @
                          @ if la[i]=la[j] there is just one
                             constructed variable @
                                   if la[i]==la[j];
                                       w1=-w1[.,1]+w1[.,2];
                                    else;
                                    endif;
         else;
 w1=Xcon;    @ This matrix may have 1 or 2 columns corresponding to the
              number of explanatory varibles which must be transformed @
                                   if la[i]==la[j];
                                        w1=sumc(Xcon');
                                    else;
                                    endif;
         endif;
  @  "w when lambda=0.8 and 1 ";
      w1[1:10,.]; wait;     @

                    cw=cols(w1);  @ cw may be 1 or 2 @
              jk=cw;
      q=1; do until q==cw+1;
                   if cw==2;   @ if there are 2 constructed variables @
                X2=X1~w1[.,jk];
         @ In the first step of the loop we construct a t test for
           w1[.,1] when w1[.,2] is already in the model. In the second
           step we construct a t test for w1[.,2] when w1[.,1] is
           already in the model @
                    else;
                X2=X1;
                    endif;
                A=Id-X2*invpd(X2'X2)*X2';
                r=z'A*z;
                zAw=z'A*w1[.,q];   wAw=w1[.,q]'A*w1[.,q];
                Sz=sqrt((r-zAw^2/wAw)/(n-cols(X2)-1));

                @ n-cols(X2)-1 is the number of columns of matrix X @
                @ Score test for first const. var. @
                   if real(Sz) < 0.000000001;
                    tr=miss(0,0);
                   else;
                     tr=zAw/(Sz*sqrt(wAw));
                   endif;
                if q==1;
                Sc[i,j]=tr;
                       if la[i] ne la[j] and d[1]==0;
                           Sc[i,j]=-Sc[i,j];
                        else;
                        endif;
                else;
                Sc1[i,j]=tr;
                endif;
      jk=jk-1; @ Now take the first column of Mat. of the constr. var. @
          q=q+1;
      endo;
    j=j+1;
    endo;
   i=i+1;
  endo;
 retp(Sc,Sc1);
 endp;

