/*
**  LMSBS
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
**
**  {mE,BB,BS} =lmsBS(X,y,la,d)         To calculate the least median
**                                      of squares when 
**                                      both sides of a model are transformed
*/


/*
**
** LMS
**
** PURPOSE:  to calculate the least median of squares
**           See Rousseuuw (1984) JASA
**           or the least trimmed squares when both sides are transformed
**
**
** FORMAT:
**
** {mE,BB,BS} = lmBSs(X,y,la,d)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             la = nla x 1  vector which specifies for which values
**                  of the transformation parameter \lambda
**                  you want to compute LMS
**             d = scalar or 2 x 1 vector
**                 If d is a scalar LMS is calculated
**                    d specifies the number of
**                    subsamples to be used in the calculation
**                    of the LMS
**                    (Remark: d=1 ALL SUBSAMPLES ARE EXTRACTED)
**                 If d is a 2 x 1 vector LTS is calculated
**                    d[1] specifies the number of
**                    subsamples to be used in the calculation
**                    of the LTS least trimmed squares
**                    (Remark: d=1 ALL SUBSAMPLES ARE EXTRACTED)
**
** OUTPUT:     mE = n x nla matrix which contains the standardized
**                  residuals. Each column corresponds to
**                  a particular element of vector lambda
**                  If d is a scalar the LMS residuals are standardized
**                     using the sqrt of the Least Median of Squares
**                  If d is a 2 x 1 vector the LTS res. are standardized
**                     using the sqrt of the Least Trimmed Squares
**             BB = p x 2 matrix. In the first col there is vector
**                  la in the 2nd col we can find the value of the
**                  least median of squares (trimmed squares) for each
**                  element of  vector la
**             BS = p x nla matrix which contains the units forming
**                  the best subset for each element of vector la
**
**     See  also LMS,LMSG,ROB
**
**
**  REMARK2:  Vector y is transformed using Box Cox
**
**  REMARK3:  If the value of the LMS (LTS) is equal to zero
**            the residuals in matrix mE are not standardized
**
**  REMARK4: If the minimum value of y is \leq 0 and la is a scalar
**           equal to 1 then we do not use Box-Cox transformation
**           for vector y
**
*/


proc (3)=lmsBS(X,y,la,d);
   local n,k,nss,s1,sub,s,x1,y1,j,b,i,med,mlms,slms,jj,nla,mE,ny,e,
         ij, mSt,BB,BS,nX;
n=rows(x);
k=2;  @ dimension of the elemental subsets @
nla=rows(la);
mE=zeros(n,nla);
BB=la~zeros(nla,1);
BS=zeros(k,nla);
ij=0; @ ij is a dummy index linked to rhe storage of the best sample @
 if d[1]==1;
   {nss}=elms(n,k); @ nss contains in the rows all possible subsamples @
   sub=rows(nss); @ number of subsamples used in the calculation of LMS @
 else;
        s1=zeros(k,d[1]);
     sub=d[1];
 endif;
 jj=1; do until jj==nla+1;
    @ Do not use Box Cox if the minimum value of y is smaller than 0 @
    if minc(y)<=0 and la[jj]/= 1 ;
         errorlog "The minimum value for vector y is smaller than 0";
                   "and the transformation parameter you supplied
                   is different from 1.";
          end;
    elseif minc(y)<=0 and la[jj]==1;
     ny=y;
    else;
               if abs(la[jj])<0.0001;
      ny=ln(y);
      nX=ln(X);
        else;
      ny= y.^la[jj];
      nX= X.^la[jj];
        endif;
    endif;

    i=1; do until i==sub+1;
                    ij=ij+1; @  ij=dummy index @
            start:
         if d[1]==1;
             s=(nss[i,.]');
         else;
                good:
             s=without(k,n);
             s=sortc(s,1);
         endif;



       x1=nX[s,.];
       y1=ny[s,.];
           "extracting subsample";; i;
          @  s; @
         if  abs(det(x1'x1)) < 1e-20 and i < sub;
                       if d[1]==1;
                          i=i+1;
                         else;
                       endif;
                goto start;
          elseif abs(det(x1'x1)) < 1e-4 and i==sub;
                goto end;
           else;
          endif;
        if d[1] ne 1;
            j=1; do while j<i;
        @ this if checks whether the subsample had already been extracted @
                 if sumc(s.==s1[.,j])==k;
                    goto start;
                 else;
                 endif;
               j=j+1;
             endo;
            s1[.,i]=s;
          else;
         endif;


       b=invpd(x1'x1) *x1'y1;
       e=sortc((ny-nX*b).^2,1);
          @ If rows(d) > 1 compute the LTS @
          if rows(d)>1;
              @ med=mean of the first [n/2]+1 residulas @
            @  med=meanc(e[1:floor(0.5*n)+1]); @
              med=meanc(e[1:floor(0.5*n)+floor(0.5*(k+1)) ]);
          else; @ if d is a scalar LMS is computed @
              @  e=e[k+1:n];   @
              @ med=median of the squared residuals different from zero @
              @ med=e[round(rows(e)/2)]; @
              med=e[trunc(0.5*(n+k+1))];
          endif;

         if ij==1;
           MST=med;
         else;
         endif;


       if med <= mst;
          @ <= is necessary because the best could be the first @
          mst=med;
          Mlms=med; @ "Least median of squares is";; med;  wait; @
          Slms=s;   @ Slms contains the best elemental subset @
       else;
       endif;
           cls;
     i=i+1;
    endo;
            end:
    @  For the best subset compute the standardized lms or lts @
    x1=nX[Slms,.];
    y1=ny[Slms,.];

    b=invpd(x1'x1) *x1'y1;
     @    Mlms; wait;  @
    if Mlms>0;
      e=(ny-nX*b)/ sqrt(Mlms);
    else;
     "LMS value equal to zero";
     e=(ny-nX*b);
    endif;

    @ Matrix mE contains the residuals from lms
                each column correspond to a value of trasf. parameter @
    mE[.,jj]=e;
    @ Store value for least median of squares @
    BB[jj,2]=Mlms;
    @ Store the indexes of the units forming the best subsample @
    BS[.,jj]=Slms;
        ij=0; @ ij is reset equal to 0 @
  jj=jj+1;
 endo;

 retp(mE,BB,BS);
endp;

