/*
**  LMSG
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
**
** {mE} =lmsg(X,y,la,f,d)               To calculate the least median
**                                      of squares  when both response
**                                      and explanatory have been
**                                      transformed
**
*/


/*
**
** LMSG
**
** PURPOSE:  to calculate the least median of squares on transformed
**           data.
**           See Rousseuuw (1984) JASA
**
**
**
** FORMAT:
**
** {mE} = lmsg(X,y,la,f,d)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             la = k x 1  vector which specifies for which values
**                  of the transformation parameter \lambda
**                  you want to compute  LMS
**             f =  scalar which specifies the number of samples
**                  to be used in the calculation of the LMS
**                  IF f=1 ALL SUBSAMPLES ARE EXTRACTED
**             d =  dummy variable which specifies the two
**                  variables which need to
**                  be transformed. (LMS will be done on the transformed
**                  variables)
**                        Example: d=0|1; response and first column
**                        of matrix X are transformed
**                        Example: d=2|3 second and third column
**                        of matrix X are transformed
**
** OUTPUT:     mE = n x k^2 matrix which contains the standardized
**                  residuals. Each column corresponds to
**                  a particular combination of values of the
**                  vector lambda
**                  For example if la=0.2|0.5   d=0|1
**                  mE[.,1] res. from y^0.2  X[.,1]^0.2
**                  mE[.,2] res. from y^0.2  X[.,1]^0.5
**                  mE[.,3] res. from y^0.5  X[.,1]^0.2
**                  mE[.,4] res. from y^0.5  X[.,1]^0.5
**
**  REMARK    this program does not add a column of 1 in matrix X to
**            create the intercept.
**
** WARNING    DO NOT SPECIFY 0 AS A TRANSFORMATION FOR THE
**              EXPLANATORY
*/


proc (1)=lmsg(X,y,la,f,d);
   local n,k,G,nss,s1,sub,s,x1,y1,j,b,i,med,mlms,slms,jj,nla,mE,ny,e,
         ii,ij, mSt,nX,dd,pp;

n=rows(x);
k=cols(x);  @ dimension of the elemental subsets @
nla=rows(la);
G=exp(meanc(ln(y)));  @ Geometric mean of the observations @
mE=zeros(n,nla^2);



                 ij=0; @ ij is a dummy index linked to the storage of
                         the best sample @
 if f==1;
   {nss}=elms(n,k); @ nss contains in the rows all possible subsamples @
   sub=rows(nss); @ number of subsamples used in the calculation of LMS @
 else;
        s1=zeros(k,f);  @ Matrix s1 will contain (in the column) all the
                          subsamples @
     sub=f;
 endif;


   nX=X; @ Matrix nX will contain the transformed variables @
         @ The columns of matrix X which must be transformed are
           always the same. @
 ii=1; do until ii==nla+1;
                      @ TRANSFORM THE RESPONSE @
       if d[1]==0; @ if the response must be transformed @
               pp=1; @ cols of the matrix of constructed variables @
               dd=d[2]; @ scalar which tells which X col. must be transformed @
                                 if abs(la[ii])<0.0001;
            ny=G*ln(y);
                                  else;
            ny=(y.^la[ii]-1)/(la[ii]*G^(la[ii]-1));
                                  endif;
       else;
              ny=y; @ do not transform the response variable @
              pp=rows(d); @ cols of matrix of constructed variables @
              dd=d; @ vector which tells which X cols. must be transformed @
       endif;
                       @ TRANSFORM EXPLANATORY @
  jj=1; do until jj==nla+1;
                     if pp==2;
                         nX[.,dd[1]]=X[.,dd[1]]^la[ii];
                         nX[.,dd[2]]=X[.,dd[2]]^la[jj];
                     else;
                        nX[.,dd]=X[.,dd]^la[jj];
                     endif;

    i=1; do until i==sub+1;
    "Examining subsample number";; i;

                     ij=ij+1; @  ij=dummy index @
             start:
          if f==1;
                  s=(nss[i,.]');
            else;
                  s=ceil(n*rndu(k,1));
           endif;

        x1=nX[s,.];  @ x1 = subset of transformed X matrix @
        y1=ny[s,.];  @ y1 = subset of transfomed y variable @
         if ABS(det(x1))<0.0000001 and i < sub;
             @    "found det==0"; wait:     @
                   @ if f=1 before going to start increase i by 1. In
                     this case you can examine another row of matrix
                     nss which contains all the possible subsamples @
                        if f==1;
                           i=i+1;
                          else;
                        endif;
                 goto start;
          elseif ABS(det(x1)) < 0.000001 and i==sub;
                 goto end;
            else;
          endif;

        if f ne 1;
             j=1; do while j<i;
       @ this if checks whether the subsample had already been extracted @
                  if sumc(s.==s1[.,j])==k;
                     goto start;
                  else;
                     s1[.,i]=s;
                  endif;
             j=j+1;    endo;
         else;
       endif;


       b=invpd(x1'x1) *x1'y1;    @ ols on transformed subset @
       e=sortc((ny-nX*b).^2,1);  @ res. on overall tranf. var. @
       e=e[k+1:n];
               @ med=median of the squared residuals different from zero @
       med=e[round(rows(e)/2)];
       if ij==1;
         MST=med;
          else;
          endif;

        if med <= mst;
         @ <= is necessary because the best could be the first @
       mst=med;
       Mlms=med; @ "Least median of squares is";; med;   @
       Slms=s;   @ Slms contains the best elemental subset @
       else;
            endif;
             cls;
     i=i+1;
    endo;

                   end:
        @  For the best subset compute the standardized lms @
        x1=nX[Slms,.];
        y1=ny[Slms,.];

       b=invpd(x1'x1) *x1'y1;    @ b on best subset of transf. var. @
       e=(ny-nX*b)/sqrt(Mlms);  @ standardized residuals from transf.
                                  variables @
       mE[.,jj+nla*(ii-1)]=e;
           @ Matrix mE contains the residuals from lms
             each column correspond to a combination of the values of
             trasf. parameter @
   jj=jj+1;
   endo;
 ii=ii+1;
endo;
      retp(mE);
  endp;

