/*
**  LMSGLM
**  written by MARCO RIANI
**
**  FORMAT                                      PURPOSE           Line
** =======================================================================
**
** {mE,BB,BS,cntr}=lmsglm(X,y,q,fr,d,k,mint,sh,tol,bin,off)
**                                              To calculate the
**                                              least
**                                              median of squares for
**                                              generalized linear
**                                              models
*/


/*
**
** LMSGLM
**
** PURPOSE:  to calculate the least median of squares
**           in generalized linear models
**
**
**
** FORMAT:
**
** {mE,BB,BS,cntr} = lmsglm(X,y,q,fr,d,k,mint,sh,tol,bin,off)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             q  = scalar which specifies the kind of GLM
**                  q=1 ==> Poisson link
**                  q=2 ==> Logit link
**                  q=3 ==> Complementary Log log link
**                  q=4 ==> Gamma link
**                  q=5 ==> Inverse Gaussian link
**                  q=6 ==> Probit link
**                  q=7 ==> Arc sin link
**                  q=8 ==> Modified logit link
**             fr = n x 1 vector of frequencies
**                  if fr = scalar 0 a n x 1 vector of 1s
**                  is created automatically by the program
**                 to be used in the calculation of the LMS
**             d = scalar which specifies the number of subsamples
**                 IF d=1 ALL SUBSAMPLES ARE EXTRACTED
**             k = scalar denoting the dimension of the
**                 subsamples to be extracted
**           mint= scalar. if mint=1 the null
**                 model is also considered
**           sh  = scalar which specifies the kind of residuals used
**                 in least median of squares calculation
**                  sh=1 ==> simple residuals
**                  sh=2 ==> Pearson residuals
**                  sh=3 ==> Deviance residuals
**                  sh=4 ==> Studentized Pearson residuals
**                        McN p. 397 eq. (12.4)
**                  sh=5 ==> Studentized Deviance residuals
**                        McN p. 397  eq. (12.5)
**           tol = tolerance level to be used in the maximization
**           bin = 2 x 1 vector. This option is useful when
**                 dealing
**                 with binary data
**                 if bin[1] <> 0 and maxc(fr)==1 program takes into
**                 account the size of y.==0
**                 an of y.==1. In this case, therefore the lms
**                 is the maximum of the two
**                 separate lms
**                 if bin[2] =  threshold value.
**                   The subsamples whose lms
**                   is smaller than bin[2] are not considered
**                 REMARK: the lms is calculated skipping the
**                 units forming bsb, because deviance for binary
**                 data gives no information in terms of
**                 goodness of fit
**           off = n x 1 vector which defines an offset
**                 REMARK: off = scalar 0 implies no offset
**
** OUTPUT:     mE = n x 1 matrix which contains the
**                  residuals specified in scalar sh for all the units
**             BB = scalar which contains the value of the
**                  least median of squares
**             BS = p x 1 matrix which contains the units forming
**                  the best subset
**          cntr  = 2 x 1 vector
**                  cntr[1]= counts the number of iterations
**                           in which there has not been convergence
**                  cntr[2]= 1 there is not conv. in the lms subset
**                           0 there is conv. in the lms subset
**
**
**  REMARK1    this program does not add a column of 1 in matrix X to
**             create the intercept.
**
**  REMARK2    This programs calls routine glm.g
**             on order to fit the generalized linear model
**             an routine predglm.g for prediction
*/


proc (4)=lmsglm(X,y,q,fr,d,k,mint,sh,tol,bin,off);
   local n,p,nss,s1,sub,s,x1,y1,j,b,i,med,mlms,slms,mE,e,
         ij,mSt,BB,BS,Rs,Hi,R,fr1,W,cnt,cntr,seq,eq_0,eq_1,
         e0,e1,dif0,dif1,med0,med1,off1;
n=rows(x);
p=cols(x);
mE=zeros(n,1);
BS=zeros(k,1);
ij=0; @ ij is a dummy index linked to the storage of the best sample @
 if d==1;
   {nss}=elms(n,k); @ nss contains in the rows all possible subsamples @
   sub=rows(nss); @ number of subsamples used in the calculation of LMS @
 else;
        s1=zeros(k,d);
     sub=d;
 endif;

 if bin[1] /=0;
    seq=seqa(1,1,n);
    eq_0=selif(seq,y.==0);
    eq_1=selif(seq,y.==1);
 else;
 endif;

 cntr={0,0}; @ cntr= counter for the subsamples in which convergence
                 was not obtained @
  i=1; do until i==sub+1;
              ij=ij+1; @  ij=dummy index @
      start:
   if d==1;
        s=(nss[i,.]');
   else;
          good:
        s=ceil(n*rndu(k,1));
              @ Extract without replacement @
              if rows(unique(s,1))<k;
                 goto good;
              else;
              endif;
        s=sortc(s,1);
    endif;



 x1=X[s,.];
 y1=y[s];
 fr1=fr[s];
 if off ne 0;
   off1=off[s];
 else;
   off1=0;
 endif;
     "extracting subsample";; i;
      @  s; @
   if  (rank(x1)<p or rows(unique(y1,1))==1) and i < sub;
                 if d==1;
                    i=i+1;
                   else;
                 endif;
          goto start;
    elseif (rank(x1)<p or rows(unique(y1,1))==1) and i==sub;
          goto end;
     else;
    endif;
  if d ne 1;
      j=1; do while j<i;
  @ this if checks whether the subsample had already been extracted @
           if sumc(s.==s1[.,j])==k;
              goto start;
           else;
           endif;
         j=j+1;
       endo;
      s1[.,i]=s;
    else;
   endif;
@ Use Procedure glm to estimate beta coefficients @
  {B,RS,Hi,W,R,cnt}=glm(X1,y1,q,fr1,mint,tol,0,off1);

  cntr[1]=cntr[1]+cnt; @ count the number of subsamples in
                         which ILS did not converge @
   if cnt==1;
      med=1e+7;
   goto skip;
   else;
   endif;
@ Use Procudere predglm glm proc.
  to compute residuals for the overall sample @

  {RS}=predglm(X,y,q,fr,B[.,1],off);

  if bin[1]==0 or maxc(fr)>1;
    e=sortc(RS[.,sh].^2,1);    @ sort squared glm residuals @
    e=e[p+1:n];
    @ med=median of the squared deviance residuals different from zero @
    med=e[round(rows(e)/2)];
  else;
                @   e0=sortc(RS[eq_0,sh].^2,1);  Sort all units @
   dif0=setdif(eq_0,s,1);   @ select units of univ 0 belonging to ncl @
    e0=sortc(RS[dif0,sh].^2,1);
    med0=e0[round(rows(e0)/2)];
                @  e1=sortc(RS[eq_1,sh].^2,1); Sort all units @
   dif1=setdif(eq_1,s,1);   @ select units of univ 1 belonging to ncl @
    e1=sortc(RS[dif1,sh].^2,1);
    med1=e1[round(rows(e1)/2)];
    med=maxc(med0|med1);
       @ bin[2] specifies whether to use or not the threshold @
       if minc(e0)<bin[2] or minc(e1)<bin[2];
       @  minc(e0); minc(e1); wait;  @
         med=10^6;
       else;
         sglobal=sglobal~s;
         mglobal=mglobal~med;
       endif;
  endif;

       skip:
if ij==1;
  MST=med;
   else;
   endif;


 if med <= mst;
  @ <= is necessary because the best could be the first @
mst=med;
Mlms=med; @ "Least median of squares is";; med;  wait; @
Slms=s;   @ Slms contains the best elemental subset @
 "Best elemental subset";;
    Slms;
else;
     endif;
        @   cls; @

       i=i+1;
    endo;
            end:
 @  For the best subset compute the standardized lms @
 X1=X[Slms,.];
 y1=y[Slms];
 fr1=fr[Slms];
 if off ne 0;
   off1=off[Slms];
 else;
   off1=0;
 endif;

{B,RS,Hi,W,R,cntr[2]}=glm(X1,y1,q,fr1,mint,tol,0,off1);
  {RS}=predglm(X,y,q,fr,B[.,1],off);
  @ "Rs"; Rs;   cntr[2];  wait;   @
b=B[.,1];
ME=RS[.,sh];    @ ME stores (sh) residuals @

 @ Store value for least median of squares @
 BB=Mlms;
 @ Store the indeces of the units forming the best subsample @
 BS=Slms;

      retp(mE,BB,BS,cntr);
  endp;

