/*
**  LMSNLS
**  written by MARCO RIANI
**
**  FORMAT                                      PURPOSE           Line
** =======================================================================
**
** {mE,Mlms,Slms,cntr} = lmsnls(b0,dd,d,k,sh)
**                                              To calculate the
**                                              least
**                                              median of squares for
**                                              non linear regression
**                                              models
**
*/


/*
**
** LMSNLS
**
** PURPOSE:  to calculate the least median of squares
**           for non linear regression models
**
**
**
** FORMAT:
**
** {mE,Mlms,Slms,cntr} = lmsnls(b0,dd,d,k,sh)
**
** INPUT:      b0 = r x 1  vector which specifies the starting
**                  values for beta coeffcients and other
**                  parameters (see scalar dd below)
**             dd = scalar which specifies the method of
**                  minimization
**                  if  dd==1 we use unweighted minimization
**                    see Seber and Wild eq. 2.166 with w_i=1
**                    \sum_{i=1}^n [y_i-f(\x_i,\theta)]^2
**                    In this case we estimate only the \theta
**                    (regression) coefficients
**                  elseif dd==2 we use direct minimization
**                    see Seber and Wild eq. 2.166
**                    \sum_{i=1}^n \ln w_i-0.5n w_i[y_i-f(\x_i,\theta)]^2
**                    where w_i=f(\x_i,\theta)^(2 \phi-2)
**                    In this case we estimate both the \theta
**                    (regression) coefficients and transformation
**                    parameter phi (lambda)
**                  else we use Power transform for both sides
**                    see Seber and Wild eq. 2.148
**                    In this case we estimate both the \theta
**                    (regression) coefficients, the transformation
**                    parameter phi (lambda) and \sigma^2
**             d  = scalar which specifies the number of subsamples
**                  IF d=1 ALL SUBSAMPLES ARE EXTRACTED
**             k  = scalar denoting the dimension of the
**                  subsamples to be extracted
**            sh  = scalar which specifies the kind of residuals used
**                  in least median of squares calculation
**                   sh=1 ==> simple residuals
**                   sh=2 ==> studentized residuals
**                   sh=3 ==> projected residuals
**                   else ==> studentized projected residuals
**
** OUTPUT:     mE = n x 4 matrix which contains the residuals
**                  for all the units
**                  1st col = simple residuals
**                  2nd col = studentized residuals
**                  3rd col = projected residuals
**                  3rd col = studentized projected residuals
**             BB = scalar which contains the value of the
**                  least median of squares
**             BS = k x 1 matrix which contains the units forming
**                  the best subset
**          cntr  = 2 x 1 vector
**                  cntr[1]= counts the number of iterations
**                           in which there has not been convergence
**                  cntr[2]= 1 there is not conv. in the lms subset
**                           0 there is conv. in the lms subset
**
**      REMARK 1      THE MAIN FILE MUST CONTAIN 2 GLOBAL
**                    VARIABLES
**                    yglobal  = n x 1 vector containing response
**                    Xglobal  = n x p matrix containing expl. variables
**                    YOU MUST ALSO CREATE A GAUSS PROC WHOSE NAME
**                    MUST BE funcnls WHICH SPECIFIES THE
**                    LINK BETWEEN X
**                    AND THE PARAMETERS VECTOR b
**                    For example for the Michaelis Menten model
**                    (Seber and Wild p. 148 @
**                      proc (1)=funcnls(psi);
**                          local X;
**                          X=Xglobal;
**                          retp(  (psi[1]*X)./(psi[2]+X) );
**                      endp;
**
**
**  REMARK2    This programs calls routine nls.g
**             on order to fit the non linear model
**             and routine prednls.g for prediction
**
**
**  REMARK3    If k=cols(X) we take the median of n-cols(X) units
**             if k>cols(X) we take the median of all the n units
**
**
*/


proc (4)=lmsnls(b0,dd,d,k,sh);
   local X,y,n,p,nss,s1,sub,s,x1,y1,j,b,i,med,mlms,slms,mE,e,
         ij,mSt,BB,BS,Rs,Hi,R,fr1,W,cnt,cntr,seq,cnv,gam,res,
         Lev,S2;
X=Xglobal;
y=yglobal;
n=rows(x);
            if dd==1;
              p=rows(b0);
            elseif dd==2;
              p=rows(b0)-1;
            else;
              p=rows(b0)-2;
            endif;
mE=zeros(n,1);
BS=zeros(k,1);
ij=0; @ ij is a dummy index linked to the storage of the best sample @
 if d==1;
   {nss}=elms(n,k); @ nss contains in the rows all possible subsamples @
   sub=rows(nss); @ number of subsamples used in the calculation of LMS @
 else;
        s1=zeros(k,d);
     sub=d;
 endif;


 cntr={0,0}; @ cntr= counter for the subsumples in which convergence
                 was not obtained @
  i=1; do until i==sub+1;
              ij=ij+1; @  ij=dummy index @
      start:
   if d==1;
        s=(nss[i,.]');
   else;
          good:
        s=ceil(n*rndu(k,1));
              @ Extract without replacement @
              if rows(unique(s,1))<k;
                 goto good;
              else;
              endif;
        s=sortc(s,1);
    endif;


 sglobal=s;

 x1=X[s,.];
 y1=y[s];
     "extracting subsample";; i;
        s;
   if  (rank(x1)<cols(X) or rows(unique(y1,1))==1) and i < sub;
                 if d==1;
                    i=i+1;
                   else;
                 endif;
          goto start;
    elseif (rank(x1)<cols(X) or rows(unique(y1,1))==1) and i==sub;
          goto end;
     else;
    endif;
  if d ne 1;
      j=1; do while j<i;
  @ this if checks whether the subsample had already been extracted @
           if sumc(s.==s1[.,j])==k;
              goto start;
           else;
           endif;
         j=j+1;
       endo;
      s1[.,i]=s;
    else;
   endif;

   @ Use Procedure nls to estimate beta coefficients @
   Xglobal=X[s,.];
   yglobal=y[s];
   {b,Lev,E,S2,gam,cnv}=nls(b0,dd);
   Xglobal=X;
   yglobal=y;
 @ "cnv"; cnv; wait;  @
  cntr[1]=cntr[1]+cnv; @ count the number of subsamples in
                         which ILS did not converge @
   if cnv==1;
      med=1e+7;
   goto skip;
   else;
   endif;

    b=b[1:p,1];

   @ Use Procedure prednls to compute residuals for the overall sample @
   {RS}=prednls(b);
  @ 1st col of RS contains simple residuals
    2nd col of RS contains projected residuals
    2rd col of RS contains studentized projected residuals @
           if sh==1 or sh==2 or sh==3;
              e=sortc(RS[.,sh]^2,1);
           else;
              e=sortc(RS[.,4]^2,1);
           endif;
     if p==k;
       e=e[p+1:n];
     else;
     endif;
       @ med=median of the squared residuals different from zero @
       med=e[round(rows(e)/2)];

       skip:
   if ij==1;
       MST=med;
   else;
   endif;

         if med <= mst;
             @ <= is necessary because the best could be the first @
               mst=med;
               Mlms=med; @ "Least median of squares is";; med;  wait; @
               Slms=s;   @ Slms contains the best elemental subset @
               "Best elemental subset";; Slms;
         else;
         endif;
       i=i+1;
    endo;
            end:
 @  For the best subset compute the lms @
 Xglobal=X[Slms,.];
 yglobal=y[Slms];

   {b,Lev,res,S2,gam,cnv}=nls(b0,dd);
    b=b[1:p,1];
    cntr[2]=cnv;
    Xglobal=X;
    yglobal=y;
    {mE}=prednls(b);

      retp(mE,Mlms,Slms,cntr);
  endp;

