/*
**  LRADDEL (deletion diagnostics for LDA (QDA) using likelihood ratio)
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {lrde,est}=lraddel(X,vsz,la,quadr,h)
**
**                                To compute del. diagnostics
**                                for likelihodd ratio  in AD
**                                The purpose is to test the null
**                                hypothesis that the transformation
**                                parameters are as those specified in
**                                vector la.
**
**
*/

/*
** LRADDEL
**
** PURPOSE:  To compute del. diagnostics for lik. ratio
**
** FORMAT:   {lrde,est}=lraddel(X,vsz,la,quadr,h)
**
** INPUT:     X = n x p data matrix; n observations
**                and p variables
**          vsz = g x 1 (where g is the number of groups) vector
**                which contains the sizes of the
**                groups
**           la = p x 1 vector specifies how the p variables in matrix
**                X should be transformed.
**                For example la={0,0.5} suggests a log transformation
**                for the first variable and s sqrt tr. for the second.
**        quadr = scalar dummy variable
**               if discr <> 1 QDA is used
**               else  LDA is used
**            h = scalar. If h=1 a plot which contains the values
**                of the del. likelihood ratio appears on the screen
**
** OUTPUT: lrde =  n x 2 matrix
**                 1st col. index from 1 to n
**                 2nd col. lr deletion diagnostic test
**          est =  n x p matrix
**                 mle deletion estimates for
**                 each variable separately
**
*/


 proc (2)=lraddel(X,vsz,la,quadr,h);

      local seq,seq1,i,n,p,ch,nla,pos,pot,likmax,lr,lrde,plik,est,
            lamax,bsb,In,Id,ident,vsz1,sz,j,lik,un,modsz,g,Xb,bs;

  n=rows(X); nla=rows(la);
  seq=seqa(1,1,n);
  bs=seq[1:n-1];
  seq1=seq[1:n-1];
  p=cols(X);
  g=rows(vsz);
  est=zeros(n,p);
 lrde=seq~miss(zeros(n,1),0);
 @ We use procedure flr for computing
   maximum likelihood estimates of transformation
   parameters using all the observations @
    @ row vector plik contains mle of transformation parameters @

    @ We use procedure fwdmle in order to compute
      max lik. of transformation parameters without unit i
      and procedure flr in order to compute the lik ratio test @
    ch=seqa(1,1,p)~la;

           id=0;
           i=1; do until i==g+1;
           id=id|reshape(i,vsz[i],1);
             i=i+1;
           endo;
          @ Ident Identifier vector  @
          Ident=id[2:n+1];
          vsz1=0|vsz|10^40;

 i=1;    do until i==n+1;

    bsb=setdif(seq,i,1);
     Xb=X[bsb,.];

    sz=vsz;
     j=1; do until j==g+1;
       if i>sumc(vsz1[1:j]) and i<=sumc(vsz1[1:j+1]);
         sz[j]=vsz[j]-1;
         goto modsz;
       else;
       endif;
        j=j+1;
    endo;
       modsz:
       @  bsb'; wait;
          "sz"; sz; wait;  @
  if quadr==1;
    {lamax}=fwdmleQD(bs,sz,Xb,sz,0,ch,0);
    {lr,lik,Un}=flrQD(bs,sz,Xb,sz,0,ch,lamax[.,2:p+1],0,0);
   else;
     {lamax}=fwdmleLD(bs,sz,Xb,sz,0,ch,0);
     {lr,lik,Un}=flrLD(bs,sz,Xb,sz,0,ch,lamax[.,2:p+1],0,0);
  endif;
      est[i,.]=lamax[1,2:p+1];
      lrde[i,2]= lr[1,2];
    i=i+1;
  endo;

    if h==1;
                 _pdate="";
                  title("Deletion values of the Lik. ratio test");
                  xy(lrde[.,1],lrde[.,2]);
                 graphset;
                      begwind;
                      if p<=3;
                        window(p,1,0);
                      elseif p<=6;
                        window(ceil(p/2),2,0);
                      elseif p<=12;
                        window(ceil(p/3),3,0);
                      else;
                        window(6,4,0);
                      endif;
                      makewind(80,6.8555,0,0,1);
                         i=1; do until i==p+1;
                                 _PMSGSTR =0 $+ "X" $+ FTOCV(i,1,0);
                                 _PMSGCTL=0.7~6~0.3~0~2~2;
                              xy(0,est[.,i]);
                      nextwind;
                         i=i+1;
                  endo;
                  endwind;
              else;
              endif;
      retp(lrde,est);
    endp;

