/*
**  MULTOUT
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {pos,pot,malaF,rdet,In,Un}=multout(X,bsb,forced,euclid);
**
**                                To detect multivariate outliers
**                                monitoring Mahalanobis distances
**                                in each step of the forward
**                                procedure
*/

/*
** MULTOUT
**
** PURPOSE:  to detect multivariate outliers.
**           A lot of statistics are monitered as the fwd search
**           progresses
**
** FORMAT:   {pos,pot,malaF,rdet,In,Un}=multout(X,bsb,forced,euclid);
**
** INPUT:     X = n x p data matrix; n observations
**                and p variables
**          bsb = m x 1 vector containing the units
**                belonging to the initial subset
**       forced = scalar which controls the interchange
**                  if scalar = 2 ==> no unit can leave the subset
**                  if scalar = 1 ==> units forming initial bsb cannot leave
**                  if scalar = 0 ==> every unit can leave the subset
**                if not scalar, the units indexed by vector forced 
**                will enter the subset before the others and when 
**                m=rows(forced) the units inside forced will not be
**                able to leave the subset anymore 
**       euclid = scalar.
**                if euclid ==1; The progressing and the monitoring
**                   refers to Euclidean distances
**                else the progressing and the monitoring
**                   is based on Mahalanobis distances 
**
**
**
** OUTPUT:  pos = (n-m+1) x 6 matrix.
**                1st col = forward search index
**                2nd Col m-th ranked MD
**                3rd average of the first m ranked MD
**                4th average of the first m-1 MD
**                5th average of the MD of the unis forming BSB
**                6th max of MD among those forming BSB
**          pot = (n-m) x 3 matrix
**                1st col = forward search index
**                2nd col (m+1)-th ranked MD
**                3rd col min. Mahalanobis distances among the units
**                            which form the group of potential outliers
**       malaF =  n x (n-m+2) matrix of MD in
**                each step
**                of the fwd search
**                The first row is a forward search for the first unit
**                First column is a sequence from 1 to n
**        rdet =  (n-m) x 3  matrix
**                1st col = forward search index
**                2nd col = ratio between |S_(i)| (determinant of
**                the covariance matrix excluding the units of noclean)
**                and |S| (determinant of covariance matrix)
**                3rd col = trace of S_(i)
**           IN = (n-m+1) x 11 matrix.
**                1st col = forward search index
**                2nd col = unit with max MD in bsb (mm-th MD)
**                3rd col = unit with (mm+1)-th MD
**                   ...
**                11th col = unit with (mm+10)-th MD
**           Un = (n-m) x 11 Matrix which contains the unit(s)
**                included in bsb at each step of the fwd search
**                REMARK: in every step the new subset is compared
**                with the old subset. Un contains the unit present
**                in the new subset but not in the old one
**
**    The output of the distances refers to Mahalanobis distances
**    only if input scalar euclid is different from 1
**    See also: MULTOUT1
**
*/


 proc (6)=multout(X,bsb,forced,euclid);

   local seq,ym,vc,malaf,In,Un,Unit,ds,rdet,va,varX,tr,bs,out,
       nn,n,mm,i,gg,pos,pot,zs,min,mala,ii,oldbsb,ncl,sncl,sbsb;

nn=rows(X);
seq=SEQA(1,1,nn);
mm=rows(bsb);
min=mm;
bs=bsb;
@   The first column of pos contains the dimension
    of the sample used to compute Mahalanobis distances
    2nd Col m-th ranked MD
    3rd average of the first m ranked MD
    4th average of the first m-1 MD
    5th average of the MD of the unis forming BSB
    6th max of MD among those forming BSB  @
  pos=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,5);

@   pot has two columns the first contains the dimension of the
    sample used in order to compute Mahalanobis distances,
    2nd col (m+1)-th ranked MD
    3rd col min. Mahalanobis distances among the units
                 which form the group of potential outliers @
  pot=seqa(mm,1,nn-mm)~zeros(nn-mm,2);

   @ Malaf= nn x (nn-mm+2) matrix of Mahalanobis distances
       in each step of the forward search @
   Malaf=seq~zeros(nn,nn-mm+1);

   @    in= Matrix whose 2:4th column contains the unit included in bsb
        at each step of the fwd search @
   In=seqa(mm,1,nn-mm+1)~miss(zeros(nn-mm+1,10),0);

   @    UN= Matrix whose 2nd column:10th col contains the unit(s)
           just included @
   Un=seqa(mm+1,1,nn-mm)~miss(zeros(nn-mm,10),0);

   @   rdet= (nn-mm) x 2 matrix.
       1st col = index related to the steps of the forward search
       2nd col = |S_(i)| / |S|
                  where S_(i) is the determinant of the cov.
                  matrix of n-i observations with centroid based on
                  (n-i) observations
       3rd col = trace(S_(i))/ trace(S) @

   rdet=seqa(mm,1,nn-mm)~zeros(nn-mm,2);
   varX=vcx(X);
   ds=det(varX);
   tr=sumc(diag(varX));

/***  calculation of Mahal. dist. on initial clean data set ***/

mala=seq~zeros(nn,1);

mm=mm; do until mm==nn+1;  @ nn= numero tot. di unita' stat. @
 @ "mm"; mm;  @
       ym=meanc(X[bsb,.]);
       va=vcx(X[bsb,.]);
@ 2nd col of Matrix Mala contains Mahalanobis distances _or Euclidean distances @

if euclid ==1;
             i=1;  do until i==nn+1;
                 mala[i,2]=sqrt((X[i,.]-ym')*(X[i,.]'-ym));
               i=i+1;
            endo;

else;
       vc=invpd(va);
             i=1;  do until i==nn+1;
                 mala[i,2]=sqrt((X[i,.]-ym')*vc*(X[i,.]'-ym));
               i=i+1;
            endo;
endif;

Malaf[.,mm-min+2]=mala[.,2];


zs=sortc(mala,2);

               zs=zs|miss(zeros(10,2),0);
               @ Store in matrix In unit ranked mm-th (mm+10)-th @
               IN[mm-min+1,2:11]=zs[mm:mm+9,1]';
  @ sortc MD among those which belong to bsb @
  sbsb=sortc(mala[bsb,2],1);

                @   mm-th , average of the first mm,  trimmed av.  @
 pos[mm-min+1,2:6]=zs[mm,2]~meanc(zs[1:mm,2])~meanc(zs[1:(mm-1),2])~
             @ mean of bsb ,   max among mcl   @
             meanc(mala[bsb,2])~sbsb[mm];

     if mm<nn;

 @ rdet contains in the 2nd and 3rd col. functions of Wilks ratio @

  rdet[mm-min+1,2]=det(va)/ds;
  rdet[mm-min+1,3]=sumc(diag(va))/tr;

 ncl=setdif(seq,bsb,1);
 sncl=sortc(mala[ncl,2],1);

@ pot contains mm+1 th MD and  minimum of Mahalanobis distances among
  the units which form the group of potential outliers @
pot[mm-min+1,2:3]=zs[mm+1,2]~sncl[1];

@ the dimension of the clean data set increases of one unit.
  vector y contains the indexes corresponding to the units of
  the new basic subset @
       oldbsb=bsb;
      if rows(forced)==1 and forced==2;
          mala[bsb,2]=-ones(mm,1);
      elseif rows(forced)==1 and forced==1;
          mala[bs,2]=-ones(min,1);
      elseif rows(forced)==1 and forced==0;
      
      else;
          if mm<rows(forced);
            out=setdif(seq,forced,1);
            mala[out,2]=1e+16*ones(nn-rows(forced),1);
          else;
            mala[forced,2]=zeros(rows(forced),1);
          endif;
      endif;
      @  mm;  sortc(bsb,1)'; wait;  @
       zs=sortc(mala,2);

       bsb=zs[1:mm+1,1];

             unit=setdif(bsb,oldbsb,1);
             if rows(unit)<=10;
              UN[mm-min+1,2:rows(unit)+1]=(unit');
             else;
              "Warning: interchange greater than 10";
              UN[mm-min+1,2:11]=unit[1:10]';
             endif;

      else;
    endif;


                    mm=mm+1;
                  endo;
   retp(pos,pot,Malaf,rdet,In,Un);
endp;

