/*
**  NLS
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
**
**  {b,Lev,E,S,gam,cnv}=nls(b0,dd)            To fit a non linear
**                                          regression model via
**                                          least squares
**
*/


/*
**
** NLS
**
** PURPOSE: To fit a non linear  regression model via
**          least squares
**
** FORMAT:
**
** {b,Lev,E,S,gam,cnv} = nls(b0,dd)
**
** INPUT:      b0 = k x 1  vector which specifies the starting
**                  values for beta coeffcients
**             dd = scalar which specifies the method of
**                  minimization
**                  if  dd==1 we use unweighted minimization
**                    see Seber and Wild eq. 2.166 with w_i=1
**                    \sum_{i=1}^n [y_i-f(\x_i,\theta)]^2
**                    In this case we estimate only the \theta
**                    (regression) coefficients
**                  elseif dd==2 we use direct minimization
**                    see Seber and Wild eq. 2.166
**                    \sum_{i=1}^n \ln w_i-0.5n w_i[y_i-f(\x_i,\theta)]^2
**                    where w_i=f(\x_i,\theta)^(2 \phi-2)
**                    In this case we estimate both the \theta
**                    (regression) coefficients and transformation
**                    parameter phi (lambda)
**                  else we use Power transform for both sides
**                    see Seber and Wild eq. 2.148
**                    In this case we estimate both the \theta
**                    (regression) coefficients and transformation
**                    parameter phi (lambda) and \sigma^2
**
**
**      REMARK 1      THE MAIN FILE MUST CONTAIN 2 GLOBAL
**                    VARIABLES
**                    yglobal  = n x 1 vector containing response
**                    Xglobal  = n x p matrix containing expl. variables
**                    YOU MUST ALSO CREATE A GAUSS PROC WHOSE NAME
**                    MUST BE funcnls WHICH SPECIFIES THE
**                    LINK BETWEEN X
**                    AND THE PARAMETERS VECTOR B
**                    For example for the Michaelis Menten model
**                    (Seber and Wild p. 148 @
**                      proc (1)=funcnls(psi);
**                          local X;
**                          X=Xglobal;
**                          retp(  (psi[1]*X)./(psi[2]+X) );
**                      endp;
**
**
** OUTPUT      b = k x 4 matrix which contains the output of est.
**                 procedure. More precisely
**                 if d==1     b[.,1]     contains the regression
**                                        coefficients
**                 elseif d==2 b[1:k-1,1] contain the regression coeff.
**                             b[k,1]     contains the mle estimate of
**                                        transformation parameter
**                                        (lambda)
**                 else b[1:k-2,1] contains the regression coeff.
**                      b[k-1,1]   contains the mle estimate of
**                                 transformation parameter (lambda)
**                      b[k,1]     contains the mle esimate of
**                                 \sigma^2
**                 2nd col. contains standard errors
**                 3rd col. contains t-statistics
**                 4th col. p-values
**           Lev = n x 1 vector which contains the leverage h_ii for
**                 every unit
**                 h_ii=f_i' (F'F)^(-1) f_i that is the i-th diagonal
**                 element of the so-called hat matrix
**                 F (F'F)^(-1) F' See Seber p. 211
**                 F= matrix of partial derivates wrt parameters
**                 evaluated at mle of b
**             E = n x 3 matrix
**                 1st col. contains simple residuals
**                 2nd col. contains studentized residuals
**                 3rd col. contains projected residuals
**                 4rd col. studentized projected residuals
**             S = 2 x 1 vector which contains two estimates of  \sigma^2
**                 S[1,1]= (y-f(\hat(beta)))^2/(n-p)
**                         if n=p S[1,1]= missing value
**                 S[2,1]= \eps'(I_n-P12) \eps/(n- trace(P12))
**                         eq. 4.170 p. 180
**                         if n<=p+0.5*p*(p+1)
**                         S[2,1]= missing value
**           gam = 2 x 1 vector  if n >= p+0.5*p*(p+1)
**                 element [1,1] = parameter-effects curvature
**                 element [2,1] = intrinsic curvature
**                 Remark: both curvatures need to be small
**                         if the tangent-plane approximaion
**                         is to be considered valid (p. 131)
**                 Remark: if n<p+0.5*p*(p+1) gam= missing value
**          cnv =  dummy which informs about convergence
**                 of the iterative procedure
**                 cnv=0 ==> normal convergence
**                 cnv=1 ==> convergence not reached
**
**
**          REMARK2: if the absolute value of one of the estimated
**                  parameters is greater than 10000 the program
**                  assumes that convergence has not been reached
**
**          REMARK3: Projected residuals can be computed only
**                   if n>= p+0.5*p*(p+1)
**                   where p is the number of estimated
**                   regression coefficients
**          REMARK4: The program uses all 6 optmization
**                   algorithms given in optmum
**                   before claiming non convergence
**
**
**   See also: FWDNLS
*/

proc (6)=nls(b0,dd);
   local X,y,n,b,b1,b_new,fvalue,grad,retcode,mlogl,Fssm,F,endd,C,
         p,q,h1,i,h,h2,mu,W,Bet,ss,ind,rho,Q1,U,D,k,Kssm,cn,Lev,S2,
         r,s,sa,Gm,Assm,AssmT,AssmN,At,Att,Atn,tol,star,res,cnv,ini,
         star1,d2,rd1,Xs,P12,PPc,Pres,Mssm,j,l,rd,d1,ml,mlt,mln,E,gam,
         spres,Ercm,RR;
            y=yglobal;
            X=Xglobal;
            n=rows(X);
            if dd==1;
              p=rows(b0);
            elseif dd==2;
              p=rows(b0)-1;
            else;
              p=rows(b0)-2;
            endif;
            q=0.5*p*(p+1);
            @ Matrix B will contain
              beta coeff. s.e., t-statistics and p-values @
            B=miss(zeros(rows(b0),4),0);
            @ Vector S2 will contain two estimates of \sigma^2 @
            S2=miss(0|0,0);

              cn=0;
               rest:
             if dd==1; @ Unweighted minimization @
               {b1,fvalue,grad,retcode}=OPTMUM(&nls1,b0);
             elseif dd==2; @ Direct minimization @
               {b1,fvalue,grad,retcode}=OPTMUM(&nls2,b0);
               @ Retransform \lambda @
               b1[p+1]=2.5*sin(b1[p+1]);
             else;  @ Power transform for both sides @
               {b1,fvalue,grad,retcode}=OPTMUM(&nls3,b0);
               @ Retransform \lambda \sigma @
               b1[p+1]=2.5*sin(b1[p+1]);
               b1[p+2]=exp(-b1[p+2]);
             endif;
               @ Create scalar cnv which informs about
                 convergence of the iterative process @
               if retcode ==0 and maxc(abs(b1))<50000;
                    cnv=0;
               elseif cn<6;
                    cn=cn+1;
                    _opalgr=cn;
                    b0=b0+rndn(rows(b0),1);
                    goto rest;
               else;
                    b=miss(0,0);
                    Lev=b;
                    E=b;
                    S2=b;
                    gam=b;
                    cnv=1;
                    goto endd;
               endif;
               b_new=b1[1:p];
               @ Store in the first col. of matrix B
                 the estimated coefficients @
               B[.,1]=b1;
        @ F= n x p matrix containing first derivatives @
          F=gradp(&funcnls,b_new);
        @ ss=sqrt(||y-mu||^2)/(n-p) @
         {mu}=funcnls(b_new);
         @ Vector res contains the residuals @
         res=y-mu;
         if n>p;
           ss=sqrt(moment(res,0)/(n-p));
         else;
           ss=miss(0,0);
         endif;

            @ Store estimate of disperion parameter @
            S2[1,1]=ss^2;
         @ Store in vector Lev the leverage for every unit @
         trap 1;
         if ismiss(invpd(F'F))==0;
                C=invpd(F'F);
             @   F'F;   Xglobal=F'F;
                Xglobal; "mmm"; wait;  @
                Lev=diag(F*C*F');
                @ Store studentized residuals @
                res=res~res[.,1]./(ss*sqrt(1e-10+ones(n,1)-Lev));

                @ Compute standard errors and t-statistics of beta
                  coefficients @
                B[1:p,2]=ss*sqrt(diag(C));
                B[1:p,3]=B[1:p,1]./B[1:p,2];
                /* pvalue t ratios (2 sided test) */
              if n>p;
                B[1:p,4]=2*cdftc(abs(B[1:p,3]),n-p);
              endif;
          else;
                   Lev=miss(0,0);
          endif;

  if n>=p+q;

              Fssm=zeros(n,q);
              h1=miss(zeros(p,p),0);
              h1=diagrv(upmat(h1),zeros(p,1));
             @   RR=qr(F);
                RR=inv(RR);  @

              i=1; do until i==n+1;
                  Xglobal=X[i,.];
                  h=hessp(&funcnls,b_new);
                  h2=h+h1;
                   @ Row vector of 2nd derivatives
                   arranged in symmetric storage mode @
                    Fssm[i,.]=(packr(vec(h2))');
                i=i+1;
              endo;
                  Xglobal=X;
             @ Ercm= Effective residual curvature matrix
               Ercm=zeros(q,1); @
              @   j=1; do until j==q+1;
                   Ercm[j]=res[.,1]'  @

       @ ind =vector which containes the indexes of the cols of matrix
              Fssm sorted in non decreasing order @
       ind=rev(sortind(minc(abs(Fssm))));

       @ Pivot the columns of matrix Fssm @
       Fssm=Fssm[.,ind];
        rho=ss*sqrt(p);
       @ D= n x (p+q) matrix @
       D=(F~Fssm)/(rho);
       @ QR decomposition of matrix D @
       {Q1,U}=qqr(D);

      @ {Q1,U2}=qqr(F/rho);
       Q1 is unique
       U is divided by rho In other words U=U2*rho
             "jjjj"; wait; @

      @ K = inv of the upper triangular matrix R_{11}
            See equation 4.37 @
               K=inv(U[1:p,1:p]);
            @    Remark:  RR*rho=K;
             K=RR*rho; @

       @ sa = p*(p+1)/2  x 2 matrix containing
         all possible subsamples of size 2 which can
         be extracted with replacement from p elements  @
       {sa}=elmsr(p,2);


      @ Mssm= q x q matrix composed of products of elements in K
              See equation 4.77 @
      Mssm=zeros(q,q);
              l=1; @ l = dummy index used for the cols of matrix Mssm @
               s=1; do until s==p+1;
                 r=s; do until r==p+1;
                     i=1; do until i==q+1;
                         if sa[i,1]==sa[i,2];
                            Mssm[i,l]=K[sa[i,1],r]*K[sa[i,2],s];
                          @ sa[i,1]~r~sa[i,2]~s; WAIT;      @
                         else;
                            Mssm[i,l]=K[sa[i,1],r]*K[sa[i,2],s]
                                       +K[sa[i,2],r]*K[sa[i,1],s];
                                 @   sa[i,1]~r~sa[i,2]~s~0~
                                  sa[i,2]~r~sa[i,1]~s; WAIT; @
                         endif;
                       i=i+1;
                     endo;
                  l=l+1;
                 r=r+1;
                 endo;
               s=s+1;
              endo;
          @ Apply to the rows of matrix Mssm the same interchange
            which was applied to the columns of matrix Fssm @
          Mssm=(Mssm[ind,.]);
         @ Gm = n x q matrix containing g_11 q_12 ... g_pp (Remark: pp=q)
           This matrix will be necessary for the construction
           of projected residuals @
         Gm=Fssm*Mssm;   @ eq. 4.45  @
         @ Assm = (p+q) x q containing intrinsic curvature arrays (which
           will be stored in matrix AssmN) and parameters effects
           curvature (which will be stored in matrix AssmT) @
         Assm=Q1'Gm;   @ eq. 4.77 @
         AssmT=Assm[1:p,.];   @ eq. 4.78 @
         @ At= p x p matrix which will contain a generic face of matrix
               A..^T and A..^N (A..^T and A..^N
               are defined in equations 4.56 and 4.57 @
         At=zeros(p,p);
         @ Att= p^2 x p matrix which will contain the p faces
                of matrix A..^T @
         ATT=zeros(p*p,p);
          l=1; do until l==p+1;
            j=0;
            i=0; do until i==p;
                At[i+1,i+1:p]=AssmT[l,j+1:j+p-i];
                 j=j+p-i;
               i=i+1;
              endo;
                ATT[(l-1)*p+1:l*p,.]=diagrv(upmat(At)'+at,diag(at));
            l=l+1;
          endo;

          @ Rescale matrice Att  @
           Att=att/rho;

          @ Specify the tolerance level for the curvatures @
             tol=1e-15;

          @ Maximize eq. 4.58 with respect to the unit vector d @
            @ Parameters effect curvature @
            d1=ones(p,1);
            d=d1/sqrt(d1'd1);
            j=0;
            star1:
           rd=0;
         i=1; do until i==p+1;
              rd=rd+(d'Att[(i-1)*p+1:i*p,.]*d)*Att[(i-1)*p+1:i*p,.]*d;
              rd1=rd./sqrt(rd'rd);
            i=i+1;
         endo;
            if rd1'*d>1-tol;
                      mL=0;
                     i=1; do until i==p+1;
                          ml=ml+(d'Att[(i-1)*p+1:i*p,.]*d)^2;
                      i=i+1;
                     endo;
                     mlT=sqrt(ml);
              else;
               j=j+1; "Iteration for parameters curvatures";;j;
               d=rd1;
               goto star1;
            endif;

           @ rows(Assm); cols(Assm); wait;  @
         AssmN=Assm[p+1:p+q,.]; @ eq. 4.79 @
            @ AtN= p*q x p matrix which will contain the q faces
                of matrix A..^N @
            ATN=zeros(p*q,p);

          l=1; do until l==q+1;
            j=0;
            i=0; do until i==p;
                At[i+1,i+1:p]=AssmN[l,j+1:j+p-i];
                 j=j+p-i;
               i=i+1;
              endo;
                ATN[(l-1)*p+1:l*p,.]=diagrv(upmat(At)'+at,diag(at));
            l=l+1;
          endo;

          @ Rescale matrice  Atn @
           Atn=Atn/rho;

          @ Maximize eq. 4.59 with respect to the unit vector d @
          @ Intrinsic curvature @
             d=ones(p,1);
             d=d/sqrt(d'd);
             j=0;
           star:
              rd=0;
         i=1; do until i==q+1;
              rd=rd+(d'Atn[(i-1)*p+1:i*p,.]*d)*Atn[(i-1)*p+1:i*p,.]*d;
              rd1=rd./sqrt(rd'rd);
            i=i+1;
         endo;
            if rd1'*d>1-tol;
                   mL=0;
                       i=1; do until i==q+1;
                          ml=ml+(d'ATN[(i-1)*p+1:i*p,.]*d)^2;
                        i=i+1;
                       endo;
                         mlN=sqrt(ml);
              else;
              j=j+1; "Iteration for intrinsic curvature";;j;
              d=rd1;
               goto star;
            endif;
         @ Store the two curvatures in vector gam @
         gam=mlT|mlN;


      @ Compute the projected residuals @
      @ Xs= (I_n -Q_p*Q_p')*Gm  @
      Xs=Q1[.,1:p]*(Q1[.,1:p]');
      Xs=diagrv(Xs,ones(n,1)-diag(Xs));
      @ Xs = n \times q @
      Xs=Xs*Gm; @ last equation on p. 182 @
      @ PP Matrix which projects into \Omega_2 @
        ini:
        if abs(maxc(Xs))<1e+2;
            Xs=Xs*100;
           goto ini;
        endif;

      if rank(Xs)<q;
           E=res~miss(zeros(n,2),0);
      else;
           P12=Xs*invpd(Xs'Xs)*Xs';
           @ PPc matrix which projects into the
             orthogonal complement of \Omega_2 (PPc = I_n - P12) @
           PPc=diagrv(P12,ones(n,1)-diag(P12));
           @ Pres= n x 1 vector of Projected residuals @
           Pres=PPc*res[.,1];
           Spres=(Pres./sqrt(S2[2,1]*diag(PPc)));
           E=res~Pres~spres;
           @ estimates \sigma^2 using eq. 4.170 @
           S2[2,1]=res[.,1]'PPc*res[.,1]/(n-sumc(diag(P12)));
     endif;

  else;
     E=res~miss(zeros(n,2),0);
     gam=miss(0,0);
  endif;

     endd:

 retp(B,Lev,E,S2,gam,cnv);
endp;

/* nls1 is the procedure which creates the function to minimize */
proc (1)=nls1(th);   @ Unweighted minimization eq.2.166 with w_i=1 @
    retp(sumc((yglobal-funcnls(th))^2));
endp;

proc (1)=nls2(th);   @ Direct minimization eq. 2.166 @
    local X,y,n,phi,fit;
      X=Xglobal;
      y=yglobal;
      n=rows(X);
      phi=2.5*sin(th[rows(th)]);
      th=th[1:rows(th)-1];
      {fit}=funcnls(th);
               if minc(fit)<=0;
                  @ fit=fit-minc(fit)+rndu(n,1)/100;  @
            fit=substute(fit,fit.<0,fit-minc(fit)+rndu(n,1)/100);
                else;
               endif;
    retp(-sumc(ln(fit^(2*phi-2)))+n*ln(sumc(fit^(2*phi-2).*(y-fit)^2)));
endp;


proc (1)=nls3(th); @ Power transform both sides min. eq. 2.148 @
    local n,rth,la,fit,X,y,sla,sig;
      y=yglobal;
      X=Xglobal;
      n=rows(X);
      rth=rows(th);
      la=2.5*sin(th[rth-1]);  @ th = \theta | \lambda | \sigma^2 @
      sig=exp(-th[rth]);
      th=th[1:rth-2];
      {fit}=funcnls(th);
               if minc(fit)<=0;
                  fit=abs(fit);
                else;
               endif;
               if sig<=0;
                 sig=abs(sig);
               else;
               endif;
    sla=sumc( ((y^la-fit^la)/la)^2  );
 retp(0.5*n*ln(sig)+0.5*(1/sig)*sla -(la-1)*sumc(ln(y)));
endp;


