/*
**  PCA.G
**  written by MARCO RIANI
**
**  FORMAT                                 PURPOSE       Line
** =======================================================================
** {val,vect,cor,Comp}=pca(X,k,st,pr);
**                                      To preform principal
**                                      component analysis
**
**
*/


/*
**
** PCA.G
**
** PURPOSE:  to compute principal component analysis
**
** FORMAT:   {val,vect,cor,comp}=pca(X,k,st,pr);
**
** INPUT:     X    =  n x p matrix containing
**                    n observations for p variables
**            k    =  scalar specifying the number of princ. comp
**                    which must be considered
**            st   =  scalar
**                    if st==1 the variables are standardized
**                    else the variables are simply centerd
**                    removing the arithmetic means
**            pr   =  scalar
**                    if pr ==1 the output of principal component is
**                               shown on the screen
**                    else no output is shown on the screen
**
** OUTPUT:    val  =  p x 3 matrix
**                    1st col = eigenvalues
**                    2nd col = proportion of variance explained
**                    3rd col = Cumulative proportion of variance expl.
**            vect  = p x k matrix
**                    which contains in the columns the
**                    first k standardized eigenvectors
**             cor  = p x k matrix
**                    which contains in the columns the
**                    correlations (loadings) between the p variables
**                    and the first k principal component
**            comp  = n x k matrix containing the
**                    principal component scores
**
** REMARK:   unbiased variance covariance matrix is used
**
*/


proc (4)=pca(X,k,st,pr);
    local n,p,la,vect,las,seqr,cumla,Val,cor,Comp,i;

 n=rows(X); p=cols(X);

     if st==1;
       X=stand(X);
     else;
       X=X-(meanc(X)'); @ Matrix whose sum of the cols is 0 @
     endif;

    {la,vect}=eigrs2(X'X/(n-1));
    las=sumc(la);
    seqr=seqa(p,-1,p);
    la=la[seqr];    @ First k Eigenvalues in descending order @
    vect=vect[.,seqr];  @ First k Normalized eigenvectors @
    cumla=la/las;   @ Proportion of variance explained by the first k
                      eigenvalues @
    Val=la~cumla~cumsumc(cumla);
    @ cor = p x k matrix which contains corr.
            between variables and comp. @
    cor=(vect.*sqrt(la'))./stdc(X);

    @ Comp= n x k matrix containing the first k principal components @
     Comp=X*vect;

    if pr==1;
    "Number.      Eig.      Prop. explained   Cum. prop.  ";
     i=1;  do until i>p;
          format 4,0; /ld  i;;
          format 14,3; val[i,1];; val[i,2];; val[i,3];
        i=i+1;
      endo;
     "  ";
     "Analysis of Correlations between variables and components";
     "Var                               Components ";
     "          ";; format 9,0; /ld seqa(1,1,k)';
     i=1;  do until i>p;
          format 4,0; /ld "X" i;;
          format 9,3; cor[i,1:k];
        i=i+1;
      endo;
    endif;

  retp(val[1:k,.],vect[.,1:k],cor[.,1:k],comp[.,1:k]);
 endp;

