/*
**  PREDGLM
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
**
**  {RS}=predglm(X,y,q,fr,b_new,off)    To produce residuals
**                                      from a glm using
**                                      as input vector of beta coeff.
**
*/


/*
**
** PREDGLM
**
** PURPOSE:  to produce residuals from  a generalized linear model,
**
**
**
**
** FORMAT:
**
** {RS} = predglm(X,y,q,fr,b_new,off)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             q  = scalar which specifies the kind of GLM
**                  q=1 ==> Poisson link
**                  q=2 ==> Logit link
**                  q=3 ==> Complementary Log log link
**                  q=3.05 ==> Gamma link (Box Cox la=2)
**                  q=3.1 ==> Gamma link (Box Cox la=1)
**                  q=3.2 ==> Gamma link (Box Cox la=0.5)
**                  q=3.3 ==> Gamma link (Box Cox la=0)
**                  q=3.4 ==> Gamma link (Box Cox la=-0.5)
**                  q=3.5 ==> Gamma link (Box Cox la=-1)
**                  q=3.6 ==> Gamma link (Power family la=3)
**                  q=3.7 ==> Gamma link (Power family la=2)
**                  q=3.8 ==> Gamma link (Power family la=1)
**                  q=3.9 ==> Gamma link (Power family log)
**                  q=4   ==> Gamma link (Power family la=-1)
**                            (inverse)
**                  q>4  and q<5 Gamma link all values of lambda
**                       between 0 and -1
**                       For example q=4.25 ==> la=-0.25
**                       For example q=4.5 ==> la=-0.5
**                  q=5 ==> Inverse Gaussian link
**                  q=6 ==> Probit link
**                  q=7 ==> Arc sin link
**                  q=8 ==> Modified logit link
**             fr = n x 1 vector of frequencies
**                  if fr = scalar 0 a n x 1 vector of 1s
**                  is created automatically by the program
**          b_new = p x 1 vector of coefficients
**           off = n x 1 vector which defines an offset
**                 REMARK: off = scalar 0 implies no offset
**
** OUTPUT:     RS =  n x 5 matrix
**                  1st col. simple residuals  y_i - \mu_i
**                  2nd col. Pearson residuals  McN p. 37 eq. (2.11)
**                  3rd col. deviance residuals  rD McN p. 30
**                  4th col. Studentized Pearson residuals
**                           McN p. 397 eq. (12.4)
**                  5th col. Studentized Deviance residuals
**                           McN p. 397  eq. (12.5)
**
**  REMARK1    this program does not add a column of 1 in matrix X to
**             create the intercept.
**
*/


proc (1)=predglm(X,y,q,fr,b_new,off);
   local  n,p,i,j,dm,Xb,one,
          th,mu,V,z,W,e,rP,Cb,rd,df,RS,phi,H,Hi,rSpr,rSDr,lam;

   n=rows(X);
   p=cols(X);
   df=n-p;
   one=ones(n,1);
   mu=one; @ This is simply an inizialization for mu @
   dm=one; @ This is simply an inizialization for dm @
   V=one;  @ Initialization @
   if fr==0;
    fr=one;
   else;
   endif;

   if q==3.05;
       lam=2;
   elseif q==3.1;
       lam=1;
   elseif q==3.2;
       lam=0.5;
   elseif q==3.4;
       lam=-0.5;
   elseif q==3.5;
       lam=-1;
   else;
   endif;



                   th=X*b_new+off;

                 if q==1;     @ Poisson link @
                        i=1; do until i==n+1;
                           if th[i]>25;
                             mu[i]=1e+10;
                             dm[i]=1e-10;
                             V[i]=1e+10;
                           elseif th[i]<-25;
                             mu[i]=1e-10;
                             dm[i]=1e+10;
                             V[i]=1e-10;
                           else;
                             mu[i]=exp(th[i]);
                             dm[i]=1/mu[i];
                             V[i]=mu[i];
                           endif;
                          i=i+1;
                         endo;

               /*******    mu=exp(th);
                   dm=1/(mu+1e-12);
                        @ dm= derivative of \eta wrt mu @
                   V=mu;         @ V= variance @
                    *********/

                 elseif q==2; @ Logit  link @
                        i=1; do until i==n+1;
                           if th[i]>35;
                             mu[i]=1-1e-14;
                           else;
                             mu[i]=exp(th[i])/(1+exp(th[i]));
                           endif;
                          i=i+1;
                         endo;

                   dm=1/((mu+1e-12).*(1.000000001-mu));
                   V=mu.*(1-mu) ./fr + 1E-12;
                   @ V=variance (fr are the frequencies) @

                 elseif q==3; @ Complementary log log link @

                        i=1; do until i==n+1;
                           if th[i]>5;
                             mu[i]=1-1e-10;
                           elseif th[i]< -18;
                             mu[i]= 1e-10;
                           else;
                             mu[i]=1-exp(-exp(th[i]));
                           endif;
                          i=i+1;
                         endo;

                       @ mu=1-exp(-exp(th));  old @
                   dm= -1/( (1-mu+1e-12).*ln(1-mu+1e-12) );
                   V=mu.*(1-mu) ./fr+ 1e-12;

                 elseif q>3 and q <=3.5;  @ GAMMA LINK (Box Cox) @
                   if q ne 3.3;
                    mu=  (1+lam*th)^(1/lam);
                    dm=  mu^(lam-1);
                   else;
                    mu= exp(th) @ +1e-8 @ ;
                    dm= 1/mu;
                   endif;
                    V=(mu^2)./fr;


                 elseif q==3.6;  @ GAMMA LINK (third power) @
                   mu=  th^(1/3);
                   dm=  3*(mu^2);
                   V=(mu^2)./fr;

                 elseif q==3.7;  @ GAMMA LINK (square) @
                   mu=  sqrt(abs(th));
                   dm=  2*mu;
                   V=(mu^2)./fr;

                 elseif q==3.8;  @ GAMMA LINK (identity) @
                   mu=  th;
                   dm=   1;
                   V=(mu^2)./fr;

                 elseif q==3.9;  @ GAMMA LINK (log) @
                   mu=  exp(th);
                   dm=   1/mu;
                   V=(mu^2)./fr;

                 elseif q==4; @ GAMMA LINK (inverse) @
                   mu=  1/th;
                   dm= -  1/(mu^2);
                   V=(mu^2)./fr;

        /****
                 elseif q==4.2; @ GAMMA LINK (-0.5) @
                   mu= 1/(th^2);
                   dm= -0.5*(mu^(-1.5));
                   V=(mu^2)./fr;
         **/
                 elseif q>4 and q<5; @ GAMMA LINK (4+lambda) @
                   mu= 1/(th^(round(1/(q-4))));
                   dm= (4-q)*(mu^(3-q));
                   V=(mu^2)./fr;

                 elseif q==5;  @ INVERSE GAUSSIAN @
                   mu=1/sqrt(th);
                   dm= -2/(mu^3);
                   V= mu^3;

                 elseif q==6;  @ INVERSE GAUSSIAN @
                   mu= cdfn(th);   @ CDF evaluated in th @
                   dm=1/pdfn(TH);  @ inv. of density evaluated at th @
                   V=mu.*(1-mu)./fr;

                 elseif q==7; @ ARC SIN LINK @
                  i=1; do until i==n+1;
                    if th[i]>= pi/2;
                     mu[i]=1 -1e-8;
                     dm[i]=1e+6;
                    elseif th[i]<= - pi/2;
                     mu[i]=0 +1e-8;
                     dm[i]=1e+6;
                    else;
                     mu[i]=  0.5 * (1+sin(th[i]));
                     dm[i]=  2/sqrt(1-((2*mu[i]-1)^2));
                    endif;
                   i=i+1;
                  endo;
                  V= 1e-6+ mu.*(1.-mu)./fr;

                 elseif q==8; @ MODIFIED LOGIT LINK @
                        i=1; do until i==n+1;
                           if th[i]> 13.8 ;
                             mu[i]=1-1e-6;
                             dm[i]=1e+6;
                             V[i]=1e-6;
                           elseif th[i]< -13.8;
                             mu[i]=1e-6;
                             dm[i]=1e+6;
                             V[i]=1e-6;
                           else;
                             mu[i]=exp(th[i])/(1+exp(th[i]));
                             dm[i]=1/((mu[i]).*(1-mu[i]));
                             V[i]=mu[i].*(1-mu[i]) ./fr[i];
                           endif;
                          i=i+1;
                         endo;
                 else;
                 endif;
                    @  V; dm;   @
                   W=sqrt((dm.^2).*V);   @ sqrt of weight vector n x 1 @
                   Xb=X./W;              @ weighted regressors n x p @
                        Cb=invpd(Xb'Xb);

     /* MODEL CHECKING THROUGH RESIDUALS */
                  e=y-mu;    @ simple residuals @

               @  rD  = Deviance Residuals @
              @ RESIDUAL DEVIANCE (SPLUS) = SCALED DEVIANCE (mcN p. 34)
                = sumc(rD^2)  = -2 lik   @
               if q==1;     @ Poisson link @
                     rD=zeros(n,1);
                  i=1; do until i==n+1;
                     if y[i]>0;
                      rD[i]=2*(y[i].* ln(y[i]./mu[i]) - (y[i]-mu[i])  );
                     else;
                      rD[i]= 2* mu[i];
                     endif;
                    i=i+1;
                   endo;
                 phi=1;

               elseif q==2 or q==3 or q==6 or q==7 or q==8; @ binomial
               link  @
                     rd=zeros(n,1);
                 if maxc(fr)>1;
                    i=1; do until i==n+1;
                      if abs(y[i]-mu[i])<1e-09;
                        rd[i]=0;
                      elseif y[i]==0;
                        rd[i]=2*(fr[i]*ln(1/(1-mu[i]-1e-12)));
                      else;
                    rd[i]=2*(fr[i]*y[i]*ln(y[i]/(mu[i]+1e-12))+
                    (fr[i]-fr[i]*y[i])*ln((1-y[i]+1e-12)/
                    (1-mu[i]+1e-12)));
                      endif;
                    i=i+1;
                    endo;
                 else;
                   @ binary data @
                    i=1; do until i==n+1;
                      if mu[i]==1;
                        rd[i]= -2*(y[i]*ln(mu[i])+
                        (1-y[i])*ln(1-mu[i]+1e-12));
                      elseif mu[i]==0;
                        rd[i]= -2*(y[i]*ln(mu[i]+1e-12)+
                        (1-y[i])*ln(1-mu[i]));
                      else;
                        rd[i]= -2*(y[i]*ln(mu[i])+
                        (1-y[i])*ln(1-mu[i]));
                      endif;
                    i=i+1;
                    endo;
                 endif;
                 phi=1;

               elseif q>3 and q<5; @ Gamma link @
                     rD=zeros(n,1);
                 i=1; do until i==n+1;
                   if y[i]>0;
                    rD[i]=2*(fr[i])*(
                    - ln(y[i]/mu[i]) + (y[i]-mu[i])/mu[i]);
                   else;
                    rD[i]=2*(ln(mu[i]) -1);  @p. 290 McN @
                   endif;
                  i=i+1;
                 endo;
                  phi=sumc((e./mu)^2)/df;

               elseif q==5;  @ Inverse Gaussian link @
                  rD= ((y-mu).^2) ./ ((mu.^2).*y) ;
                  phi=sumc((e./mu)^2)/df;
               else;
               endif;
                if ismiss(rD)==1;
                  "STOP: THERE IS A MISSING VALUE IN VECTOR rD"; wait;
                  @ th[21:34]~rd[21:34]; wait;  @
                else;
                endif;

               @ Now we take the sign square root of the elements of rD @
                  i=1;  do until i==n+1;
                    if y[i]-mu[i] > 0;
                      rD[i]=sqrt(rd[i]);
                    else;
                      rD[i]=-sqrt(rd[i]);
                    endif;
                     i=i+1;
                   endo;

               @  rP= PEARSON RESIDUALS  eq. 2.11 p. 37 McN @
                  rP=e./sqrt(V);
               @ HAT MATRIX @
                  H=Xb*Cb*Xb';
                  Hi=diag(H);  @ Hi contains generalized leverage values @
             if maxc(Hi)<1;
               @  rSPR= Studentized Pearson Residuals ( 12.4 p. 397 McN) @
                  rSPR= e./  sqrt(phi * V.*(1-Hi));
               @  rSDR= Studentized Deviance Residuals (12.5 p. 397 McN) @
                  rSDR=rD./sqrt(phi *(1-Hi));
             else;
                 rSPR=miss(one,1);
                 rSDR=rSPR;
             endif;


                /* Store all the residuals  inside a matrix
                 1st col. simple residuals
                 2nd col. Pearson residuals
                 3rd col. deviance residuals   */
                 RS=e~real(rP)~real(rD)~rSPR~rSDR;
   retp(RS);
endp;

