/*
**  PREDNLS
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
**
**  {RS}=prednls(b_new)                 To produce residuals
**                                      from a NLS model using
**                                      as input a vector of beta coeff.
**
*/


/*
**
** PREDNLS
**
** PURPOSE:  to produce residuals from a NLS model,
**
**
**
**
** FORMAT:
**
** {RS} = prednls(b_new)
**
** INPUT:    b_new = p x 1 vector of coefficients
**
** OUTPUT:     RS =  n x 3matrix
**                  1st col. simple residuals  y_i - \mu_i
**                  2nd col. studentized residuals
**                  3rd col. projected residuals
**                  4th col. studentized projected residuals
**
**
**      REMARK        THE MAIN FILE MUST CONTAIN 2 GLOBAL
**                    VARIABLES
**                    yglobal  = n x 1 vector containing response
**                    Xglobal  = n x p matrix containing expl. variables
**                    YOU MUST ALSO CREATE A GAUSS PROC WHOSE NAME
**                    MUST BE funcnls WHICH SPECIFIES THE
**                    LINK BETWEEN X
**                    AND THE PARAMETERS VECTOR B
**                    For example for the Michaelis Menten model
**                    (Seber and Wild p. 148 @
**                      proc (1)=funcnls(psi);
**                          local X;
**                          X=Xglobal;
**                          retp(  (psi[1]*X)./(psi[2]+X) );
**                      endp;
**
*/

proc (1)=prednls(b_new);
   local X,y,b,Fssm,F,
         p,q,h1,i,h,h2,mu,Bet,ss,ind,rho,Q1,U,D,k,Kssm,
         r,s,S2,sa,Gm,res,N,Rs,Lev,C,
         d2,rd1,Xs,P12,PPc,Pres,Mssm,j,l,rd,d1,ml,mlt,mln,E;
            y=yglobal;
            X=Xglobal;
            p=rows(b_new);
            n=rows(X);
            q=0.5*p*(p+1);


        @ F= n x p matrix containing first derivatives @
          F=gradp(&funcnls,b_new);

       @ ss=sqrt(||y-mu||^2)/(n-p) @
        {mu}=funcnls(b_new);
        @ Vector res contains the residuals @
        res=y-mu;
        if n>p;
         ss=sqrt(moment(res,0)/(n-p));
        else;
         ss=1;
        endif;

                C=invpd(F'F);
                Lev=diag(F*C*F');
                @ Store studentized residuals @
                res=res~res[.,1]./(ss*sqrt(ones(n,1)-Lev));


   if n>=100+p+q;
           Fssm=zeros(n,q);
           h1=miss(zeros(p,p),0);
           h1=diagrv(upmat(h1),zeros(p,1));

           i=1; do until i==n+1;
               Xglobal=X[i,.];
               h=hessp(&funcnls,b_new);
               h2=h+h1;
               @ Row vector of 2nd derivatives
               arranged in symmetric storage mode @
                 Fssm[i,.]=(packr(vec(h2))');
             i=i+1;
           endo;
        Xglobal=X;


       @ ind =vector which containes the indexes of the cols of matrix
              Fssm sorted in non decreasing order @
       ind=rev(sortind(minc(abs(Fssm))));
       @ PIvot the columns of matrix Fssm @
       Fssm=Fssm[.,ind];
        rho=ss*sqrt(p);
       @ D= n x (p+q) matrix @
       D=(F~Fssm)/(rho);
       @ QR decomposition of matrix D @
       {Q1,U}=qqr(D);

       @ sa = p*(p+1)/2  x 2 matrix containing
         all possible subsamples of size 2 which can
         be extracted with replacement from p elements  @
       {sa}=elmsr(p,2);

      @ K = inv of the upper triangular matrix R_{11}
            See equation 4.37 @
        K=inv(U[1:p,1:p]);

      @ Mssm= q x q matrix composed of products of elements in K
              See equation 4.77 @
      Mssm=zeros(q,q);
       l=1; @ l = dummy index used for the cols of matrix Mssm @
        s=1; do until s==p+1;
          r=s; do until r==p+1;
              i=1; do until i==q+1;
                  if sa[i,1]==sa[i,2];
                     Mssm[i,l]=K[sa[i,1],r]*K[sa[i,2],s];
                  else;
                      i~l~q;
                     Mssm[i,l]=K[sa[i,1],r]*K[sa[i,2],s]
                               +K[sa[i,2],r]*K[sa[i,1],s];
                  endif;
                i=i+1;
              endo;
           l=l+1;
          r=r+1;
          endo;
        s=s+1;
       endo;
        @ Apply the same colum interchange which was applied
          to Fssm @
        Mssm=(Mssm[ind,.]);
       @ Gm = n x q matrix containing g_11 q_12 ... g_pp (Remark: pp=q)
         This matrix will be necessary for the construction
         of projected residuals @
       Gm=Fssm*Mssm;   @ eq. 4.45  @

      @ Compute the projected residuals @
      @ Xs= (I_n -Q_p*Q_p')*Gm  @
      Xs=Q1[.,1:p]*(Q1[.,1:p]');
      Xs=diagrv(Xs,ones(n,1)-diag(Xs));
      Xs=Xs*Gm; @ last equation on p. 182 @
      @ P12 Matrix which projects into \Omega_2 @
      P12=Xs*inv(Xs'Xs)*Xs';
      @ PPc matrix which projects into the
        orthogonal complement of \Omega_2 (PPc = I_n - P12) @
      PPc=diagrv(P12,ones(n,1)-diag(P12));
      @ Pres= n x 1 vector of Projected residuals @
      Pres=PPc*res[.,1];
      RS=res~Pres;

            @ Second estimate of \sigma^2 (eq. 4.170 is used) @
            S2=res[.,1]'PPc*res[.,1]/(n-sumc(diag(P12)));
            @ Store in matrix RS studentized projected residuals @
            RS=RS~(RS[.,3]./sqrt(S2*diag(PPc)));
   else;
    Rs=res~miss(zeros(n,2),2);
   endif;
 retp(RS);
endp;

