/*
**  QDA.G
**  written by MARCO RIANI
**
**  FORMAT                                 PURPOSE       Line
** =======================================================================
** {post1,Lnd,M,MD,eqva}=qda(Z,sz,X,vsz,prior);
**                                      To compute posterior
**                                      probabilities using
**                                      quadr. discr. analysis
**
*/


/*
**
** QDA.G
**
** PURPOSE:  to compute posterior probabilities using quadratic
**           discriminant analysis
**
** FORMAT:   {post1,Lnd,M,MD,eqva}=qda(Z,sz,X,vsz,prior);
**
** INPUT:     Z    = (n1+n2+...+ng) x p matrix containing
**                    observations from the g populations.
**                    Units forming matrix Z will be used
**                    to obtain quadratic discriminant rule.
**                    This is the so called "training set".
**            sz   = g x 1 (where g is the number of groups) vector
**                   which contains the sizes of the
**                   groups forming the training set
**                   e.g.: if sz =50|40|30
**                   then size of 1st gr. is 50
**                   then size of 2nd gr. is 40
**                   then size of 3rd gr. is 30
**            X     = (nn1+nn2+...+nng) x p matrix containing
**                    the units forming the validation set
**                    nni (i=1, ..g) can be greater or smaller than ni
**            vsz   = g x 1 (where g is the number of groups) vector
**                   which contains the sizes of the
**                   groups forming the validation set
**                   e.g.: if sz =50|40|30
**            prior = g x 1 vector containing prior probabilities
**
** OUTPUT:   post1 = nn x 1 vector of posterior probabilities
**             Lnd =  1 x g vector containing ln(det(var cov)) for
**                    the g group
**              M  =  g x p matrix.
**                    First row contains the means of gr. 1
**                    ...
**                    Last row contains the means of gr. g
**             MD  = (nn1+nn2+...+nng) x 1 vector
**                   containing (squared) Mahalanobis distances of
**                   each unit from the centroid of its
**                   population
**            eqva  = scalar which contains the value of the
**                    test statistics for the equality of
**                    var cov matrix (See program eqvar.g)
**
** REMARK1:   Pooled unbiased variance covariance matrix is used
** REMARK2:   This procedure does not report but also
**            computes matrix post
**            post  = (nn1+nn2+...+nng) x (g+2) matrix.
**                    1st col = identifier vector
**                    2nd col =post. prob. of belonging to gr. 1
**                    ...
**                    (g+1)th col =post. prob. of belonging to gr. g
**                    (g+2)th col = group to which the unit is
**                    assigned
**
*/


proc (5)=qda(Z,sz,X,vsz,prior);
    local n,nn,g,M,MD,mdt,Vz,iVz,Lnd,big,ident,
    ln1,ln2,p,dist,edd,s,post,Post1,i,k,Id,eqva;

n=rows(Z); p=cols(Z);   @ Z = training set @
nn=rows(X);             @ X = Validation set @
g=rows(sz);           @ ng = number of groups @
sz=0|sz;
vsz=0|vsz;
    @  Matrix M contains (in the rows) the vector of the means for
       the groups
       first row refers to first group
       ...
       last row refers to the last group  @
    M=zeros(g,p);

    @  Matrix VZ contains the varcov matrices of every
       group
       first p rows refer to first group
       ...
       last p rows refer to last group  @
    VZ=zeros(p*g,p);

    @  Matrix iVZ contains inv(varcov) of every
       group
       first p rows refer to first group
       ...
       last p rows refer to last group  @
    iVZ=zeros(p*g,p);

    @  Lnd is a row vector of dimension equal to the number of groups
       It contains the ln of the var cov matrix of each group
       first element refers to first group
       ...
       last element refers to last group  @
    Lnd=zeros(1,g);

    @ Post=(nn x g+2) matrix
      1st col contains the identifier vector
      2st col contains posterior prob. of belonging to gr. 1
      ...
      (g+1)th col contains posterior prob. of belonging to last gr.
      (g+2)th col contains to which group the unit has been assigned @

    @ Post1= nn x 1 matrix of posterior probabilities of each unit
      from its own universe @
    Post1=zeros(nn,1);
           id=0;
           i=1; do until i==g+1;
           id=id|reshape(i,vsz[i+1],1);
             i=i+1;
           endo;
          Ident=id[2:nn+1];
          @ Identifier vector  @
          BIG=10^5*ones(nn,1);   @ vector of big numbers @
    Post=Ident~zeros(nn,g)~BIG;

   @ MD= nn x 1 vector containing Mahalanobis distances of each unit
         from its own population @
   MD=zeros(nn,1);

   @ Store means, var cov, inv(var cov) and ln(det(var cov))
     for each group @

  i=1; do until i==g+1;
       M[i,.]=(meanc(Z[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]))';

    VZ[(1+p*(i-1)):p*i,.]=vcx(Z[(sumc(sz[1:i])+1):sumc(sz[1:i+1]),.]);
    iVZ[(1+p*(i-1)):p*i,.]=invpd(VZ[(1+p*(i-1)):p*i,.]);
    Lnd[1,i]=ln(det(VZ[(1+p*(i-1)):p*i,.]));
        i=i+1;
   endo;


dist=zeros(nn,g);  @ Matrix of sample distances  @
MDt=dist; @ Matrix of Mahalanobis distances from every population @

   i=1;  do until i==nn+1;
          k=1; do until k==g+1;
              @ MD of each unit from every population @
              MDt[i,k]= (X[i,.]-M[k,.])
             *iVz[((k-1)*p+1):(k*p),.]*((X[i,.]-M[k,.])');

            dist[i,k]=0.5*(Lnd[1,k]+ MDt[i,k])-ln(prior[k]);
              if dist[i,k]<post[i,g+2];
               post[i,g+2]=k;
              else;
              endif;
          k=k+1;
         endo;
    i=i+1;
   endo;
      @ Convert distances to posterior probabilities @
     edd=exp(-dist);
     s=sumc(edd');   @ denominator of posterior probabilities @
     post[.,2:(g+1)]=edd./s;
      @  rows correspond to units
         cols correspond to groups
    e.g. post[r,(s+1)]= posterior prob. of unit r of belonging to
                        group s @

   @ Vector MD contains Mahalanobis distances of each unit from
     its own population @
@ "nnn";; cols(mdt); rows(mdt); wait;
    mdt;  @
i=1; do until i==nn+1;
    MD[i]=MDt[i,ident[i]];
   post1[i]=post[i,ident[i]+1];  @ post1= nn x 1 vector of posterior
                                  prob.
                                 of each
                                 unit from its own universe @
   i=i+1;
endo;
 @ Call the procedure for testing the equality of var cov matrix @

 {eqva}=eqvar(Z,sz[2:rows(sz)],0);

  retp(post1,Lnd,M,MD,eqva);
 endp;

