/*
**  QQNORM.G
**  written by MARCO RIANI
**
**  FORMAT                      PURPOSE                     Line
** =======================================================================
** {tst}=QQNORM(X,d)         To Produce a graphical display to test 
**                           normalility distribution of data.
**
*/

/*
**
** QQNORM
**
** PURPOSE:   To Produce a graphical display to test
**            normality
**
** FORMAT:   {tst}=QQNORM(X,d)
**
** INPUT:     X = n x 1 vector of data
**            d = scalar.
**                if d=1  program displays on the screen the
**                quantile against quantile plot + a straight line
**                passing through the 1st and third quartile
**                is produced
**                else nothing is printed on the screen
**
** OUTPUT:    tst = n x 3 matrix which contains the inputs for
**                  the qqnorm plot
**                  1st col=  indexes of the units
**                            the order is based on X values
**                  2nd col=  input data for X axis
**                            We use F^{(-1)} (i-3/8)/(n+1/4)
**                            where F^{(-1)} is the inverse of the
**                            cumulative normal distribution
**                  3rd col=  input data for y axis
**                            This is simply input X ordered
**                REMARK: If the
**                underlying hypothesis of normality is true,
**                then the plot will be
**                approximately a straight line.  The extreme
**                points have more variability than points towards
**                the center. A plot from qqnorm that is bent down
**                on the left and bent up on the right means that
**                the data have longer tails than the Gaussian.
**                References:  Chambers, J. M., Cleveland, W. S., Kleiner, B. and Tukey,
**                             P. A. (1983). Graphical
**                             Methods for Data Analysis. Wadsworth, Belmont, California.
**
**
** SEE ALSO  SIMENV.G
**
*/


proc (1)=qqnorm(y,d);
 local n,tst,seq,X,ind,q1,q3,m,x0,y0,x1,y1,x2,y2;
    n=rows(y);
    ind=sortind(y);
    y=sortc(y,1);
    seq=seqa(1,1,n);
    x=cdfni((seq-3/8)/(n+0.25));  @ See Minitab Ryan et al. (1985) @
       @ REMARK: BMDP uses (seq-1/3)/(n+1/3) @
       @ See Mage (1982) The American Statistician for other choices @
    tst=ind~x~y;
    tst=sortc(tst,2);
     if d==1;
          _plctrl=-1;
      q1=trunc((n+2)/4);
      q3=n-q1+1;
      @ Slope (y1-y0)/(x1-x0) @
      m=(tst[q3,3]-tst[q1,3])/(tst[q3,2]-tst[q1,2]);
      x0=tst[q1,2]; y0=tst[q1,3];
      x1=minc(x);
      y1=m*(x1-x0)+y0;
      x2=maxc(x);
      y2=m*(x2-x0)+y0;
      @ _pline=1~6~tst[q1,2:3]~tst[q3,2:3]~1~2~6.4;   @
      _pline=1~6~x1~y1~x2~y2~1~2~6.4;
          xy(X,y);
      else;
      endif;
     retp(tst);
endp;


