/*
**  QUELPLOT
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {out}=quelplot(x,y,r,d);            To prepare the inputs for the
**                                      xy plot which contains the
**                                      confidence ellipses
*/


/*
**
** QUELPLOT
**
** PURPOSE:  to compute the inputs for the xy plot which
**           contains confidence elliplses
**
** FORMAT:   {out}=quelplot(x,y,r,d)
**
** INPUT:     x     = N x 1 vector of data
**            y     = N x 1 vector of data
**            r     = if r = scalar 0 means are used
**                    if r = scalar <> 0 univariate medians are used
**                    ir r = 2 x 1 vector, the vaues of this
**                    vector are coordinates for robust centroid
**
**            d     = scalar which specifies the confidence
**                    level of the outer ellipse (hinge)
**                    This coefficient is based on a F distribution
**                    with 2 and N-2 degrees of freedom
** OUTPUT:    out   = k x 1 which contains the indexes of the units
**                    identified as bivariate outliers
**
*/

/*** the following procedure calculates the quelplot ***/
proc (1)=quelplot(x,y,r,d);
 local Tx,Ty,Sx,Sy,F,i,
       E,ran,c,Em,R1,R2,Th,The1,The2,Xx,Yy,R1max,R2max,
       The1max,The2max,Xmax,Ymax,nn,Xs,Ys,ter,lin,linn,emax,simX,
      simY,wx,wy,min,max,cost,a,b,aut,costh,sinth,m,fuo,rot,
       new1,xnew,ynew,new2,ch,freq,jk,freq1,fre2,ysim,mess1,
       seq,t,n;

n=rows(x);

 /****** In the following lines there is a robust centroid *********/
 if r==0;
  Tx=meanc(x);
  Ty=meanc(y);
 elseif r==1;
  Tx=median(x);
  Ty=median(y);
 else;
  Tx=r[1]; Ty=r[2];
 endif;
 Sx=sqrt((x-Tx)'(x-Tx)/(n-1));    @ Sx is the robust std of x @
 Sy=sqrt((y-Ty)'(y-Ty)/(n-1));   @ Sy is the robust std of y @

 R=(x-Tx)'(y-Ty)/((n-1)*(Sx*Sy));   @ robust correlation @

 Xs=(x-Tx)/Sx;   @ robust standardization of x_i  @
 Ys=(y-Ty)/Sy;   @ robust standardization of y_i  @

 /* Now we calculate Mahalanobis distances from
 centroid previously defined */
 E=sqrt((Xs.^2+Ys.^2-2*R*Xs.*Ys)/(1-R^2));

 /** the 50th percentile of an F with 2 and rows(x)-2 d.o.f.***/
 ran=sortc(E,1);
 if (n/2-round(n-n/2))==0;
   Em=(ran[n/2]+ran[n/2+1])/2;
 else;
   Em=ran[(n+1)/2];
 endif;         @  Em  is the median  of vector E @


 /** the square of the MAha. distance is distributed like an F **/
 @ n number of rows of x (y) @
 eMAX= (2*(n-1))/(n-2)* d;
 @ d = constant which specifies the condidence level of the outer
      ellipse @
 emax=sqrt(emax);


  /*** code for the hinge  ***/
  R1=Em*sqrt((1+R)/2);
  R2=Em*sqrt((1-R)/2);
  th=seqa(0,2*pi/3600,3600);
  The1=R1*cos(th);
  The2=R2*sin(th);
  Xx=(The1+The2);
  Yy=(The1-The2);

      @  XX~YY;   WAIT;   @


  /*** code for the fence ***/
  R1max=Emax*sqrt((1+R)/2);
  R2max=Emax*sqrt((1-R)/2);
  The1max=R1max*cos(th);
  The2max=R2max*sin(th);
  Xmax=(The1max+The2max);
  Ymax=(The1max-The2max);

  /***** Xs and Ys are the robust standardized values ******/
  simX=Xs; simY=Ys;
  f=Xs~Ys;
  @ simX and simY are the symbols to be plotted in the main plot  @
  Wx=ones(rows(Xs),1);
  Wy=ones(rows(Xs),1);

 scale(Xmax|simX,Ymax|simY);
           begwind;
           window(1,1,0);
           makewind(80,6.8555,0,0,1);


  /**** plot of the hinge together with the fence ***/
  ter=1;
  lin=ones(rows(ter),1)~6*ones(rows(ter),1)~
      0~-10*ones(rows(ter),1)~0~100*ones(rows(ter),1)
      ~zeros(rows(ter),1)~7*ones(rows(ter),1)~0.2*ones(rows(ter),1);
   linn=ones(rows(ter),1)~6*ones(rows(ter),1)~
       -10*ones(rows(ter),1)~0~100*ones(rows(ter),1)~0
       ~zeros(rows(ter),1)~7*ones(rows(ter),1)~0.2*ones(rows(ter),1);
   _pline=lin|linn;


  /***** rotation of the coordinates ******/
  cost=EMAX^2;
  a=1/(cost*(1-r^2));
  c=a;
  b=-r*a;


  AUT=eig((a~b)|(b~c));
  costh=b/sqrt((c-aut[2])^2+b^2);
  sinth=(aut[2]-c)/sqrt((c-aut[2])^2+b^2);

  /****
  M is the orthogonal matrix which enables the rotation of the axes
  ****/
  M=(costh~sinth)|
  (-sinth~costh);

  /**** fuo contains the coord. of the fires of the ellipse */
  fuo=SQRT(ABS(1/AUT[1]-1/AUT[2]));


ROT=(xmax')|(ymax');
/*** NEW1 contains the coordinates of the rotated ellipse ***/
new1=M*rot;
xnew=new1[1,.]';
ynew=new1[2,.]';

/*** NEW2 contains the coordinates of the rotated points ***/
NEW2=M*(f');


/********* ch is the fixed distance from the fires of the ellipse
******/
   i=10;
ch=sqrt((xnew[i]+fuo)^2+ynew[i]^2)+sqrt((xnew[i]-fuo)^2+ynew[i]^2);

/*** freq e' is the vector which contains potential outliers
      for a particular pair of variables ***/


 freq=zeros(rows(x),1);

jk=1; do until jk==rows(f)+1;
if sqrt((new2[1,jk]+fuo)^2+new2[2,jk]^2)
+sqrt((new2[1,jk]-fuo)^2+new2[2,jk]^2)>ch;
freq[jk,1]=freq[jk,1]+1;
        else;
    endif;
    jk=jk+1;
endo;

freq1=freq;


seq=seqa(1,1,n);
ysim=delif(seq,freq1.==0);
@ missrv converts missing values in specified numbers @
ysim=missrv(YSIM,235455);

 F=SimX~SimY; @ SimX and Simy are the robust standardized vectors @
 @ _psym indicates the location of the points in the main plot @
_psym=F~1*ONES(n,1)~0.7*ones(n,1)~6*ONES(n,1)~
1.5*ONES(n,1)~4*ONES(n,1);
_pnumht=0.10|0.10;
_pltype=6|6;
    if YSIM[1].==235455;
@ if there is no bivariate outlier just plot the hinge and the fence @
xy(Xx~xmax,Yy~ymax);
else;
@ if there are some bivariate outliers check who they are @
 i=1;  do until i==rows(ysim)+1;

_PMSGCTL=xs[ysim[i]]~ys[ysim[i]]~0.1~0~1~2;
  mess1= 0 $+ FTOCV(YSIM[i],1,0);
_PMSGSTR=mess1;

xy(Xx~xmax,Yy~ymax);

i=i+1;
endo;
endif;
   endwind;
  retp(ysim);
    endp;
