/*
**  REGRESSI
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {B,Sres,R,NO,RA,SC}=regressi(y,X,d,pr);
**
**                                To extract some useful quantities
**                                from regression
*/

/*
** REGRESSI
**
** PURPOSE:  to performa linear least squares
**           and to extract a lot of quantities
**
** FORMAT:   {B,Sres,R,NO,RA,SC}=regressi(y,X,d,pr);
**
** INPUT:     y = n x 1 vector (response variable)
**            X = n x p data matrix of expl. variables;
**                n observations and p variables
**            d = scalar specifying whether or not y is a time series
**                if d==1 program produces:
**                Durbin Watson st + autocorrelation tests in matrix SC
**           pr = scalar
**                if pr=1 it is possible to see all the
**                statistic printed on the screen
**
**
** OUTPUT:  B =  p x 4 matrix
**               1st col = b coefficients
**               2nd col = Std. error for each b
**               3rd col = t-stat for each coeff.
**               4th col = P-value of t-stat
**          Sres =  n x 1 vector of standardized (studentized) residuals
**               (y_i -\hat{y}_i)/ (s *sqrt(1-h_i))
**               see Atkinson, 1985 pag. 18 eq. 2.1.14
**          R =  2 x 2 matrix
**               R[1,1] = R^2
**               R[2,1] = adjusted R^2
**               R[1,2] = est. of var. of res.
**                        (S^2= \sum_i^n  e_i^2/(n-k) )
**          NO = 3 x 2 matrix
**               1st row = asymm. test on residuals + pvalue
**               2nd row = kurt. test on residuals  + pvalue
**               3rd row = normality test on residuals + pvalue
**          RA = 2 x 2 matrix
**               1st row = Ramsey reset (LM) + p value
**               2nd row = Ramsey reset (mod. LM) + pvalue
**          SC = scalar  or 5x2 matrix
**               if d=0; Sc=0;
**               else d= 5 x 2 matrix
**               containing tests for serial correlation
**               1,1 = heterosk. test  (LM)       1,2 = p value for 1,1
**               2,1 = heterosk. test  (mod. LM)  2,2 = p value for 2,1
**               3,1 = serial corr. test (LM)     3,2 = p value for 3,1
**               4,1 = serial cor.  test (mod-LM) 4,2 = p value for 4,1
**               5.1 el = Durbin Watson test
**
**      REMARK: if sum of squared of res. is equal to zero
**              standard errors and t statistics are
**              missing values
**
**  See also: OLS
**
*/



proc (6)=regressi(y,X,d,pr);

  local T,p,C,A,B,H,yhat,df,u,rss,ssq,see,ymd,lik,tv,rsq,rbar,
        dw,fstat,pvf,R,RA,NO,k,vc,sqb1,b2,Wq,Wqq,clq,ys,us,bolss,
        ress,ttv,rnew,prl2,SC,seq,zer,WW,W,wsw,cl,i,sres;
T=rows(x);
p=cols(X);
   @ B = p x 4 matrix which will contain  beta coeff. std err. t stat
     and p values of tstat. @
   B=miss(zeros(p,4),0);
   @ R = 2 x 2 matrix  R^2 squared and adj. R^2  and S^2 @
   R=zeros(2,2);
   @ NO = 3 x 2 matrix (asymm. kurt and norm. 2nd col= p values @
   NO=zeros(3,3);
   @ RA = 2 x 2 matrix contianing ramsey test for functional form @
   RA=miss(zeros(2,2),0);
   @ SC = 5 x 2 matrix contianing serial correlation tests  @
   SC=miss(zeros(5,2),0);


df=T-p;     k=p-1;
C=invpd(X'X);
A=C*X';
H=X*A;
B[.,1]=A*y;
yhat=X*B[.,1];                     /* estimated values            */
u=y-yhat;                          /* residuals                   */
rss=u'u;                           /* residuals sum of squares    */
ymd=y-meanc(y);
tv=ymd'ymd;                        /* total sum of squares        */

if rss>0.000000001;
    /* estimate  of var. of error  */
    ssq=rss/df;  R[1,2]=ssq;
   /* S.E. of regression           */
    see=sqrt(ssq);
   /* Maximum of Log-likelihood   */
   lik=-T*ln(sqrt(rss/T)*(sqrt(2*pi)))-0.5*T;
    /* sres=standardized residuals */
    /* sres is missing value if some diagonal elements of matrix H
       are equal to 1  else is a vector */

       if maxc(diag(H))<1;
            sres=u./(see*sqrt(1-diag(H)));
       else;
            sres=miss(0,0);
       endif;


    R[1,1]=1-rss/tv;              /* R squared                */
    R[2,1]=1- ssq/(tv/(T-1));     /* Adjusted R square         */
    vc=ssq*C;                        /* var. co. matrix              */
    B[.,2]=sqrt(diag(vc));           /* st. err of bols             */
    B[.,3]=B[.,1]./B[.,2];           /* t-statistics            */
    if rss>0 and df>0 and k>0;
    fstat=((tv-rss)/(k))/(rss/df);   /* f-test                      */
    else;
    fstat=0;
    endif;
    pvf=cdffc(fstat,k,df);           /* pvalue of F-test            */
    B[.,4]=2*cdftc(abs(B[.,3]),df);  /* pvalof t ratios (2 sided test) */

               /* NORMALITY TEST */
    /* Measure of asymmetry */
    sqb1=(sumc(u.^3)/T) / (sumc(u.^2)/T)^(3/2);
    /* Measure of Kurtosis  */
    b2=(sumc(u.^4)/T) / (sumc(u.^2)/T)^2;
    /* Asymmetry test   */
    NO[1,1]=  (T/6)*  sqb1  ^2  ;
    NO[1,3]=sqb1; 
    /* Kurtosis test        */
    NO[2,1]=  (T/24)*((  b2  -3)^2)  ;
    NO[2,3]=b2;
    /* Normality test       */
    NO[3,1]=NO[1,1]+NO[2,1];
      if NO[1,1]>100;
         NO[1,2]=0;
      else;
         NO[1,2]=cdfchic(NO[1,1],1);       /* pvalue of As-test   */
      endif;
      if NO[2,1]>100;
         NO[2,2]=0;
      else;
         NO[2,2]=cdfchic(NO[2,1],1);        /* pvalue of Kur-test  */
      endif;
      if NO[3,1]>100;
         NO[3,2]=0;
      else;
          NO[3,2]=cdfchic(NO[3,1],2);      /* pvalue of Nor-test  */
      endif;
 else;
 sres=miss(0,0);
 R[1,1]=1;
 endif;
            /* Ramsey's RESET test of functional form   */
            /* Lagrange multiplier version  */
               Wq=yhat.^2;


                   Wqq = (Wq'Wq-Wq'H*Wq);
               if rss > 1e-6 and abs(Wqq)<1e+8 and Wqq ne 0;
                   RA[1,1]=T*((u'*Wq*inv(Wqq )*Wq'u)/(u'u) );
                /* pvalue for Lagrange multiplier test   */
                   RA[1,2]=cdfchic(RA[1,1],1);
                   clq=cols(Wq);
                /* Modified LM test */
                     if T ne RA[1,1];
                         RA[2,1]=((T-p-clq)/clq)*(Ra[1,1]/(T-Ra[1,1]));
                         /* p-value for modified LM test */
                         RA[2,2]=cdffc(RA[2,1],clq,T-p-clq);
                      else;
                      endif;
               else;
               endif;


  @ DO Heteroscedasticity  and residual corr. test only if d=1 @

  if d==1;
           format 8,12;
                   @ HETEROSKEDASTICITY TEST    @
        ys=ones(T,1)~(yhat.^2);
                             /* Based on the regression of  squared */
        us=u.^2;             /* residuals on squared fitted values */
        bolss=inv(ys'ys)*ys'us;
        c=inv(ys'ys);
        ress=(us-ys*bolss)'(us-ys*bolss);
        ttv=(us-meanc(us))'(us-meanc(us));
        SC[1,1]= T*(1-ress/ttv);     /* L. M. test based on T*R^2    */
        SC[1,2]=cdfchic(SC[1,1],1);      /* p value for L. M.  test      */
         @  Model1 = u^2 = \alpha + \beta * yhat^2
            fstat1= regression deviance of Model1 /
                   residual deviance of Model1.   @

        SC[2,1]=((ttv-ress)/(1))/(ress/(T-2));   /* Heteroscedasticity test */
        SC[2,2]=cdffc(SC[2,1],1,T-2);  /* pvalue of  Heteroscedasticity test */

                   @ SERIAL CORRELATION TEST @
        /* lagrange multiplier version of the test for residual serial
          correlazion  Godfrey 1978a 1978b, Econometrica */
        ww=u.*ones(T,p);
        seq=seqa(1,1,p);
        zer=seqa(0,0,p);
        w=shiftr(ww',seq,zer);  wsw=w';
        w=wsw[.,1];          /* take just the first column of wsw */
        /* Lagrange multiplier version   */
        SC[3,1]=T*( (u'*W*inv(W'W-W'H*W)*W'u)/(u'u) );
        /* pvalue for Lagrange multiplier test  */
        SC[3,2]=cdfchic(SC[3,1],1);
        cl=cols(w);
        SC[4,1]=((T-p-cl)/cl)*(sc[3,1]/(T-sc[3,1])); /* Modified LM test   */
        SC[4,2]=cdffc(SC[4,1],cl,T-p-cl);   /* p-value for modified LM test */

        SC[5,1]=(sumc((u[2:rows(x)]-u[1:rows(x)-1]).^2))/u'u; @ Durbin Watson @
    else;
  cl=miss(0,0); @ Initialize cl @
  endif;

if pr==1;
    @ Printing part @

"/****************************************************************************/
/*          ORDINARY LEAST SQUARES ESTIMATION                                */
/****************************************************************************/";
  print;  format /rd 10,0;
  "Observations :"   T;;  "       Degrees of freedom:"     df;
   format 10,3;
   "R-squared    :"  R[1,1];; "       Dw                :"     SC[5,1];
 "R-Bar-Squared:"  R[2,1];
 "Residual SS  :"  rss;; "       S.E. of Regression:"     see;
 "Total SS     :"  tv;;
  format 1,0;  "       F(" k "," df ")=";; format 4,2;  fstat;;
 format 4,2; /rd "   P-value="     pvf;
 format 10,3;
 "Max Log-lik  :"  lik;; "       Mean of Dep. Var. :"    meanc(y);
 print;
 "Var.         Coeff.       Std. Error      t-stat.       P-value";
  i=1;  do until i>k+1;
       format 4,0; /ld "X" i;;
       format 14,6; B[i,1];; B[i,2];; B[i,3];;
       format 11,3;  /rd B[i,4];
     i=i+1;
   endo;
                       format 7,5;
 "
 /***************************************************************************/
 /*          DIAGNOSTIC    TESTS                                            */
 /***************************************************************************/";

if d==1;
    "A:";;"Heterosced.test     LM ";;  format /m1 /rd 1,0;
     "CHI-SQ(" 1 ")=";;format 7,5; SC[1,1];;     "       Pvalue=";;
    format
    7,5;  SC[1,2];
    "B:";;"Heterosced.test     mod-LM";; format /rd  1,0; " F(" 1 ", " T-2
    ")= ";; format 7,5;  SC[2,1];;  "  Pvalue=";;  SC[2,2];
    "C:";;"Serial Correlation  LM ";;  format /m1 /rd 1,0;
     "CHI-SQ(" cl ")=";; format 7,5; SC[3,1];;"      Pvalue=";;
     format  7,5; SC[3,2];
      "D:";;"Serial Correlation  mod-LM";;
    format 1,0;" F(" cl ", " T-cols(x)-cl ")= ";;
    format 7,5;  SC[4,1];;  " Pvalue=";;  SC[4,2];
else;
endif;
 "E:";;"Ramsey RESET        LM ";;  format /m1 /rd 1,0;
  "CHI-SQ(" clq ")=";; format 7,5; Ra[1,1];;"       Pvalue=";;
 format
 7,5; RA[1,2];
 "F:";;"Ramsey RESET        mod-LM";;format 1,0;" F(" clq ", "
 T-cols(x)-clq ")";;
  "= ";; format 7,5; RA[2,1];; "  Pvalue=";; RA[2,2];

 "G:";;"Asymmetry  test     ";;    "CHI-SQ(1 )=";; NO[1,1];;
  "          Pvalue=";;NO[1,2];
 "H:";;"Kurtosis   test     ";;
   "CHI-SQ(1 )=";; NO[2,1];;"          Pvalue=";;NO[2,2];
 "I:";;"Normality  test     ";;
   "CHI-SQ(2 )=";; NO[3,1];;"          Pvalue=";;NO[3,2];
                  format /m1 /rd 9,6;
  else;
  endif;

   retp(B,Sres,R,NO,RA,SC);
 endp;

