/*
**  RFLR (Forward likelihood ratio)
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {pos,pot,lr,likmax,In,Un}=Rflr(Y,X,bsb,ch,lamax,d);
**
**                                To monitor the likelihodd ratio
**                                in every step of the forward procedure
**                                The purpose is to test the null
**                                hypothesis that the transformation
**                                parameters are as those specified in
**                                vector ch.
**                                Differently from procudure flr.g in this case
**                                we also assume that the columns of Y have a regression structure
**                                given by matrix X
**
*/

/*
** RFLR
**
** PURPOSE:  The principal purpose is to monitor the likelihood ratio in
**           every step of the forward procedure. The purpose is to test
**           the null
**           hypothesis that the transformation parameters are as those
**           specified in vector la. Test is based on lik. ratio of
**           Mahalanobis distances of residuals in each step of the
**           forward search
**           It also produces average, maximum of Mahalanobis distances
**           of clean (and min of noclean) in  each step of
**           the forward procedure
**
** FORMAT:   {pos,pot,lr,likmax,In,Un}=Rflr(Y,X,bsb,ch,lamax,d);
**
** INPUT:     Y = n x p matrix of response variables
**            X = n x q data matrix; n observations
**                 and q explanatory variables
**          bsb = m x 1 vector. It contains
**                the indexes of the m units which are
**                declared as potential outliers.
**                Initial centroid is determined skipping
**                these units.
**           ch     = c x 2 vector ( 1 \leq c \leq p)
**                    1st column = indexes of the variables (VAR)
**                                 whose MLE of transformation
**                                 parameter is required
**                    2nd column = values of the transformation
**                                 parameters to be used for ordering MD
**                    REMARK: if ch=0 (scalar zero) flr
**                             concerns all variables in matrix Y and
**                             the fwd search is done using
**                             untranformed values
**                    EXAMPLE: if ch=(1~0)|
**                                   (4~0.5)
**                             Maximation concerns VAR 1 and 4
**                             Ordering of MD in each step of the fwd
**                             search uses log transformation for VAR 1
**                             and sqrt transformation for VAR 4
**        lamax = (n-m+1) x p matrix which contains (in the rows) the
**                values of MLE transformation parameters in each step
**                of the forward search
**           d  = l x 1 vector which enables the user to monitor the
**                likelihood grid of the parameters.
**                For examples if l={24,33} you can monitor the
**                likelihood grid when there are respectively 24 and 33
**                observations in the basic subset
**                The program also drwas vertical bands in
**                corredpondence of the 95% confidence
**                interval observations for the transformation parameter
**                of each variable
**
**
** OUTPUT:  pos =  (n-m+1) x 3 matrix.  The first column of pos
**                 contains the
**                 dimension  of the sample used to compute Mahalanobis
**                 distances Columns 2,3 respectively contain
**                 Max and Average of Mahalanobis distances of the units
**                 forming the basic subset
**          pot =  (n-m) x 2 matrix. The first column contains the
**                 dimension of the sample used in order to compute
**                 Mahalanobis distances, the second
**                 reports the minimum of Mahalanobis distances among
**                 the units which form the group of potential
**                 outliers
**                 Remark: if n=m pot is a missing value
**            lr = (n-m+1 x 2) matrix. The first column contains the
**                 dimension of
**                 the basic subset, the second the values of the
**                 likelihood ratio in each
**                 step of the forward search
**        likmax = (n-m+1 x 2) matrix. the first column contains the
**                 size of the basic subset, the second the value of the
**                 maximazied likelihood
**           In  = (n-m+1) x 11 matrix
**                 1st col = fwd search index
**                 2nd col = unit with m-th MD (biggest MD in bsb)
**                 3rd col = unit with (m+1)-th MD (smallest among
**                           the units which are not in the subset
**                  ......
**                 11th col = unit with (m+10)-th MD
**           Un = (n-m) x 11 Matrix which contains the unit(s)
**                included in bsb at each step of the fwd search
**                REMARK: in every step the new subset is compared
**                with the old subset. Un contains the unit(s) present
**                in the new subset but not in the old one
**
**    REMARK: The forward search is based on Mahalanobis distances
**            of the residuals of the regressions
**
**    See Also: FLR, FLRLD, FLRQD
*/


 proc (6)=Rflr(Y,X,bsb,ch,lamax,d);

   local seq,stal,stal1,stal2,ij,ym,vc,
       nn,n,p,mm,i,gg,pos,pot,zs,min,mala,Yb,Cb,B,E,ii,lr,mess,
       j,G,Xb,dZ,dZmax,Zmax,la1,gri,lat,k,Zt,Ztmax,Likmax,nla1,Hb,Yh,
       Eg,Egmax,Yhmax,In,Un,Unit,nbsb,Yt,la,jj,sel,lim,confint,on;
nn=rows(X);
p=cols(Y);
seq=SEQA(1,1,nn);

mm= rows(bsb);
min=mm;
@   The first column of pos contains the dimension
    of the sample used to compute Mahalanobis distances
    Columns 2,3 respectively contain
    Max and Average of Mahalanobis distances of the units
    forming the basic subset
    Column 4 contains Max of Mahalanobis distances of all
    the units  @
  pos=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,2);

  if mm <nn;
    @ pot has two columns the first contains the dimension of the
      sample used in order to compute Mahalanobis distances, the second
      reports the minimum of Mahalanobis distances among the units
      which form the group of potential outliers @
    pot=seqa(mm,1,nn-mm)~zeros(nn-mm,1);

    @    Un= Matrix which contains the unit(s) included in
             bsb at each step of the fwd search @
    Un=seqa(mm+1,1,nn-mm)~miss(zeros(nn-mm,10),0);
  else;
    pot=miss(0,0);
    Un= miss(0,0);
  endif;
  @ Matrix lr which will contain the LR test in each step of the
    forward search @
  lr=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,1);

  @ matrix in will contain in the 2nd col. the unit which has enterd
    the basic subset, in third col the one with the smallest MD among
    Noclean, in 4th col the unit with second smallest MD among nocl. @
  In=pos[.,1]~miss(zeros(nn-mm+1,10),0);


  @ Matrix likmax which will contain the max of the profile
  likelihood in each step of the
  forward search @
       Likmax=lr;

/*** mala in the column will contain MD  ***/
mala=seq~zeros(nn,1);

       if ch==0 and rows(ch)==1;
           sel=seqa(1,1,p);
           la=ones(p,1);
       else;
           la=ch[.,2];
           sel=ch[.,1];
       endif;

          @ Create Yt=matrix of transformed values to be used
             in the forward search (for ordering MD) @
          Yt=Y;
        j=1; do until j==rows(sel)+1;
           if la[j] NE 0;
                  Yt[.,sel[j]]= Y[.,sel[j]]^la[j];
              else;
                  Yt[.,sel[j]]=ln(Y[.,sel[j]]);
             endif;
           j=j+1;
        endo;

mm=mm; do until mm==nn+1;
                 Xb=X[bsb,.];
                 Yb=Y[bsb,.];
                   Cb=invpd(Xb'Xb)*Xb';
                B= Cb*Yt[bsb,.];
               E=Yt-X*B; @ residuals for the overall sample @
               ym=meanc(E[bsb,.]);
               vc=invpd(vcx(E[bsb,.]));

  Zt=Yb;
  Ztmax=Zt;


 jj=1; do until jj==rows(ch)+1;
j=sel[jj];

G=exp(meanc(ln(Yb[.,j])));
     if la[jj] NE 0;
             Zt[.,j]=(Yb[.,j]^la[jj]-1)/(la[jj]*G^(la[jj]-1));
       else;
             Zt[.,j]=G *ln(Yb[.,j]);
      endif;
    if lamax[mm-min+1,j] NE 0;
             Ztmax[.,j]=(Yb[.,j]^lamax[mm-min+1,j]-1)/(
                    lamax[mm-min+1,j]*G^(lamax[mm-min+1,j]-1));
       else;
             Ztmax[.,j]=G*ln(Yb[.,j]);
      endif;

    jj=jj+1;
 endo;
             Hb=Xb*Cb;
             Yh=Hb*Zt;

         Eg=Zt-Yh;   @ residuals in correspondence of null @
            Yhmax=Hb*Ztmax;
         Egmax=Ztmax-Yhmax;  @ residuals in correspondence of MLE @

       dZ=det(vcx(Eg));
       dZmax=det(vcx(Egmax));

      lr[mm-min+1,2]=mm*ln(dZ/dZmax);
      Likmax[mm-min+1,2]=-mm*ln(dZmax);


       if maxc(mm.==d)==1;
        @ vector on will be used in matrix _pline to draw
          99% confidence bands of transformation parameter @
         on=ones(3,1);
       @ The following if checks if the user has specified a
         common transformation for all the parameters @
    if minc(maxc(lmax[.,2:cols(lmax)]')-minc(lmax[.,2:cols(lmax)]'))>1e-7;
           begwind;
            if rows(la)<=3;
              window(2,rows(la),0);
            elseif rows(la)<=4;
              window(2,2,0);
            elseif rows(la)<=6;
              window(2,3,0);
            elseif rows(la)<=12;
              window(3,4,0);
            elseif rows(la)<=15;
              window(3,5,0);
            else;
              window(4,6,0);
            endif;
            makewind(80,6.8555,0,0,1);

             la1=seqa(-1,0.001,2100);
             nla1=rows(la1);
             xtics(la1[1],la1[nla1],0.1,1);
             Gri=zeros(rows(la1),1);

               i=1; do until i==p+1;

                   @ vector lim will contain the x coordinates of the
                     confidence interval for lambda
                     lim[1,1] = lower limit
                     lim[2,1] = lower limit
                     lim[3,1] = mle   @
                   lim=miss(zeros(3,1),0);

                     G=exp(meanc(ln(Yb[.,i])));
                     Zt=Ztmax;   k=1;   confint=0;
                        j=1; do until j ==nla1+1;
                           if abs(la1[j]) >1e-6;
                    Zt[.,i]=(Yb[.,i]^la1[j]-1)/(la1[j]*G^(la1[j]-1));
                           else;
                             Zt[.,i]=G *ln(Yb[.,i]);
                           endif;

                           Yh=Hb*Zt;
                           Eg=Zt-Yh;
                           Gri[k]=-mm*ln(det(vcx(Eg)));

                            @ The following if finds the conf. interval
                              for each transformation parameter @
                            if Likmax[mm-min+1,2]-Gri[k]<=0.5*3.84
                                   and confint==0;
                                    lim[1]=la1[j];
                                    confint=1;
                            elseif Likmax[mm-min+1,2]-Gri[k]>=0.5*3.84
                                   and confint==1;
                                    lim[2]=la1[j];
                                  confint=2;
                                         @  "DIFFF";
                                     lim[1]-lamax[mm-min+1,i];
                                     lamax[mm-min+1,i]-lim[2]; wait; @
                            else;
                            endif;

                           j=j+1;   k=k+1;
                           endo;
                       _pnumht=0.3;
                      if i==1;
                   ytics(round(minc(Gri))-12,round(maxc(gri))+3,10,1);
                      else;
                       endif;
            mess= 0 $+ "m=" $+ ftocv(mm,2,0) $+ " " $+ "y" $+
                             FTOCV(i,1,0);
                            _PMSGSTR =mess;
                       _PMSGCTL=0.7~1~0.3~0~2~2;
@ "variable";; i;
lim; wait; @
                   lim[3]=lamax[mm-min+1,i];
                   
                _pline=on~6*on~lim~-1e+10*on~lim~1e+10*on~2*on~
                        2*on~0.4*on;
             @   output file=limits on;
                           lim;
                  output file=limits off;  @

              @     " variable";; i; "Gri"; la1~Gri; wait;  @
                  xy(la1,Gri);
            nextwind;
               i=i+1;
               endo;
            endwind;
     else; @ if there is a common transfomation @
                   @ vector lim will contain the x coordinates of the
                     confidence interval for lambda
                     lim[1,1] = lower limit
                     lim[2,1] = lower limit
                     lim[3,1] = mle   @
                   lim=miss(zeros(3,1),0);

             la1=seqa(-1,0.001,1100);
             nla1=rows(la1);
             xtics(la1[1],la1[nla1],0.1,1);
             Gri=zeros(rows(la1),1);


                     G=exp(meanc(ln(Yb)));
                     Zt=Ztmax;   k=1;   confint=0;
                        j=1; do until j ==nla1+1;
                           if abs(la1[j]) >1e-6;
                            Zt=(Yb^la1[j]-1)./(la1[j]*(G')^(la1[j]-1));
                           else;
                             Zt=G' .*ln(Yb);
                           endif;

                           Yh=Hb*Zt;
                           Eg=Zt-Yh;
                           Gri[k]=-mm*ln(det(vcx(Eg)));

                            @ The following if finds the conf. interval
                              for each transformation parameter @
                            if Likmax[mm-min+1,2]-Gri[k]<=0.5*3.84
                                   and confint==0;
                                    lim[1]=la1[j];
                                    confint=1;
                           @ Likmax[mm-min+1,2]; Gri[k]; la1[j]; wait; @
                            elseif Likmax[mm-min+1,2]-Gri[k]>=0.5*3.84
                                   and confint==1;
                                    lim[2]=la1[j];
                                  confint=2;
                           @ Likmax[mm-min+1,2]; Gri[k]; la1[j]; wait; @

                           lim[3]=lamax[mm-min+1,2];
                               @   "mm-min+1";; mm-min+1;  wait;
                                  lamax[mm-min+1,.]; wait;
                                           "DIFFF";
                                     lim[1]-lamax[mm-min+1,i];
                                     lamax[mm-min+1,i]-lim[2]; wait; @
                            else;
                            endif;

                           j=j+1;   k=k+1;
                           endo;

                       _pnumht=0.3;
                       ytics(round(minc(Gri))-20,round(maxc(gri))+5,2,1);
                   mess= "m=" $+ ftocv(mm,3,0) $+ " " $+ "la=" $+
                             FTOCV(la[1],1,2) $+ "  Common transf.";
                            _PMSGSTR =mess;
                       _PMSGCTL=0.7~1~0.3~0~2~2;


                _pline=on~6*on~lim~-1e+10*on~lim~1e+10*on~2*on~
                        2*on~0.4*on;

              @     " variable";; i; "Gri"; la1~Gri; wait;  @
                  xy(la1,Gri);

     endif;

         else;
       endif;



i=1;  do until i==nn+1;
mala[i,2]=sqrt((E[i,.]-ym')*vc*(E[i,.]'-ym));
            i=i+1;
         endo;
zs=sortc(mala,2);

   @ pos has 3 columns: the first contains the dimension of clean data
      set, Columns 2, and 3 respectively contain
    Max, Average, of MD of residuals of the units
    forming the basic subset @

                @   maximum       average  @
   pos[mm-min+1,2:3]=zs[mm,2]~meanc(zs[1:mm,2]);

   zs=zs|miss(zeros(10,2),0);
   IN[mm-min+1,2:11]=(zs[mm:mm+9,1]');

     if mm<nn;
       @ pot contains minimum of Mahalanobis distances among the units
         which form the group of potential outliers @
       pot[mm-min+1,2]=zs[mm+1,2];

     @ the dimension of the clean data set increases of one unit.
       vector bsb contains the indexes of the units
       forming the new basic subset @
     nbsb=zs[1:mm+1,1];

             @ Store in matrix Un the unit(s) included in this step @
             unit=setdif(nbsb,bsb,1);
             if rows(unit)<=10;
              UN[mm-min+1,2:rows(unit)+1]=(unit');
             else;
              "Warning: interchange greater than 10";
              UN[mm-min+1,2:11]=unit[1:10]';
             endif;
      bsb=nbsb;
                      else;
     endif;


                    mm=mm+1;
                  endo;
  @   graphset;  @
   retp(pos,pot,lr,likmax,In,Un);
endp;

