/*
**  RFWDMLE
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {mL}=Rfwdmle(Y,X,bsb,ch);           To create a matrix which
**                                     contains MLE of transformation
**                                     parameters at each step
**                                     of the forward search
**                                     Differently from procudure fwdmle.g in this case
**                                     we also assume that the columns of Y have a regression structure
**                                     given by matrix X
**
**
*/

/*
**
** RFWDMLE
**
** PURPOSE:  to create a matrix which contains Max. Lik. Est.
**           of transformation parameters at each step of
**           the forward search
**
**
** FORMAT:    {mL}=Rfwdmle(Y,X,bsb,ch);
**
** INPUT:
**            Y     = n x k matrix of response variables
**            X     = n x p matrix of the data
**           bsb    = m x 1 vector which contains the indexes of the
**                    units forming the initial subset
**           ch     = c x 2 vector
**                    1st column = indexes of the variables (VAR)
**                                 whose MLE of transformation
**                                 parameter is required
**                    2nd column = values of the transformation
**                                 parameters to be used for ordering MD
**                    REMARK: if ch=0 (scalar zero) maximization
**                             concerns all variables in matrix Y and
**                             the fwd search is done using
**                             untranformed values
**                    EXAMPLE: if ch=(1~0)|
**                                   (4~0.5)
**                             Maximation concerns VAR 1 and 4
**                             Ordering of MD in each step of the fwd
**                             search uses log transformation for VAR 1
**                             and sqrt transformation for VAR 4
**
** OUTPUT:    Plik  = (n-m+1)  x (p+1) matrix. In the first column
**                    it contains the dimension of the basic subset used
**                    in the forward search. From column 2 to (p+1) we
**                    can find max. lik. estimates of the transformation
**                    parameters in each step of the forward search
**
** REMARKS: the input vector bsb can be conveniently created by
**          program UNIBIV.G
**          The transformtion used in the maximization is 2*sin(\lambda)
**          The initial \lambda values are =1 (no transformation)
** BACKGROUND:
**          The forward search refers to Mahalanobis distances on
**          residuals from each regression
**
** WARNINGS: three global variables named xglobal and chglobal
** and Yglobal are created
**
**
*/

 clearg Xglobal,Yglobal,chglobal;


proc (1)=assr(la0);    /* procedure which transforms the parameters */
     local j,nla,la;
         nla=rows(la0);
         la=zeros(nla,1);
              j=1; do until j==nla+1;
                          la[j]=   3*sin(  la0[j]  )  ;
                          @ la[j]=la0[j]/sqrt(1+la0[j]^2); @
              j=j+1;
              endo;
    retp(la);
endp;


proc (1)=likr(la);   @ procedure which computes max. lik. @
     local j,nla,G,z,nXc,dZ,p,i,Xg,Yg,Yh,Eg,seqp,dif;
           Xg=Xglobal; Yg=Yglobal;
            p=cols(Yg); seqp=seqa(1,1,p);
             {la}=assr(la);
                  G=exp(meanc(ln(Yg[.,chglobal])));
                   G=G'; @ Vector of geometric means @
             @ transform  Yg[.,chglobal] @
       Yg[.,chglobal]=(Yg[.,chglobal].^(la')-1)./((la').*G.^((la')-1));
       if rows(chglobal)<p;
           dif=setdif(seqp,chglobal,1);
           Yg[.,dif]=Yg[.,dif]-1;
       else;
      endif;


       Yh=Xg*inv(Xg'Xg)*(Xg')*Yg;
       Eg=Yg-Yh;

       dZ=100*det(vcx(Eg));

   retp(dZ);
endp;

proc (1)=Rfwdmle(Y,X,bsb,ch);
       local nn,seq,Plik,mm,la0,zs,mala,min,dZ,E,B,Xb,Yb,Ytb,Yt,sel,
             psi,fvalue,grad,retcode,vc,i,ym,psin,p,pp,geom,prov,la,j;
          nn=rows(X);
          p=cols(X);
          seq=SEQA(1,1,nn);
          mm= rows(bsb);
          min=mm;
          pp=cols(Y);

       if ch==0 and rows(ch)==1;
           sel=seqa(1,1,pp);
           la=ones(pp,1);
       else;
           la=ch[.,2];
           sel=ch[.,1];
       endif;
           chglobal=sel;

          Plik=seqa(mm,1,nn-mm+1)~ones(nn-mm+1,pp);

          /*  calculation of Mahal. dist. on initial clean data set */
           mala=seq~zeros(nn,1);
           la0=ones(rows(chglobal),1);
           la0=sin(la)+1e-4;

          @ Create Yt=matrix of transformed values to be used
             in the forward search (for ordering MD) @
          Yt=Y;
        j=1; do until j==rows(sel)+1;
           if la[j] NE 0;
                  Yt[.,sel[j]]= Y[.,sel[j]]^la[j];
              else;
                  Yt[.,sel[j]]=ln(Y[.,sel[j]]);
             endif;
           j=j+1;
        endo;


       mm=mm; do until mm==nn+1;  @ nn= numero tot. di unita' stat. @
           @     "mm"; mm;
               rows(bsb);
                sumc(bsb); vcx(bsb); wait;  @

               if mm>min;
                  la0=psi;
               else;
               endif;

                  Xglobal=X[bsb,.];
                  Yglobal=Y[bsb,.];
                   Xb=Xglobal;
                   Yb=Yglobal;
                  {dZ}=likr(la0);
                  Ytb=Yt[bsb,.];
                  {psi,fvalue,grad,retcode}=OPTMUM(&likr,la0);
                   @ Retransform the MLE parameters @
                            {psin}=assr(psi);
                    Plik[mm-min+1, (chglobal)+1]=(psin');

                B= invpd(Xb'Xb)*Xb'*Ytb;
                @ residuals for the overall sample
                  using transformed Y @
                E=Yt-X*B;
               if cols(E)>1;
                    @ Search on Mahalanobis distances @
                       ym=meanc(E[bsb,.]);
                      vc=invpd(vcx(E[bsb,.]));
                   @ Mahalanobis dist. on residuals @
                                    i=1;  do until i==nn+1;
                         mala[i,2]=sqrt((E[i,.]-ym')*vc*(E[i,.]'-ym));
                                         i=i+1;
                                     endo;
                 else;
                  mala[.,2]=abs(E);
                 endif;
                      zs=sortc(mala,2);
                        @ the dimension of the clean data set increases
                          of one unit. vector bsb contains the
                          indexes corresponding to the units of
                          the new basic subset @
                                if mm<nn;
                                    bsb=zs[1:mm+1,1];
                             @  setdif(seqa(1,1,nn),bsb,1); wait;  @
                                 else;
                                 endif;

             mm=mm+1;
          endo;
  retp(Plik);
endp;

