/*
**  ROB
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
**
**  {mE,Blad}=rob(X,y,la,d)         To calculate robust regression
**                                  coefficients which can be
**                                  conveniently used as starting
**                                  points for the fwd search
**
*/


/*
**
** ROB
**
** PURPOSE:  to calculate regression coeff. using robust methods
**
**
**
** FORMAT:
**
** {mE,Blad} = rob(X,y,la,d)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             la = k x 1  vector which specifies for which values
**                  of the transformation parameter \lambda
**                  you want to compute the robust regression coeff.
**             d  = scalar which specifies the kind of robust regression
**                  if d=1; beta coeff. (b) minimizes
**                          \sum_{i=1}^n |y_i - b * x_i|
**                          b is found using procedure optmum
**                  if d=2; beta coeff. (b) minimizes
**                          \sum_{i=1}^q (y_i - b * x_i)^2
**                               where q= [n/2]+ 1
**                  if d=3; beta coeff. (b) minimizes
**                          \sum_{i=1}^q (y_i - b * x_i)^2
**                               where q= [n/2]+ [(p+1)/2]
**                  else;   beta coeff. (b) minimize
**                          \sum_{i=1}^n |y_i - b * x_i|
**                          Differently from d=1 in this case
**                          beta coeff. are solved as
**                          a re-iterated weighted
**                          least squares problem where the weights are
**                          the inverse of the absolute values of the
**                          residuals. (Suggestion of Schoenberg)
**
** OUTPUT     mE = n x k matrix which contains residuals in
**                 correspondence of  of the b_lad coeff.
**          Blad = p x k matrix which contains the \beta coeff.
**                 such that
**                 f(y_i , x_i'b_lad) is minimized.
**                 The kind of function f is specified in
**                 input scalar d
**                 Each column of matrix Blad corresponds to
**                 a particular value of vector la
**
**   See also: LMS, LMSG
**
**  REMARK 1  this program does not add a column of 1 in matrix X to
**            create the intercept.
**
*/


proc (2)=rob(X,y,la,d);
   local n,k,G,b,i,med,mlms,slms,jj,nla,Blad,mE,ny,b_old,b_new,w,
         fvalue,grad,retcode,mlogl;
 n=rows(x);
 k=cols(x);
 nla=rows(la);
 G=exp(meanc(ln(y))); @ Geometric mean of th observations @
 Blad=zeros(k,nla);   @ Initialize the matrix which contains `betas @
 mE=zeros(n,nla);     @ initialize the matrix which contains the
                        residuals @

 jj=1; do until jj==nla+1;
            if abs(la[jj])<0.0001;
              ny=G*ln(y);
            else;
              ny=(y.^la[jj]-1)/(la[jj]*G^(la[jj]-1));
            endif;

          if d==1;
            clearg yglobal,Xglobal;
            yglobal=ny;
            Xglobal=X;
                b_old=invpd(X'X)*X'ny;
                {mlogl}=minabs(b_old);
                       @ "b_old";; b_old; wait;   @
              {b_new,fvalue,grad,retcode}=OPTMUM(&minabs,b_old);
          elseif d==2 or d==3;
            clearg yglobal,Xglobal,dglobal;
            yglobal=ny;
            Xglobal=X;
            dglobal=d;
               b_old=invpd(X'X)*X'ny;
               {mlogl}=mintri(b_old);
                       @  "b_old";; b_old; wait;  @
              {b_new,fvalue,grad,retcode}=OPTMUM(&mintri,b_old);
          else;
               b_old = 0;
               b_new = invpd(X'X)*X'ny;  @ Initialize b_new with b_ols @
               w = X;
               do until abs(b_new - b_old) < 1e-12;
                   b_old = b_new;
                   b_new = invpd(w'X)*(w'ny);
                   w = X./abs(ny - X * b_new);
               endo;
          endif;

   @ Matrix Blad contains the robust beta coeff.
     Each column corresponds to a value of trasf. parameter @
   Blad[.,jj]=b_new;
   @ Store residuals @
   mE[.,jj]=ny-X*b_new;

               jj=jj+1;
         endo;

      retp(mE,Blad);
  endp;

/* This are the procedures which  creates the function to minimize */
clearg yglobal,Xglobal;

proc (1)=minabs(psi);
local mlogl;
  mlogl=sumc( abs(yglobal-xglobal*psi) );
    retp(mlogL);
endp;

proc (1)=mintri(psi);
local e,e1,mlogl,n,q,seq,p;
  n=rows(yglobal);
  p=cols(Xglobal);
  seq=seqa(1,1,n);
  e=yglobal-xglobal*psi;
  e1=sortc(seq~e,2);
  if dglobal==2;
     q=trunc(n/2)+1;
  else;
     q=trunc(n/2)+ trunc((p+1)/2);
  endif;
  mlogl=sumc( (yglobal[e1[1:q,1]]-Xglobal[e1[1:q,1],.]*psi).^2 );
    retp(mlogL);
endp;

