/*
** SCGLM
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** ======================================================================
**
**  {Tl} =scglm(X,y,q,fr,b_new,phi,off)     To test the link in GLM
**                                          through the score test
**
*/


/*
**
** SCGLM
**
** PURPOSE:  to calculate the score test for GLM
**
**
**
**
** FORMAT:
**
** {Tl} = scglm(X,y,q,fr,b_new,phi,off)
**
** INPUT:      X  = n x p matrix of explanatory variables
**             y  = n x 1 vector which contains the response variable
**             q  = scalar which specifies the kind of GLM
**                  q=1 ==> Poisson link
**                  q=2 ==> Logit link
**                  q=3 ==> Complementary Log log link
**                  q=3.05 ==> Gamma link (Box Cox la=2)
**                  q=3.1 ==> Gamma link (Box Cox la=1)
**                  q=3.2 ==> Gamma link (Box Cox la=0.5)
**                  q=3.3 ==> Gamma link (Box Cox la=0)
**                  q=3.4 ==> Gamma link (Box Cox la=-0.5)
**                  q=3.5 ==> Gamma link (Box Cox la=-1)
**                  q=3.6 ==> Gamma link (Power family la=3)
**                  q=3.7 ==> Gamma link (Power family la=2)
**                  q=3.8 ==> Gamma link (Power family la=1)
**                  q=3.9 ==> Gamma link (Power family log)
**                  q=4   ==> Gamma link (Power family la=-1)
**                            (inverse)
**                  q>4  and q<5 Gamma link all values of lambda
**                       between 0 and -1
**                       For example q=4.25 ==> la=-0.25
**                       For example q=4.5 ==> la=-0.5
**                  q=5 ==> Inverse Gaussian link
**                  q=6 ==> Probit link
**                  q=7 ==> Arc sin link
**                  q=8 ==> Modified logit link
**             fr = n x 1 vector of frequencies
**                  if fr = scalar 0 a n x 1 vector of 1s
**                  is created automatically by the program
**          b_new = p x 1 vector of coefficients
**            phi = scalar, estimate of dispersion parameter
**           off = n x 1 vector which defines an offset
**                 REMARK: off = scalar 0 implies no offset
**
**
** OUTPUT:     Tl = 1 x 4 matrix
**               Tl[1,1]= value of the coeff. of the constructed
**                        variable
**               Tl[1,2]= standard erroe
**               Tl[1,3]= score test for goodness of link (t stat)
**               Tl[1,4]= p value
**
**
*/
 proc (1)=scglm(X,y,q,fr,b_new,phi,off);
    local n,p,th,mu,dm,V,z,W,Xb,CXb,Cb,b_add,Tl,end,i,lam;


n=rows(X);
p=cols(X);
mu=zeros(n,1);
dm=mu; @ Initialization of dm @
V=mu;

   if q==3.05;
       lam=2;
   elseif q==3.1;
       lam=1;
   elseif q==3.2;
       lam=0.5;
   elseif q==3.4;
       lam=-0.5;
   elseif q==3.5;
       lam=-1;
   else;
   endif;


Tl=miss(zeros(1,4),0);
   if fr==0;
     fr=ones(n,1);
   else;
   endif;
              @ Matrix Tl will contain the quantities of
                  interest   @

                   th=X*b_new+off;
                 if q==1;     @ Poisson link @
                        i=1; do until i==n+1;
                           if th[i]>25;
                             mu[i]=1e+10;
                             dm[i]=1e-10;
                             V[i]=1e+10;
                           elseif th[i]<-25;
                             mu[i]=1e-10;
                             dm[i]=1e+10;
                             V[i]=1e-10;
                           else;
                             mu[i]=exp(th[i]);
                             dm[i]=1/mu[i];
                             V[i]=mu[i];
                           endif;
                          i=i+1;
                         endo;
                 /********
                   mu=exp(th);   @  "mu";; mu; wait;  @
                   dm=1/mu;      @ dm= derivative of \eta wrt mu @
                   V=mu;         @ V= variance @
                  ************/
                 elseif q==2; @ Logit  link @
                   mu=exp(th)./(1+exp(th));
                   dm=1/((mu+1e-12).*(1.000000001-mu));
                   V=mu.*(1-mu) ./fr + 1E-12;
                   @ V=variance (fr are the frequencies) @

                 elseif q==3; @ Complementary log log link @
                        i=1; do until i==n+1;
                           if th[i]>5;
                             mu[i]=1-1e-10;
                           elseif th[i]< -18;
                             mu[i]= 1e-10;
                           else;
                             mu[i]=1-exp(-exp(th[i]));
                           endif;
                          i=i+1;
                         endo;
                    @    mu=1-exp(-exp(th));   @
                   dm= -1/( (1-mu+1e-12).*ln(1-mu+1e-12) );
                   V=mu.*(1-mu) ./fr+ 1e-12;

                 elseif q>3 and q <=3.5;  @ GAMMA LINK (Box Cox) @
                   if q ne 3.3;
                    mu=  (1+lam*th)^(1/lam);
                    dm=  mu^(lam-1);
                   else;
                    mu=  exp(th);
                    dm= 1/mu;
                   endif;
                    V=(mu^2)./fr;


                 elseif q==3.6;  @ GAMMA LINK (third power) @
                   mu=  th^(1/3);
                   dm=  3*(mu^2);
                   V=(mu^2)./fr;

                 elseif q==3.7;  @ GAMMA LINK (square) @
                   mu=  sqrt(abs(th));
                   dm=  2*mu;
                   V=(mu^2)./fr;

                 elseif q==3.8;  @ GAMMA LINK (identity) @
                   mu=  th;
                   dm=   1;
                   V=(mu^2)./fr;

                 elseif q==3.9;  @ GAMMA LINK (log) @
                   mu=  exp(th);
                   dm=   1/mu;
                   V=(mu^2)./fr;

                 elseif q==4; @ GAMMA LINK (inverse) @
                   mu=  1/th;
                   dm= -  1/(mu^2);
                   V=(mu^2)./fr;
/***

                 elseif q==4.2; @ GAMMA LINK (-0.5) @
                   mu= 1/(th^2);
                   dm= -0.5*(mu^(-1.5));
                   V=(mu^2)./fr;
 ***/
                 elseif q>4 and q<5; @ GAMMA LINK (4+lambda) @
                   mu= 1/(th^((1/(q-4))));
                   dm= (4-q)*(mu^(3-q));
                   V=(mu^2)./fr;

                 elseif q==5;  @ INVERSE GAUSSIAN @
                   mu= 1/sqrt(th);
                   dm=  -2/(mu^3);
                   V= mu^3;

                 elseif q==6; @ PROBIT LINK @
                   mu= cdfn(th);  @ CDF evaluated in th @
                   dm=1/pdfn(TH);  @ inv. of density evaluated at th @
                   V=mu.*(1-mu)./fr;
                 elseif q==7; @ ARC SIN LINK @
                  i=1; do until i==n+1;
                    if th[i]>= pi/2;
                     mu[i]=1- 1e-8;
                     dm[i]=1e+6;
                    elseif th[i]<= - pi/2;
                     mu[i]=0 +1e-8;
                     dm[i]=1e+6;
                    else;
                     mu[i]=  0.5 * (1+sin(th[i]));
                     dm[i]=  2/sqrt(1-((2*mu[i]-1)^2));
                    endif;
                   i=i+1;
                  endo;
                  V= 1e-6+ mu.*(1.-mu)./fr;

                 elseif q==8; @ MODIFIED LOGIT LINK @
                        i=1; do until i==n+1;
                           if th[i]>13.8;
                             mu[i]=1-1e-6;
                             dm[i]=1e+6;
                             V[i]=1e-6;
                           elseif th[i]<-13.8;
                             mu[i]=1e-6;
                             dm[i]=1e+6;
                             V[i]=1e-6;
                           else;
                             mu[i]=exp(th[i])/(1+exp(th[i]));
                             dm[i]=1/((mu[i]).*(1-mu[i]));
                             V[i]=mu[i].*(1-mu[i]) ./fr[i];
                           endif;
                          i=i+1;
                         endo;
                 else;
                 endif;

                   z=th+(y-mu).*dm-off;      @ Taylor expansion @
                   W=sqrt((dm.^2).*V);  @ sqrt of weight vector n x 1 @


                    @ th^2 is the added variable @
                   Xb=(X~(th.^2))./W;   @ Xb n x (p+1) matrix  @
                   CXb=Xb'Xb;
                   if rank CXb < p+1;
                      goto end;
                   else;
                   endif;
                   Cb=inv(CXb);
                   @ b_add has (p+1) columns @
                   b_add=Cb*Xb'(z./W); @ new estimate of \beta @

               /* Store the last element of b_add
               in the first column of matrix Tl */
                @ "b_add";    b_add; wait; @
               Tl[1,1]=b_add[p+1];
               /* Standard error and t statistics stored in matrix B */
               Tl[1,2]=sqrt(phi*Cb[p+1,p+1]);
               Tl[1,3]=Tl[1,1]/Tl[1,2];
              @  pvalue of t ratio (2 sided test)  @
               Tl[1,4]=2*cdftc(abs(Tl[1,3]),n-p-1);
      end:
  retp(Tl);
 endp;

