/*
**  SIMENV.G
**  written by MARCO RIANI
**
**  FORMAT                      PURPOSE                     Line
** =======================================================================
** {ts}=simenv(X,y,b,d)         To produce a simulation envelope for qqplots
**
*/

/*
**
** SIMENV
**
** PURPOSE:   To Produce a simulation envelope for quantile quantile plot
**
** FORMAT:   {ts}=simenv(X,y,b,d)
**
** INPUT:     X = n x p matrix of explanatory variables
**            y = n x 1 vector containing response
**            b = scalar=0 of (p+1) x 1 vector
**                if  b== scalar 0 we use OLS estimate of b
**                        and s (s=standard error of regression)
**                else
**                    b[1:p] contains estimate of b to be used for
**                    computing fitted values
**                    b[p+1] estimate of (s) standard error of regression
**            d = scalar dummy variables
**                if d==1 qqnorm plot of standardized residuals with a
**                    simulation envelope is produced on the screen
**                else no plot is produced
**
** OUTPUT:    ts =  n x 5 matrix which contains the inputs for
**                  the qqnorm plot with simulation envelope
**                  1st col=  indexes of the units
**                            the order is based on X values
**                  2nd col=  input data for X axis
**                            We use F^{(-1)} (i-3/8)/(n+1/4)
**                            where F^{(-1)} is the inverse of the
**                            cumulative normal distribution
**                  3rd col=  input data for y axis
**                            (standardized resdiuals)
**                  4th col=  lower envelope for standardized res.
**                  5th col=  upper envelope for standardized res.
**
**
** SEE ALSO  QQNORM.G
**
*/


proc (1)=simenv(X,y,b,d);
    local n,p,C,H,M,s,yhat,res,Sres,i,epsilon,eo,e1,tst,low,up,ts;
       n=rows(X);
       p=cols(X);
             C=invpd(X'X);
             H=X*C*X';
             M=diagrv(-H,1-diag(H));
          if b==0;
             b=C*X'y;
             s=0;
          else;
             s=b[p+1];
             b=b[1:p];
          endif;
           @ yhat= fitted values, res=simple residuals @
           yhat=X*b;
           res=y-yhat;
           if s==0;
              s=sqrt(moment(res,0)/(n-p));
           else;
           endif;
           @ Vector Sres contains standardized residuals @
           Sres=res./(s*(diag(M)));
           epsilon=rndn(n,100);
           i=1; do until i==101;
             epsilon[.,i]=sortc(M*epsilon[.,i],1);
            i=i+1;
           endo;
            e1=zeros(n,2);
            i=1; do until i==n+1;
               eo=sortc(epsilon[i,.]',1);
               e1[i,.]=eo[5]~eo[95];
              i=i+1;
            endo;
            @ Quantile plot of Standardized residuals @
            {tst}=qqnorm(Sres,0);
            @ Quantile plot of lower simulation envelope @
            {low}=qqnorm(e1[.,1],0);
            @ Quantile plot of upper simulation envelope @
            {up}=qqnorm(e1[.,2],0);
              if d==1;
                 scale(tst[.,2],e1[.,1]|e1[.,2]|tst[.,3]);
                 begwind;
                 window(1,1,0);
                 makewind(80,6.8555,0,0,1);
                         _plctrl=-1;
                         xy(tst[.,2],tst[.,3]);
                         _plctrl=0;
                         _pltype=6;
                         xy(low[.,2],low[.,3]);
                         xy(up[.,2],up[.,3]);
                 nextwind;
                 endwind;
              else;
              endif;
      ts=tst~low[.,3]~up[.,3];
     retp(ts);
endp;


