/*
**  UNIBIV
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {fre}=unibiv(X,rf,medb,d);            To detect univariate        58
**                                       and bivariate outliers.
**
*/

/*
**
** UNIBIV
**
** PURPOSE:  to detect univariate and bivariate outliers
**
** FORMAT:    {fre}=unibiv(X,rf,medb,d);
**
** INPUT:     X     = N x p matrix of the data
**            rf    = constant (>0) which specifies the confidence
**                    level of the bivariate ellipses
**                    (it must be chosen according to the F
**                    distribution with 2 and N-2 d.f.)
**           medb   = p x p matrix which contains the coordinates
**                    of the bivariate (univariate) medians for each
**                    couple
**                    of variables (the upper triangular part
**                    contains x coordinates, the lower triangular
**                    the y coordiantes)
**            d     = scalar, if d is equal to 1 a plot
**                    which contains univariate standardized
**                    boxplots on the main diagonal and bivariate
**                    confidence ellipse out of the main diagonal
**                    is produced on the screen. If d is <> 1 no
**                    plot is produced
**
**
** OUTPUT:    fre   = k x 3 matrix. The first column contains the
**                    indexes of the units which at least once have
**                    been declared univariate or bivariate outliers
**                    The second and third column respectively contain
**                    the number of times in which such
**                    units have been declared univariate and bivariate
**                    outliers
**
**
** REMARK1: Matrix medb can be created with procedure MEDB.G
**
** REMARK2: The lower and upper limits for the axes in the program
**          are controlled by the scalar v. The current value is v=4.
**
** WARNING: if d=1 GRAPHSET is invoked
**
*/


  proc (1)=unibiv(yyy,rf,medb,d);
  local a,aut,b,biv,bivT,c,cc,ch,cinf,cost,costh,DI,e,em,emax,
        f,fuo,i,il,jl,jk,k,lin,m,mess1,n,new1,new2,no,out,outy,p,psi,
        r,ran,rot,r1,r2,R1max,R2max,s,sinth,Sx,Sy,seq,th,The1,The2,
        The1max,The2max,ti,tinf,ts,tsup,Tx,Ty,univ,univT,v,x,xaxis,
        Xs,Xmax,XX,X025,Xmed,x075,xnew,xst,y,yaxis,Ys,Ymax,YY,ynew,
        ysim,ter,vert,vert2,vert3,vert4,z,fre,pms,me1,j;

n=rows(yyy);
p=cols(yyy);


  if rf<=0;
   errorlog "ERROR: The confidence threshold must be greater than 0";
      end;
   else;
  endif;

 v=4.0;  @ v specifies xlim and ylim for all the plots @

/***** bivT contains the frequency distribution of biv. outliers  ***/
/****  univT contains the frequency distribution of univ. outliers ***/
bivT=zeros(n,1);
univT=zeros(n,1);

seq=seqa(1,1,n);

S=6.855;  /***** DIMENSION OF THE SCREEN ****/
k=1;
      if d==1;
                  begwind;
       else;
      endif;
il=p-1;  DO UNTIL  il==-1; /* il is linked to the rows */
jl=0;   DO UNTIL  jl==p;   /* jl is linked to columns */
       if d==1;
                  makewind(9/p,S/p,9/p*jl,S/p*il,1);
       else;
       endif;
 x=yyy[.,jl+1]; y=yyy[.,k];


 if jl+1==k;
     @ then produce a vertical bosplot @
                       @ xst contains x standardized values @
                       xst=(x-meanc(x))/stdc(x);
                           @ x= original data @
                       xs=sortc(xst,1);  @ xs= sorted original data @

                               c=n*0.25;
                               cc=round(c);
                               cinf=xs[cc];
                     if (c-round(c))==0;
                              X025=xs[c];
                     else;
                              X025=(xs[cc]+xs[cc+1])/2;
                     endif;         @  X025  is the first quartile @


                               c=n;
                     if (c/2-round(c-c/2))==0;
                               Xmed=(xs[c/2]+xs[c/2+1])/2;
                     else;
                               Xmed=xs[(c+1)/2];
                     endif;    @  Xmed  is the median  of vector X @

                                c=n*0.75;
                                cc=round(c);
                                cinf=xs[cc];
                                z=0;
                     if (c-round(c))==0;
                              X075=xs[c];
                     else;
                              X075=(xs[cc]+xs[cc+1])/2;
                     endif;     @  X075  is the third quartile @

                             DI=X075-X025;

                             Ts=X075+DI*1.5;  @ upper truncation point @
                             Ti=X025-DI*1.5;  @ lower truncation point @
                     if Ts>xs[n];
                              Tsup=xs[n];
                     else;
                              @ Find upper truncation point @
                              i=1; do until i==n+1;
                               if xs[n-i]>Ts;
                               else;
                                Tsup=xs[n-i];
                                i=n;
                               endif;
                                i=i+1;
                               endo;
                     endif;
                     if Ti<xs[1];
                              Tinf=xs[1];
                     else;

                              @ Find lower truncation point @
                              i=1; do until i==n+1;
                               if xs[i]<Ti;
                               else;
                                Tinf=xs[i];
                                i=n;
                               endif;
                                i=i+1;
                               endo;
                     endif;
                          @   "statis";
                        tinf|X025|X075|tsup; wait;   @

                @ check for univariate outliers  @
                @ outy is a (l+1) x 2 matrix. Apart from the first row
                  which contains two zeros, the first column contains
                  the indexes of the units decleared univariate
                  outliers, the second columns reports the values @
                outy=0~0;
               i=1;    do until i==n+1;
                       if xst[i]<Tinf or xst[i]>Tsup;
                           outy=outy|(i~xst[i]);
                       else;
                       endif;
                 i=i+1;
               endo;
                @ univT contains the frequency distribution of univ.
                  outliers   @
                 if rows(outy)>1;
                            univ=zeros(n,1);
                        i=2;    do until  i==rows(outy)+1;
                           univ[outy[i,1]]=1;
                           i=i+1;
                        endo;
                           univT=univT+univ;
                 else;
                 endif;

    if d==1;
        @ lines for drawing boxplot are produced @
        ter=Tinf|x025|xmed|x075|Tsup;
        lin=ones(5,1)~6*ones(5,1)~-1*ones(5,1)~ter~1*ones(5,1)~ter
        ~zeros(5,1)~2*ones(5,1)~0.2*ones(5,1);
         vert=1~6~0~Tinf~0~x025~  0~2~0.2;
         vert2=1~6~-1~x025~-1~x075~0~2~0.2;
         vert3=1~6~1~x025~1~x075~  0~2~0.2;
         vert4=1~6~0~x075~0~Tsup~  0~2~0.2;
        _pline=lin|vert|vert2|vert3|vert4;
         ytics(-v,v,1,0);
         xtics(-2,2,1,0);

         /** the following lines put stars for outliers in boxplot **/
            if rows(outY)>1;
                      no=rows(outY);
                      outY=outY[2:no,.];
                     @ _psym is the matrix which contains the stars @
                     /* third is symbol type  fourth is the height 
                     7th is line thickness ***/
                     _psym=zeros(no-1,1)~outY[.,2]~
                          12*ones(no-1,1)~7*ONES(no-1,1)~
                          1.5*ONES(no-1,1)~ONES(no-1,1);
                          @ xloc yloc symboltype height colour  @
                      @
                      i=1;  do until i==rows(outy)+1;
                           _pmsgctl=1.5~(xst[outy[i,1]]-0.15)~0.3~0~1~2;
                               mess1= 0 $+ ftocv(outy[i,1],1,0);
                              _pmsgstr=mess1;
                               xy((-2|-2),(minc(y)-10)|(maxc(y)+10));
                             i=i+1;
                       endo;    @

                        pms=zeros(1,6);
                        me1=" ";
                        i=1;  do until i==rows(outy)+1;
                          pms=pms|(1.5~(xst[outy[i,1]]-0.15)~0.3~0~1~2);
                          me1= me1 $+ "\000" $+ ftocv(outy[i,1],1,0);
                             i=i+1;
                        endo;
                              _pmsgctl=pms;
                              _pmsgstr=me1;
                   xy((-2|-2),(minc(y)-10)|(maxc(y)+10));

            else;
                 @ if there is no univariate outlier just draw the box @
                           _psym=zeros(7,1);
                              _pmsgstr="";
                              _pmsgctl=0;
                   xy((-2|-2),(minc(y)-10)|(maxc(y)+10));
            endif;

            /**** end of univariate part ***/
    else;
    endif;
           goto pippo;
     @ after producing a boxplot go to pippo and skip the bivariate
       statements @

else;

       /*** beginning of bivariate part ***/


Ty=medb[k,jl+1];
Tx=medb[jl+1,k];
/**** Tx e Ty are the coordinates of the Bivariate  Median ****/

Sx=sqrt((x-Tx)'(x-Tx)/(n-1));      @ Sx is the robust standard deviation of x @
Sy=sqrt((y-Ty)'(y-Ty)/(n-1));      @ Sy is the robust standard deviation of y @

R=(x-Tx)'(y-Ty)/((n-1)*(Sx*Sy));   @ robust correlation @

Xs=(x-Tx)/Sx;   @ robust standardization of x_i  @
Ys=(y-Ty)/Sy;   @ robust standardization of y_i  @
/* Now we calculate Mahalanobis distances with centroid defined by the L_1 med*/
E=sqrt((Xs.^2+Ys.^2-2*R*Xs.*Ys)/(1-R^2));

@ ran is  a vector which contains sorted Mahalanobis distances from
  the robust bivariate centroid @
ran=sortc(E,1);
if (n/2-round(n-n/2))==0;
Em=(ran[n/2]+ran[n/2+1])/2;
else;
Em=ran[(n+1)/2];
endif;         @  Em  is the median  of vector E @
/** the square of the MAhl. distance is distributed like an F **/

eMAX= (2*(n-1))/(n-2)* rf;    /***** rf is the constant ****/
emax=sqrt(emax);               /***** of the F distribution ****/


/*** hinge=ellipsis containing 50% of the values ***/
R1=Em*sqrt((1+R)/2);
R2=Em*sqrt((1-R)/2);
th=seqa(0,2*pi/3600,3600);
The1=R1*cos(th);
The2=R2*sin(th);
Xx=(The1+The2);
Yy=(The1-The2);

/*** fence=ellipsis containing (1-\alpha)% of the values ***/
R1max=Emax*sqrt((1+R)/2);
R2max=Emax*sqrt((1-R)/2);
The1max=R1max*cos(th);
The2max=R2max*sin(th);
Xmax=(The1max+The2max);
Ymax=(The1max-The2max);

/***** Xs and Ys are the robust standardized values ******/
f=Xs~Ys;

/***** rotation of the coordiantes ******/
cost=emax^2;
a=1/(cost*(1-r^2));
c=a;
b=-r*a;


aut=EIG((a~b)|(b~c));
costh=b/sqrt((c-aut[2])^2+b^2);
sinth=(aut[2]-c)/sqrt((c-aut[2])^2+b^2);

/**** M is the orthogonal matrix which enables the rotation of the axes ****/
M=(costh~sinth)|(-sinth~costh);

fuo=SQRT(ABS(1/aut[1]-1/aut[2]));


ROT=(xmax')|
    (ymax');
/*** NEW1 = 2 x n matrix which contains the coordinates of the rotated
ellipse ***/ new1=M*rot;
xnew=new1[1,.]';  @ xnew = n x 1 vector which contains x coord. of
                                 rotated points @
ynew=new1[2,.]';
                  @ ynew = n x 1 vector which contains y coord. of
                                 rotated points @

/*** NEW2 = 2 x n matrix which contains the coordinates of the rotated
points ***/
NEW2=M*(f');


/* ch is the fixed distance from the fires of the ellipse */
   i=10;   @ one could choose i= integer <= nn @
ch=sqrt((xnew[i]+fuo)^2+ynew[i]^2)+sqrt((xnew[i]-fuo)^2+ynew[i]^2);

/*** biv is the 1 x n vector which contains potential outliers
      for a particular pair of variables ***/
 biv=zeros(1,n);

jk=1; do until jk==n+1;
if sqrt((new2[1,jk]+fuo)^2+new2[2,jk]^2)
+sqrt((new2[1,jk]-fuo)^2+new2[2,jk]^2)>ch;
biv[1,jk]=biv[1,jk]+1;
        else;
    endif;
    jk=jk+1;
endo;
/***** bivT contains cumulative distribution of bivariate outliers
*****/ bivT=bivT+biv';

/**** the following lines plot the hinge together with the fence ***/
     if d==1;
 xtics(-v,v,1,1);
 xaxis=1~6~0~-10~0~100~0~7~0.2;   @ Cartesian axes @
 yaxis=1~6~-10~0~100~0~0~7~0.2;
_pline=xaxis|yaxis;


  /**** the following lines plot the ellipsis ***/
@ ysim is a g x 1 vector which contains the indexes of the units
  which are bivariate outliers @
ysim=delif(seq,biv'.==0);
ysim=missrv(ysim,235455);
  @ _psym indicates the location of the points in the main plot @
_psym=F~1*ONES(n,1)~ZEROS(n,1)~6*ONES(n,1)~
1.5*ONES(n,1)~10*ONES(n,1);

               if ysim[1].==235455;
                  _pmsgstr="";
                  xy(Xx~xmax,Yy~ymax);
               else;


                        pms=zeros(1,6);
                        me1=" ";
                        i=1;  do until i==rows(ysim)+1;
                          pms=pms|xs[ysim[i]]~ys[ysim[i]]~0.3~0~1~2;
                          me1= me1 $+ "\000" $+ FTOCV(YSIM[i],1,0);
                             i=i+1;
                        endo;
                           @   _pmsgctl=pms;
                              _pmsgstr=me1; @
                              _pmsgctl=0;
                              _pmsgstr="";
 xy(Xx~xmax,Yy~ymax);
     else;
     endif;
               endif;

         pippo: /* it is linked to the statement  k=jl+1 */

    jl=jl+1;   @ jl is linked to columns @
      if d==1;
                               nextwind;
      else;
      endif;
                                                      endif;
                                               /* endif of k=jl+1 */

          endo;   /*****  k is the kth col of dataset *****/
      k=k+1;   il=il-1;     @ il is linked to rows @
   endo;
      if d==1;
             endwind;
    graphset; @ reset all graphical parameters @
      else;
      endif;
    fre=seq~univT~(bivT/2);
    
    @ Note that delif(fre,fre[.,2].==0 .and fre[.,3].==0)
    removes all the rows always inside univariate fre[.,2]
    and bivariate fre[.,3] 
    if you want to remove only the bivariate outliers 
    remove fr[.,2] @
     @ fre=delif(fre,fre[.,2].==0 .and fre[.,3].==0); @
   
    fre=delif(fre,fre[.,3].==0);
  retp(fre);
    endp;

