/*
**  RFWDMLES
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {Plik}=rfwdmleS(Y,X,bsb,la);            To create a matrix which
**                                     contains MLE of transformation
**                                     parameter in each step
**                                     of the forward search
**                                     NOTE THAT WE SEARCH FOR
**                                     A COMMON LAMBDA FOR ALL THE
**                                     VARIABLES
**
*/

/*
**
** FWDMLE
**
** PURPOSE:  to create a vector wchih contains Max. Lik. Est.
**           of common transformation parameter at each step of
**           the forward search
**
**
** FORMAT:    {Plik}=fwdmleS(Y,X,bsb,la);
**
** INPUT:     Y     = n x k matrix of response variables
**            X     = n x p matrix of the data
**           bsb    = m x 1 vector which contains the indexes of the
**                    units forming the initial subset
**            la    = scalar which specifies  which value of
**                    lambda must be used for ordering the units
**                    in all steps of the forward search
**
** OUTPUT:    Plik  = (n-m+1)  x 2 matrix. In the first column
**                    it contains the dimension of the basic subset used
**                    in the forward search. In column 2 we
**                    can find max. lik. estimate of the common
**                    transformation
**                    parameter in each step of the forward search
**
** REMARK:  the input vector bsb can be conveniently created by
**          program UNIBIV.G
**          The transformtion used in the maximization is 2.5*sin(\lambda)
**          The initial \lambda values are =1 (no transformation)
**
**
** See also: FDWMLES,FLR,FWDMLER,FWDMLEQD,FWDMLELD,MULTOUT,FWDMLE
**
** WARNING: 2 global variables named Xglobal and Yglobal
**          are created
**
*/

 clearg Xglobal,Yglobal,chglobal;


proc (1)=RassS(la0);    /* procedure which transforms the parameters */
     local la;
                        @  la=  3*(sin  (  la0  ));
                           la=3*la0/sqrt(1+la0^2);   @
                        @  la= 5*( 1-exp(la0))/(1+exp(la0)) ; @
                          la=  la0;
    retp(la);
endp;


proc (1)=RlikS(la);   /* procedure which computes max. lik. */
     local G,Xg,Yg,Yh,Eg,dZ;
           Xg=Xglobal; Yg=Yglobal;

             {la}=RassS(la);
                   G=exp(meanc(ln(Yg)));
                   G=G'; @ Vector of geometric means @
             @ transform  Yg[.,chglobal] @
       Yg=(Yg^(la)-1)./(la.*G.^(la-1));


       Yh=Xg*inv(Xg'Xg)*(Xg')*Yg;
       Eg=Yg-Yh;

       dZ=100*det(vcx(Eg));

   retp(dZ);
endp;


proc (1)=Rfwdmles(Y,X,bsb,la);
       local nn,seq,Plik,mm,la0,zs,mala,min,dZ,E,B,Xb,Yb,Ytb,Yt,sel,
             psi,fvalue,grad,retcode,vc,i,ym,psin,p,pp,geom,prov,j;
          nn=rows(X);
          p=cols(X);
          seq=SEQA(1,1,nn);
          mm= rows(bsb);
          min=mm;
          pp=cols(Y);


          Plik=seqa(mm,1,nn-mm+1)~ones(nn-mm+1,1);

          /*  calculation of Mahal. dist. on initial clean data set */
           mala=seq~zeros(nn,1);
           la0=1;
           la0=sin(la)+1e-4;

          @ Create Yt=matrix of transformed values to be used
             in the forward search (for ordering MD) @
          Yt=Y;

           if la NE 0;
                  Yt= Y^la;
              else;
                  Yt=ln(Y);
             endif;


       mm=mm; do until mm==nn+1;  @ nn= numero tot. di unita' stat. @
           @     "mm"; mm;
               rows(bsb);
                sumc(bsb); vcx(bsb); wait;  @

               if mm>min;
                  la0=psi;
               else;
               endif;

                  Xglobal=X[bsb,.];
                  Yglobal=Y[bsb,.];
                   Xb=Xglobal;
                   Yb=Yglobal;
                  {dZ}=Rliks(la0);
                  Ytb=Yt[bsb,.];
                  {psi,fvalue,grad,retcode}=OPTMUM(&Rliks,la0);
                   @ Retransform the MLE parameters @
                            {psin}=RassS(psi);
                    Plik[mm-min+1,2]=psin;

                B= invpd(Xb'Xb)*Xb'*Ytb;
                @ residuals for the overall sample
                  using transformed Y @
                E=Yt-X*B;
               if cols(E)>1;
                    @ Search on Mahalanobis distances @
                       ym=meanc(E[bsb,.]);
                      vc=invpd(vcx(E[bsb,.]));
                   @ Mahalanobis dist. on residuals @
                                    i=1;  do until i==nn+1;
                         mala[i,2]=sqrt((E[i,.]-ym')*vc*(E[i,.]'-ym));
                                         i=i+1;
                                     endo;
                 else;
                  mala[.,2]=abs(E);
                 endif;
                      zs=sortc(mala,2);
                        @ the dimension of the clean data set increases
                          of one unit. vector bsb contains the
                          indexes corresponding to the units of
                          the new basic subset @
                                if mm<nn;
                                    bsb=zs[1:mm+1,1];
                             @  setdif(seqa(1,1,nn),bsb,1); wait;  @
                                 else;
                                 endif;

             mm=mm+1;
          endo;
  retp(Plik);
endp;

