/*
**  SCOMR (Multivariate version of the score test)
**  written by MARCO RIANI
**
**  FORMAT                              PURPOSE                     Line
** =======================================================================
** {Sco,Un}=ScomR(Y,X,W,bsb,ch,r);
**
**                                To monitor the multivariate version
**                                of the score test
**                                in every step of the forward procedure
**                                The purpose is to test the null
**                                hypothesis that additional variables given in the columns
**                                of matrix W
**                                do not significantly explain Y
**
**
*/

/*
** SCOMR
**
** PURPOSE:  The principal purpose is to monitor the likelihood ratio in
**           every step of the forward procedure. The purpose is to test
**           that the additional variables given in the cols of matrix W do not significantly
**           explain Y
**
** FORMAT:   {Sco,Un}=scomR(Y,X,W,bsb,ch,r);
**
** INPUT:     Y = n x v data matrix n observations and v  response var.
**            X = n x p data matrix n observations and p expl. variables
**            W = n x p data matrix additional n observations and p
**                expl. variables
**          bsb = m x 1 vector. It contains
**                the indexes of the m units which
**                form the initial subset
**                Initial centroid is determined using only
**                these units.  (1 \leq m \leq n)
**       ch     = c x 2 matrix or scalar 0
**                1st column = indexes of the variables (VAR)
**                             to be tested for transformation
**                2nd column = values of transformation
**                             parameters
**                REMARK: if ch=0 (scalar zero) you can test
**                         the null hypothesis that no variable
**                         must be transformed
**                EXAMPLE: if ch=(1~0)|
**                               (4~0.5)
**                         LR concerns VAR 1 and 4
**                         The null hypothesis is:
**                         log transformation for VAR 1
**                         and sqrt transformation for VAR 4
**
** OUTPUT:
**            sco = (n-m+1 x 3) matrix.
**                 1st col = fwd search index
**                 2nd col = Multivariate Score test based on
**                           t statistic after first iteration
**                 3rd col = Multivariate Score test based on t
**                           statistics after final convergence
**                 4th col = Test on covariance matrices
**                           after final convergence without correction
**                 5th col = Test on covariance matrices
**                           after final convergence with correction
**                 6th col = Lik ratio test after first iteration
**                           m lm |\Sigma_0| / |\Sigma_1|
**                           |\Sigma_0| = cov matrix of residuals without
**                           additional constructed expl. variables
**                           |\Sigma_0| = cov matrix of residuals with
**                           additional constructed expl. variables
**                 7th col = Lik ratio test after final iteration
**                           m lm |\Sigma_0| / |\Sigma_1|
**            Un = (n-m) x 11 Matrix which contains the unit(s)
**                 included in bsb at each step of the fwd search
**                 REMARK: in every step the new subset is compared
**                 with the old subset. Un contains the unit(s)
**                 which are in the new subset but not in the old one
**
**
*/


 proc (2)=scomR(Y,X,W,bsb,r);

   local seq,ym,vc,la,Xt,sel,mm,
       nn,p,v,i,zs,min,sco,In,Un,Unit,mala,nbsb,sj,Ysj,Ivcw,rl,rU,
       j,G,E,Eb,B,Bj,Ew,Yt,Wb,Yb,Ytb,Xb,Xw,Vcw,eige,Enew,Ball,
       XpX,Xpy,num,den,iter,bold,df;

nn=rows(X);
seq=SEQA(1,1,nn);

if X==miss(0,0);
   X=seq;
 else;
endif;

p=cols(X);
v=cols(Y);

rl=r[.,1];
rU=r[.,2:cols(r)];

mm= rows(bsb);
min=mm;

  @ Matrix lr which will contain the LR test in each step of the
    forward search @
  sco=seqa(mm,1,nn-mm+1)~zeros(nn-mm+1,6);

  @ matrix In will contain in the 2nd col. the unit with greatest MD
    in basic subset, in third col the one with the smallest MD among
    Noclean, in 4th col the unit with second smallest MD among nocl. @
   In=seqa(mm,1,nn-mm+1)~miss(zeros(nn-mm+1,10),0);

   if nn>mm;
      @  matrix Un will contain from 2nd col. to 10th the unit(s) which
      entered the basic subset @
      Un=seqa(mm+1,1,nn-mm)~miss(zeros(nn-mm,10),0);
   else;
      Un=miss(0,0);
  endif;


/*** mala in the column will contain MD  ***/
mala=seq~zeros(nn,1);



          @ Create Yt = matrix of transformed values to be used
             in the forward search (for ordering MD) @
          Yt=Y;


mm=mm; do until mm==nn+1;

  Yb=Y[bsb,.];
  Xb=X[bsb,.];
  Wb=W[bsb,.];

  Ytb=Yb;



                B= invpd(Xb'Xb)*Xb'*Ytb;
                @ residuals for subset
                  using transformed Y @
                Eb=Ytb-Xb*B;


                @ Now compute residuals for each regression
                  using additional constructed variables @
                Ew=Eb;
                j=1;  do until j==v+1;
                   Xw=Xb~Wb[.,j];
                   Bj=invpd(Xw'Xw)*Xw'*Ytb[.,j];
                   Ew[.,j]=Ytb[.,j]-Xw*Bj;
                  j=j+1;
                endo;


               Vcw=vcxp(Ew)/1;    @ \Sigma_w = S  @
               Vc=vcxp(Eb)/1; @ \Sigma_0  @
                    @   "Vc Vcw"; Vc; Vcw; wait;  @

           bold=zeros((p+1)*v,1);
           iter=1; do until iter==101;
               IVcw=invpd(Vcw);

                @ Now apply SUR using Vcw; @

                 @ XpX= inv X'X @
                 XpX=zeros((p+1)*v,(p+1)*v);
                 @ Xpy= X'y @
                 Xpy = zeros((p+1)*v,1);

                  i=1; do until i==v+1;
                        j=1; do until j==v+1;

                            @ XpX=inv X'X @
                            XpX[(p+1)*(i-1)+1:(p+1)*i,
                                 (p+1)*(j-1)+1:(p+1)*j]=
                                  IVcw[i,j]*(Xb~Wb[.,i])'(Xb~Wb[.,j]);

                            @   Xpy= X'y @
                            Xpy[(p+1)*(i-1)+1:(p+1)*i,1]=
                            Xpy[(p+1)*(i-1)+1:(p+1)*i,1]+
                                  IVcw[i,j]*(Xb~Wb[.,i])'Ytb[.,j];
                             j=j+1;
                          endo;
                       i=i+1;
                    endo;
                           XpX=invpd(XpX);

                          ball=XpX*Xpy;

                       Enew=Ew;
                         j=1; do until  j==v+1;
                           Enew[.,j]=Ytb[.,j]-(Xb~Wb[.,j])*
                              ball[(j-1)*(p+1)+1:j*(p+1)];
                           j=j+1;
                          endo;

                   Vcw=vcxp(Enew)/1;
                  @ "tolerance";
                   mm; moment(ball-bold,0);  wait;
                   @
                if iter==1;
                  @ Now compute the multivariate score test
                  based on the first iteration @
                  num=(rl-rU*ball)'invpd(rU*XpX*rU')*(rl-rU*ball);
                  IVcw=inv(Vcw);
                  den=0;
                  i=1; do until i==v+1;
                       j=1; do until j==v+1;
                           den=den+Enew[.,i]'Enew[.,j]*IVcw[i,j];
                          j=j+1;
                        endo;
                       i=i+1;
                   endo;
                    df=((mm*v-(p+1)*v)/rows(r));
                   Sco[mm-min+1,2]= @ df * @ num @ /den @ ;


                   @ Store Lik. ratio test using first iteration @
                   Sco[mm-min+1,6]=mm*ln(det(Vc)/det(Vcw));

                else;
                endif;

                if moment(ball-bold,0)< 1e-3;
                  iter=100;
                else;
                  bold=ball;
                endif;
             iter=iter+1;
           endo;

              @ Now compute using final estimates of ball the test @
              num=(rl-rU*ball)'invpd(rU*XpX*rU')*(rl-rU*ball);
              IVcw=inv(Vcw);
              den=0;
              i=1; do until i==v+1;
                   j=1; do until j==v+1;
                       den=den+Enew[.,i]'Enew[.,j]*IVcw[i,j];
                      j=j+1;
                    endo;
                   i=i+1;
               endo;

               Sco[mm-min+1,3]= @ df  * @ num @ /den @ ;

              @ Store Lik. ratio test using first iteration @
              Sco[mm-min+1,7]=mm*ln(det(Vc)/det(Vcw));


                  @ UPDATING PART: order according to power
                    transformation not to Box Cox transformation @

                B= invpd(Xb'Xb)*Xb'*Yt[bsb,.];
                @ residuals for the overall sample
                  using transformed Y @
                E=Yt-X*B;

                    @ Search on Mahalanobis distances @
                      ym=meanc(E[bsb,.]);
                      vc=invpd(vcx(E[bsb,.]));
                        @ already computed @
                   @ Mahalanobis dist. on residuals @
                                    i=1;  do until i==nn+1;
                         mala[i,2]=sqrt((E[i,.]-ym')*vc*(E[i,.]'-ym));
                                         i=i+1;
                                     endo;

                      zs=sortc(mala,2);
                        @ the dimension of the clean data set increases
                          of one unit. vector bsb contains the
                          indexes corresponding to the units of
                          the new basic subset @
                                if mm<nn;
                                    nbsb=zs[1:mm+1,1];
                               @ Store in matrix Un the unit(s)
                               included in this step @
                               unit=setdif(nbsb,bsb,1);
                               if rows(unit)<=10;
                                UN[mm-min+1,2:rows(unit)+1]=(unit');
                               else;
                                "Warning: interchange greater than 10";
                                UN[mm-min+1,2:11]=unit[1:10]';
                               endif;
                           bsb=nbsb;


                                 else;
                                 endif;




             mm=mm+1;
          endo;
  retp(Sco,Un);
endp;

